(*
 * Convert FIR to CG code.
 * CG code is very close to the machine.
 *
 * Memory representation:
 *
 * 1. There is a "pointer table" that points to all
 *    valid live blocks in memory.
 *
 *    The pointer table has three descriptors:
 *       Frame.pointer_base: start of the pointer table
 *       Frame.pointer_next: index of the next free entry
 *       Frame.pointer_free: number of free entries left
 *       Frame.pointer_size: total number of entries in the pointer table
 *    Free entries are saved in a linked list.
 *
 * 2. Each memory block contains arbitrary "stuff",
 *    beginning with two words:
 *
 *    .---------------------------.
 *    | index into pointer table  |
 *    |---------------------------|
 *    | size (in words)           |
 *    |---------------------------|
 *    | word_0                    |
 *    |---------------------------|
 *    | ...                       |
 *    |---------------------------|
 *    | word_{size - 1}           |
 *    "---------------------------"
 *
 *    Every unsafe memory operation is checked that
 *    it reads/writes from the data area of a valid block.
 *
 *    A block is valid if it begins with an index into the
 *    pointer table, and the pointer table points back to
 *    the block.  The data area may be arbitrarily overwritten
 *    by the C program.
 *
 * 3. A pointer stored in memory is stored as two words
 *
 *    .---------------------------.
 *    | index into pointer table  |  B
 *    |---------------------------|
 *    | offset (in bytes)         |  O
 *    "---------------------------"
 *
 *    A pointer is valid if the pointer index refers to
 *    a valid pointer.
 *
 *    To make it less likely that the C program can "construct"
 *    pointers on the fly, we hash the fields.
 *
 *    The Frame contains a hash index H.
 *    The B-field contains B^H
 *
 * 4. There is a "closure table" that has an entry for each
 *    higher-order function.  Each entry has the address of
 *    the function, as well as the size of the frame argument
 *    that it expects.
 *
 *    The closure table is static: it contains an entry for
 *    each global/continuation function.
 *
 *    The closure table has two descriptors:
 *       Frame.closure_base: start of the closure table
 *       Frame.closure_size: number of entries in the table
 *
 * 5. A closure is stored in memory as two words
 *
 *    .---------------------------.
 *    | index into pointer table  |  F
 *    |---------------------------|
 *    | index into closure table  |  I
 *    "---------------------------"
 *
 *    A closure is valid
 *       a. if the pointer refers to a valid block F
 *       b. if the closure index refers to a valid function (f, size)
 *       c. if block D has the same size as the function expects
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2001 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Debug
open Symbol
open Fc_ir_type
open Fc_ir_exn_type
open Fc_ir_exn
open Fc_ir_env
open Aal_type
open Frame_type

(*
 * The code is parameterized by the Frame.
 *)
module MakeAal (Frame : FrameSig) =
struct
   (************************************************************************
    * MULTI-VARIABLE ANALYSIS
    ************************************************************************)

   (*
    * At run time, we pass around some objects as implicit
    * tuples using sets of variables.
    *
    * Unsafe pointers are passed as base/offset/bounds triples.
    * Closure are passed as frame/label pairs.
    * Safe pointers are passed around as values.
    * All else are passed as values.
    *)
   type ty_var = var

   type infer_type =
      TyInferPointer of ty_var
    | TyInferClosure
    | TyInferScalar

   (*
    * Multi environment keeps track of the var_type for each variable.
    *)
   type run_subst =
      SubstVar of ty_var
    | SubstBool of bool

   type rtenv =
      { rtenv_types : run_subst SymbolTable.t;
        rtenv_funs : var list SymbolTable.t;
        rtenv_vars : infer_type SymbolTable.t
      }

   let rtenv_empty =
      { rtenv_types = SymbolTable.create ();
        rtenv_funs = SymbolTable.create ();
        rtenv_vars = SymbolTable.create ()
      }

   let rtenv_add_fun rtenv f vars =
      { rtenv with rtenv_funs = SymbolTable.add rtenv.rtenv_funs f vars }

   let rtenv_add_ty_var rtenv v flag =
      { rtenv with rtenv_types = SymbolTable.add rtenv.rtenv_types v flag }

   let rtenv_add_var rtenv v ty =
      { rtenv with rtenv_vars = SymbolTable.add rtenv.rtenv_vars v ty }

   let rtenv_lookup_fun rtenv pos f =
      try SymbolTable.find rtenv.rtenv_funs f with
         Not_found ->
            raise (IRException (pos, UnboundVar f))

   let rtenv_lookup_fun_opt rtenv f =
      try Some (SymbolTable.find rtenv.rtenv_funs f) with
         Not_found ->
            None

   let rtenv_lookup_var rtenv pos v =
      try SymbolTable.find rtenv.rtenv_vars v with
         Not_found ->
            raise (IRException (pos, UnboundVar v))

   let rtenv_lookup_ty_var rtenv pos v =
      try SymbolTable.find rtenv.rtenv_types v with
         Not_found ->
            raise (IRException (pos, UnboundVar v))

   (*
    * Normal type and variable environments.
    *)
   let raise_exn pos v =
      IRException (pos, v)

   let tenv_lookup = tenv_lookup raise_exn
   let tenv_expand = tenv_expand raise_exn
   let venv_lookup = venv_lookup raise_exn

   (*
    * Get the type of an atom.
    *)
   let type_of_atom venv pos = function
      AtomChar _ -> TyChar
    | AtomInt _ -> TyInt
    | AtomFloat _ -> TyFloat
    | AtomNil ty -> TyArray (ty, None)
    | AtomVar v -> venv_lookup venv pos v

   (************************************************
    * UTILITIES
    *)

   (*
    * Break apart a function type.
    *)
   let is_fun_type1 tenv pos ty =
      match tenv_expand tenv pos ty with
         TyArray (ty, None) ->
            (match tenv_expand tenv pos ty with
                TyFun _ -> true
              | _ -> false)
       | TyFun _ -> true
       | _ -> false

   let is_pointer_type1 tenv pos ty =
      match tenv_expand tenv pos ty with
         TyArray _ -> true
       | _ -> false

   let dest_fun_type1 tenv pos ty =
      match tenv_expand tenv pos ty with
         TyFun (ty_vars, ty_res) -> ty_vars, ty_res
       | ty -> raise (IRException (pos, NotAFunction ty))

   (************************************************
    * MULTI-TYPE INFERENCE
    *)

   (*
    * Follow a type variable to it's final binding.
    *)
   let rtype_ty_expand rtenv pos v =
      let subst = rtenv.rtenv_types in
      let rec expand v =
         try
            match SymbolTable.find subst v with
               SubstBool flag -> v, flag
             | SubstVar v -> expand v
         with
            Not_found ->
               raise (IRException (pos, UnboundVar v))
      in
         expand v

   (*
    * Unify two types.
    * If both are pointers, take the max of the
    * expansion.
    *)
   let rtype_unify rtenv pos ty1 ty2 =
      match ty1, ty2 with
         TyInferPointer v1, TyInferPointer v2 ->
            let v1, flag1 = rtype_ty_expand rtenv pos v1 in
            let v2, flag2 = rtype_ty_expand rtenv pos v2 in
               if v1 = v2 then
                  rtenv
               else
                  let flag = SubstBool (flag1 || flag2) in
                  let rtenv = rtenv_add_ty_var rtenv v1 flag in
                     rtenv_add_ty_var rtenv v2 (SubstVar v1)
       | _ ->
            rtenv

   (*
    * Lift the expansion so that it is unsafe.
    *)
   let rtype_make_unsafe rtenv pos v =
      let v, _ = rtype_ty_expand rtenv pos v in
         rtenv_add_ty_var rtenv v (SubstBool true)

   (*
    * Make sure the atom is unsafe.
    *)
   let rtype_make_unsafe_atom rtenv pos = function
      AtomChar _
    | AtomInt _
    | AtomFloat _
    | AtomNil _ ->
         rtenv
    | AtomVar v ->
         match rtenv_lookup_var rtenv pos v with
            TyInferPointer v ->
               rtype_make_unsafe rtenv pos v
          | _ ->
               rtenv

   (*
    * Figure the run-type of a normal type.
    * If the type is a pointer, assume it is unsafe.
    *)
   let bool_rtype_of_type flag tenv rtenv pos ty =
      let ty = tenv_expand tenv pos ty in
         match ty with
            TyChar
          | TyInt
          | TyFloat
          | TyAny ->
               rtenv, TyInferScalar
          | TyArray (ty, _) ->
               (match tenv_expand tenv pos ty with
                   TyFun _ ->
                      rtenv, TyInferClosure
                 | _ ->
                      let ty_var = new_symbol_string "pointer" in
                      let rtenv = rtenv_add_ty_var rtenv ty_var (SubstBool flag) in
                         rtenv, TyInferPointer ty_var)
          | TyFun _ ->
               rtenv, TyInferClosure
          | TyStruct _
          | TyId _ ->
               raise (IRException (pos, InternalTypeError (ty, "no run-time type")))

   let safe_rtype_of_type = bool_rtype_of_type false
   let unsafe_rtype_of_type = bool_rtype_of_type true

   (*
    * Get the type of an atom.
    *)
   let rtype_atom rtenv pos = function
      AtomChar _
    | AtomInt _
    | AtomFloat _ ->
         rtenv, TyInferScalar
    | AtomNil _ ->
         let ty_var = new_symbol_string "nil" in
         let rtenv = rtenv_add_ty_var rtenv ty_var (SubstBool true) in
            rtenv, TyInferPointer ty_var
    | AtomVar v ->
         rtenv, rtenv_lookup_var rtenv pos v

   (*
    * Assign a multi-type to each variable.
    *)
   let rec rtype_expr tenv venv rtenv e =
      let pos = string_pos "rtype_expr" (atom_pos e) in
         match e with
            Fc_ir_type.LetAtom (v, ty, a, e) ->
               rtype_atom_expr tenv venv rtenv pos v ty a e
          | Fc_ir_type.LetUnop (v, ty, op, a, e) ->
               rtype_unop_expr tenv venv rtenv pos v ty op a e
          | Fc_ir_type.LetBinop (v, ty, op, a1, a2, e) ->
               rtype_binop_expr tenv venv rtenv pos v ty op a1 a2 e
          | Fc_ir_type.TailCall (f, args) ->
               rtype_tailcall_expr tenv venv rtenv pos f args
          | Fc_ir_type.LetClosure (v, ty, f, args, e) ->
               rtype_closure_expr tenv venv rtenv pos v ty f args e
          | Fc_ir_type.LetExtCall (v, f, ty, args, e) ->
               rtype_extcall_expr tenv venv rtenv pos v f ty args e
          | Fc_ir_type.IntCall (ty, f, args) ->
               rtype_intcall_expr tenv venv rtenv pos ty f args
          | Fc_ir_type.IfThenElse (op, a1, a2, e1, e2) ->
               rtype_if_expr tenv venv rtenv pos op a1 a2 e1 e2
          | Fc_ir_type.SetSubscript (a1, ty, a2, a3, e) ->
               rtype_set_subscript_expr tenv venv rtenv pos a1 ty a2 a3 e
          | Fc_ir_type.LetSubscript (v, ty, a1, a2, e) ->
               rtype_subscript_expr tenv venv rtenv pos v ty a1 a2 e
          | Fc_ir_type.LetAddrOfSubscript (v, ty, a1, a2, e) ->
               rtype_addr_of_subscript_expr tenv venv rtenv pos v ty a1 a2 e
          | Fc_ir_type.SetProject (a1, l, ty, a2, e) ->
               rtype_set_project_expr tenv venv rtenv pos a1 l ty a2 e
          | Fc_ir_type.LetProject (v, ty, a, l, e) ->
               rtype_project_expr tenv venv rtenv pos v ty a l e
          | Fc_ir_type.LetAddrOfProject (v, ty, a, l, e) ->
               rtype_addr_of_project_expr tenv venv rtenv pos v ty a l e
          | Fc_ir_type.Memcpy (a1, ty, a2, e) ->
               rtype_memcpy_expr tenv venv rtenv pos a1 ty a2 e
          | Fc_ir_type.LetAlloc (v, ty, e) ->
               rtype_alloc_expr tenv venv rtenv pos v ty e
          | Fc_ir_type.SetVar _
          | Fc_ir_type.LetTypes _
          | Fc_ir_type.LetFuns _
          | Fc_ir_type.LetFuns2 _
          | Fc_ir_type.LetString _
          | Fc_ir_type.LetAddrOfVar _
          | Fc_ir_type.LetCopy _
          | Fc_ir_type.LetApply _
          | Fc_ir_type.Return _ ->
               raise (IRException (pos, IRLevel 2))

   (*
    * Add v if it is an inner pointer.
    *)
   and rtype_atom_expr tenv venv rtenv pos v ty a e =
      let pos = string_pos "rtype_atom_expr" pos in
      let venv = venv_add venv v ty in
      let rtenv, a_ty = rtype_atom rtenv pos a in
      let rtenv = rtenv_add_var rtenv v a_ty in
         rtype_expr tenv venv rtenv e

   (*
    * Unary operation does not apply to pointers.
    *)
   and rtype_unop_expr tenv venv rtenv pos v ty op a e =
      let pos = string_pos "rtype_unop_expr" pos in
      let venv = venv_add venv v ty in
      let rtenv = rtenv_add_var rtenv v TyInferScalar in
         rtype_expr tenv venv rtenv e

   (*
    * Pointer arithmetic causes pointers to escape.
    *)
   and rtype_binop_expr tenv venv rtenv pos v ty op a1 a2 e =
      let pos = string_pos "rtype_binop_expr" pos in
      let venv = venv_add venv v ty in
      let ty1 = tenv_expand tenv pos (type_of_atom venv pos a1) in
      let ty2 = tenv_expand tenv pos (type_of_atom venv pos a2) in
         match op, ty1, ty2 with
            PlusOp, TyArray _, TyInt
          | MinusOp, TyArray _, TyInt ->
               let ty_var = new_symbol v in
               let rtenv = rtenv_add_ty_var rtenv ty_var (SubstBool true) in
               let rtenv = rtenv_add_var rtenv v (TyInferPointer ty_var) in
                  rtype_expr tenv venv rtenv e
          | _ ->
               let rtenv = rtenv_add_var rtenv v TyInferScalar in
                  rtype_expr tenv venv rtenv e

   (*
    * Tailcall.
    * Coerce all the function parameters.
    *)
   and rtype_tailcall_expr tenv venv rtenv pos f args =
      let pos = string_pos "rtype_tailcall_expr" pos in
         match rtenv_lookup_fun_opt rtenv f, args with
            Some vars, _ ->
               (* For a toplevel function, lift the types of the arguments *)
               List.fold_left2 (fun rtenv v a ->
                     let ty_var = rtenv_lookup_var rtenv pos v in
                     let rtenv, ty_atom = rtype_atom rtenv pos a in
                        rtype_unify rtenv pos ty_var ty_atom) rtenv vars args
          | None, [] ->
               rtenv
          | None, _ ->
               (* Lift the type of the remaining arguments *)
               List.fold_left (fun rtenv a ->
                     rtype_make_unsafe_atom rtenv pos a) rtenv (List.tl args)

   (*
    * Closure creation.
    * Several things have to happen:
    *    1. The first argument to the closure should be a safe pointer
    *    2. The remaining pointers should be unsafe
    *    3. There should be only one args'
    *    4. The value has type TyInferClosure
    *)
   and rtype_closure_expr tenv venv rtenv pos v ty f args e =
      let pos = string_pos "rtype_closure_expr" pos in
      let venv = venv_add venv v ty in
      let vars = rtenv_lookup_fun rtenv pos f in
      let rtenv =
         match vars, args with
            v :: vars, [frame] ->
               (* All the remaining vars become UsafePointers *)
               List.fold_left (fun rtenv v ->
                     match rtenv_lookup_var rtenv pos v with
                        TyInferPointer v ->
                           rtype_make_unsafe rtenv pos v
                      | TyInferScalar
                      | TyInferClosure ->
                           rtenv) rtenv vars
          | _, [] ->
               raise (IRException (pos, InternalError "empty closure"))
          | _ ->
               raise (IRException (pos, InternalError "closure has multiple arguments"))
      in
      let rtenv = rtenv_add_var rtenv v TyInferClosure in
         rtype_expr tenv venv rtenv e

   (*
    * External functions take arbitrary pointers.
    * We assume that if it returns a pointer, the pointer is
    * safe.
    *)
   and rtype_extcall_expr tenv venv rtenv pos v f ty args e =
      let pos = string_pos "rtype_extcall_expr" pos in
      let _, ty_res = dest_fun_type1 tenv pos ty in
      let rtenv, ty_res' = safe_rtype_of_type tenv rtenv pos ty_res in
      let venv = venv_add venv v ty_res in
      let rtenv = rtenv_add_var rtenv v ty_res' in
         rtype_expr tenv venv rtenv e

   (*
    * For internal call, we will be faking bounds,
    * so we don't actually change the function definition.
    *)
   and rtype_intcall_expr tenv venv rtenv pos ty f args =
      rtenv

   (*
    * Conditional expression.
    *)
   and rtype_if_expr tenv venv rtenv pos op a1 a2 e1 e2 =
      let pos = string_pos "rtype_if_expr" pos in
      let rtenv = rtype_expr tenv venv rtenv e1 in
         rtype_expr tenv venv rtenv e2

   (*
    * Set subscript doesn't change pointer flags.
    *)
   and rtype_set_subscript_expr tenv venv rtenv pos a1 ty a2 a3 e =
      let pos = string_pos "rtype_set_subscript_expr" pos in
      let rtenv = rtype_make_unsafe_atom rtenv pos a1 in
         rtype_expr tenv venv rtenv e

   (*
    * If the subscript operation returns a pointer,
    * it is always an escaping pointer.
    *)
   and rtype_subscript_expr tenv venv rtenv pos v ty a1 a2 e =
      let pos = string_pos "rtype_subscript_expr" pos in
      let venv = venv_add venv v ty in
      let rtenv = rtype_make_unsafe_atom rtenv pos a1 in
      let rtenv, ty = unsafe_rtype_of_type tenv rtenv pos ty in
      let rtenv = rtenv_add_var rtenv v ty in
         rtype_expr tenv venv rtenv e

   (*
    * Address-of always returns an escaping pointer.
    *)
   and rtype_addr_of_subscript_expr tenv venv rtenv pos v ty a1 a2 e =
      let pos = string_pos "rtype_addr_of_expr" pos in
      let venv = venv_add venv v ty in
      let rtenv, ty = unsafe_rtype_of_type tenv rtenv pos ty in
      let rtenv = rtenv_add_var rtenv v ty in
         rtype_expr tenv venv rtenv e

   (*
    * Set project doesn't change pointer flags.
    *)
   and rtype_set_project_expr tenv venv rtenv pos a1 l ty a2 e =
      let pos = string_pos "rtype_set_project_expr" pos in
         rtype_expr tenv venv rtenv e

   (*
    * If LetProject returns a pointer, the pointer
    * escapes.
    *)
   and rtype_project_expr tenv venv rtenv pos v ty a l e =
      let pos = string_pos "rtype_project_expr" pos in
      let venv = venv_add venv v ty in
      let rtenv, ty = unsafe_rtype_of_type tenv rtenv pos ty in
      let rtenv = rtenv_add_var rtenv v ty in
         rtype_expr tenv venv rtenv e

   (*
    * Address-of project always escapes.
    *)
   and rtype_addr_of_project_expr tenv venv rtenv pos v ty a l e =
      let pos = string_pos "rtype_addr_of_project_expr" pos in
      let venv = venv_add venv v ty in
      let rtenv, ty = unsafe_rtype_of_type tenv rtenv pos ty in
      let rtenv = rtenv_add_var rtenv v ty in
         rtype_expr tenv venv rtenv e

   (*
    * Allocation always returns a safe pointer.
    *)
   and rtype_alloc_expr tenv venv rtenv pos v ty e =
      let pos = string_pos "rtype_alloc_expr" pos in
      let venv = venv_add venv v (TyArray (ty, None)) in
      let ty_var = new_symbol v in
      let rtenv = rtenv_add_ty_var rtenv ty_var (SubstBool false) in
      let rtenv = rtenv_add_var rtenv v (TyInferPointer ty_var) in
         rtype_expr tenv venv rtenv e

   (*
    * Memcpy doesn't change anything.
    *)
   and rtype_memcpy_expr tenv venv rtenv pos a1 ty a2 e =
      let pos = string_pos "rtype_memcpy_expr" pos in
         rtype_expr tenv venv rtenv e

   (************************************************************************
    * EXP CONVERSION
    ************************************************************************)

   (*
    * Expand to unsafe/safe pointers.
    *)
   type run_type =
      TyRunUnsafePointer
    | TyRunSafePointer
    | TyRunClosure
    | TyRunScalar

   type run_var =
      RunUnsafePointer of atom * atom (* base, bounds *)
    | RunSafePointer
    | RunClosure of var               (* fun *)
    | RunScalar

   type rvenv =
      { rvenv_types : run_type SymbolTable.t;
        rvenv_vars : run_var SymbolTable.t;
        rvenv_funs : (var * var list) SymbolTable.t
      }

   (*
    * Add a type for a variable.
    *)
   let rvenv_add_type rvenv v ty =
      { rvenv with rvenv_types = SymbolTable.add rvenv.rvenv_types v ty }

   (*
    * Lookup a type for a var.
    *)
   let rvenv_lookup_type rvenv pos v =
      try SymbolTable.find rvenv.rvenv_types v with
         Not_found ->
            raise (IRException (pos, UnboundVar v))

   (*
    * Pointer environment keeps track of which variables
    * can be escaping pointers.
    *)
   let rvenv_add_unsafe_pointer rvenv v a_base a_bounds =
      { rvenv with rvenv_vars = SymbolTable.add rvenv.rvenv_vars v (RunUnsafePointer (a_base, a_bounds)) }

   let rvenv_add_closure rvenv v v_fun =
      { rvenv with rvenv_vars = SymbolTable.add rvenv.rvenv_vars v (RunClosure v_fun) }

   let rvenv_add_safe_pointer rvenv v =
      { rvenv with rvenv_vars = SymbolTable.add rvenv.rvenv_vars v RunSafePointer }

   let rvenv_add_scalar rvenv v =
      { rvenv with rvenv_vars = SymbolTable.add rvenv.rvenv_vars v RunScalar }

   let rvenv_lookup_var rvenv pos v =
      try SymbolTable.find rvenv.rvenv_vars v with
         Not_found ->
            raise (IRException (pos, UnboundVar v))

   let rvenv_lookup_closure rvenv v =
      try
         match SymbolTable.find rvenv.rvenv_vars v with
            RunClosure v_fun -> Some v_fun
          | _ -> None
      with
         Not_found ->
            None

   (*
    * Look up a function.
    *)
   let rvenv_lookup_fun rvenv pos f =
      try SymbolTable.find rvenv.rvenv_funs f with
         Not_found ->
            raise (IRException (pos, UnboundVar f))

   (*
    * Expand a type variable.
    *)
   let rec expand_ty_var subst v =
      try
         match SymbolTable.find subst v with
            SubstVar v -> expand_ty_var subst v
          | SubstBool flag -> flag
      with
         Not_found ->
            raise (Invalid_argument "expand_ty_var")

   (*
    * Produce a squashed type.
    *)
   let squash_rtenv rtenv =
      let { rtenv_types = subst; rtenv_vars = vars } = rtenv in
         SymbolTable.fold (fun rtenv v tyl ->
               let ty =
                  match List.hd tyl with
                     TyInferPointer ty_var ->
                        if expand_ty_var subst ty_var then
                           TyRunUnsafePointer
                        else
                           TyRunSafePointer
                   | TyInferClosure ->
                        TyRunClosure
                   | TyInferScalar ->
                        TyRunScalar
               in
                  SymbolTable.add rtenv v ty) (SymbolTable.create ()) vars

   (*
    * Variable environment funs include a possibly-renamed fun.
    *)
   let make_rvenv_funs funs =
      SymbolTable.fold (fun funs v vars ->
            let v' = new_symbol v in
            let vars = List.hd vars in
               SymbolTable.add funs v (v', vars)) (SymbolTable.create ()) funs

   (*
    * Initial variable environment.
    *)
   let rvenv_of_rtenv rtenv =
      { rvenv_types = squash_rtenv rtenv;
        rvenv_vars = SymbolTable.create ();
        rvenv_funs = make_rvenv_funs rtenv.rtenv_funs
      }

   (************************************************
    * TYPE ENVIRONMENT
    *)
   type tenv = (int * ty) SymbolTable.t

   let tenv_empty = SymbolTable.create ()

   let tenv_add tenv v =
      SymbolTable.add tenv v

   let tenv_lookup tenv pos v =
      try SymbolTable.find tenv v with
         Not_found ->
            raise (IRException (pos, UnboundType v))

   (*
    * Alignment is word-aligned.
    *)
   let align i =
      (i + 3) land (lnot 3)

   (*
    * Calculate the size of a value in a type.
    *)
   let rec sizeof tenv pos = function
      TyAny -> 0
    | TyChar -> 1
    | TyInt -> 4
    | TyFloat -> 8
    | TyArray (ty, Some i) ->
         sizeof tenv pos ty * i
    | TyArray (_, None) ->
         8
    | TyFun _ ->
         8
    | TyStruct fields ->
         List.fold_left (fun size (_, ty) ->
               size + align (sizeof tenv pos ty)) 0 fields
    | TyId v ->
         let size, _ = tenv_lookup tenv pos v in
            size

   (*
    * Get a type and its size.
    *)
   let rec tenv_expand tenv pos = function
      TyId v ->
         let size, ty = tenv_lookup tenv pos v in
            tenv_expand tenv pos ty
    | ty ->
         sizeof tenv pos ty, ty

   (*
    * Get the size of the type for each type identifier.
    * Have to compute a fixed point.
    *)
   let make_tenv tenv pos =
      (* Initially, assume all are 0 sized *)
      let tenv =
         tenv_fold (fun tenv v ty ->
               let ty = List.hd ty in
                  SymbolTable.add tenv v (0, ty)) (SymbolTable.create ()) tenv
      in

      (* Now iterate until the sizes converge *)
      let step tenv =
         SymbolTable.fold (fun (flag, tenv') v  ty ->
               let size, ty = List.hd ty in
               let size' = sizeof tenv pos ty in
               let tenv' = SymbolTable.add tenv' v (size', ty) in
                  flag || size' <> size, tenv') (false, SymbolTable.create ()) tenv
      in
      let rec collect tenv =
         let flag, tenv = step tenv in
            if flag then
               collect tenv
            else
               tenv
      in
         collect tenv

   (*
    * Get the value pointed-to.
    *)
   let type_of_deref2 tenv pos ty =
      match tenv_expand tenv pos ty with
         _, TyArray (ty, _) -> ty
       | _ -> raise (IRException (pos, NotAPointer ty))

   (*
    * Get a struct type.
    *)
   let dest_struct_type2 tenv pos ty =
      match tenv_expand tenv pos ty with
         _, TyStruct fields -> fields
       | _ -> raise (IRException (pos, NotAStruct ty))

   (*
    * Break apart the function type.
    *)
   let dest_fun_type2 tenv pos ty =
      match tenv_expand tenv pos ty with
         _, TyFun (ty_vars, ty_res) -> ty_vars, ty_res
       | _, ty -> raise (IRException (pos, NotAFunction ty))

   (*
    * Get the sizeof the first argument to the fun.
    *)
   let sizeof_frame tenv pos ty =
      match tenv_expand tenv pos ty with
         _, TyFun (ty_frame :: _, _) ->
            (match tenv_expand tenv pos ty_frame with
                _, Fc_ir_type.TyArray (ty, _) ->
                   sizeof tenv pos ty
              | _ ->
                   0)
       | _ -> raise (IRException (pos, NotAFunction ty))

   (*
    * Type testing.
    *)
   let is_char_type2 tenv pos ty =
      match tenv_expand tenv pos ty with
         _, TyChar  -> true
       | _ -> false

   let is_fun_type2 tenv pos ty =
      match tenv_expand tenv pos ty with
         _, TyArray (ty, None) ->
            (match tenv_expand tenv pos ty with
                _, TyFun _ -> true
              | _ -> false)
       | _, TyFun _ -> true
       | _ -> false

   let is_pointer_type2 tenv pos ty =
      match tenv_expand tenv pos ty with
         _, TyArray _ -> true
       | _ -> false

   (*
    * Calculate the offset into a record.
    *)
   let offsetof tenv pos ty l =
      let fields = dest_struct_type2 tenv pos ty in
      let rec search off = function
         (v, ty) :: fields ->
            if v = l then
               off
            else
               search (off + align (sizeof tenv pos ty)) fields
       | [] ->
            raise (IRException (pos, UnboundLabel l))
      in
         search 0 fields

   (************************************************
    * CONVERSION
    *)

   (*
    * Dummy FIR.
    *)
   let fir_nop = Return (AtomVar (Symbol.add ".nop"))

   (*
    * Atom should be an unsafe pointer.
    *)
   let make_unsafe_pointer_atom rvenv pos a cont =
      match a with
         AtomChar _
       | AtomInt _
       | AtomFloat _ ->
            raise (IRException (pos, InternalError "number should be a pointer"))
       | AtomNil _ ->
            let a_base = AtomInt 0 in
            let a_bounds = AtomInt 0 in
            let a_off = AtomInt 0 in
               cont rvenv a_base a_off a_bounds
       | AtomVar v ->
            match rvenv_lookup_var rvenv pos v with
               RunUnsafePointer (a_base, a_bounds) ->
                  cont rvenv a_base (AtomVar v) a_bounds
             | RunSafePointer ->
                  let v' = new_symbol v in
                  let v_bounds = new_symbol_string (Symbol.to_string v ^ "_bounds") in
                  let rvenv = rvenv_add_unsafe_pointer rvenv v' (AtomVar v) (AtomVar v_bounds) in
                  let blocks, exp = cont rvenv (AtomVar v) (AtomVar v') (AtomVar v_bounds) in
                     blocks, LetAtom (v', AtomInt 0,
                                      LetMem (v_bounds, AddrSafeOff (AtomVar v, AtomInt (-4)),
                                              exp))
             | RunClosure _ ->
                  raise (IRException (pos, InternalVarError (v, "closure should be a pointer")))
             | RunScalar ->
                  raise (IRException (pos, InternalVarError (v, "scalar should be a pointer")))

   (*
    * Atom should be a closure.
    *)
   let make_closure_atom rvenv pos a cont =
      match a with
         AtomChar _
       | AtomInt _
       | AtomFloat _
       | AtomNil _ ->
            raise (IRException (pos, InternalError "atom should be a closure"))
       | AtomVar v ->
            match rvenv_lookup_var rvenv pos v with
               RunClosure v_fun ->
                  cont rvenv v v_fun
             | RunSafePointer
             | RunUnsafePointer _
             | RunScalar ->
                  raise (IRException (pos, InternalError "atom should be a closure"))

   (*
    * Convert an expression.
    *)
   let rec make_expr tenv venv rvenv blocks e =
      let pos = string_pos "make_expr" (atom_pos e) in
         match e with
            Fc_ir_type.LetAtom (v, ty, a, e) ->
               make_atom_expr tenv venv rvenv blocks pos v ty a e
          | Fc_ir_type.LetUnop (v, ty, op, a, e) ->
               make_unop_expr tenv venv rvenv blocks pos v ty op a e
          | Fc_ir_type.LetBinop (v, ty, op, a1, a2, e) ->
               make_binop_expr tenv venv rvenv blocks pos v ty op a1 a2 e
          | Fc_ir_type.TailCall (f, args) ->
               make_tailcall_expr tenv venv rvenv blocks pos f args
          | Fc_ir_type.LetClosure (v, ty, f, args, e) ->
               make_closure_expr tenv venv rvenv blocks pos v ty f args e
          | Fc_ir_type.LetExtCall (v, f, ty, args, e) ->
               make_extcall_expr tenv venv rvenv blocks pos v f ty args e
          | Fc_ir_type.IntCall (ty, f, args) ->
               make_intcall_expr tenv venv rvenv blocks pos ty f args
          | Fc_ir_type.IfThenElse (op, a1, a2, e1, e2) ->
               make_if_expr tenv venv rvenv blocks pos op a1 a2 e1 e2
          | Fc_ir_type.SetSubscript (a1, ty, a2, a3, e) ->
               make_set_subscript_expr tenv venv rvenv blocks pos a1 ty a2 a3 e
          | Fc_ir_type.LetSubscript (v, ty, a1, a2, e) ->
               make_subscript_expr tenv venv rvenv blocks pos v ty a1 a2 e
          | Fc_ir_type.LetAddrOfSubscript (v, ty, a1, a2, e) ->
               make_addr_of_subscript_expr tenv venv rvenv blocks pos v ty a1 a2 e
          | Fc_ir_type.SetProject (a1, l, ty, a2, e) ->
               make_set_project_expr tenv venv rvenv blocks pos a1 l ty a2 e
          | Fc_ir_type.LetProject (v, ty, a, l, e) ->
               make_project_expr tenv venv rvenv blocks pos v ty a l e
          | Fc_ir_type.LetAddrOfProject (v, ty, a, l, e) ->
               make_addr_of_project_expr tenv venv rvenv blocks pos v ty a l e
          | Fc_ir_type.Memcpy (a1, ty, a2, e) ->
               make_memcpy_expr tenv venv rvenv blocks pos a1 ty a2 e
          | Fc_ir_type.LetAlloc (v, ty, e) ->
               make_alloc_expr tenv venv rvenv blocks pos v ty e
          | Fc_ir_type.SetVar _
          | Fc_ir_type.LetTypes _
          | Fc_ir_type.LetFuns _
          | Fc_ir_type.LetFuns2 _
          | Fc_ir_type.LetString _
          | Fc_ir_type.LetAddrOfVar _
          | Fc_ir_type.LetCopy _
          | Fc_ir_type.LetApply _
          | Fc_ir_type.Return _ ->
               raise (IRException (pos, IRLevel 2))

   (*
    * If a is a pointer type, the variable
    * is split into three parts: the base, offset, and bounds.
    *)
   and make_atom_expr tenv venv rvenv blocks pos v ty a e =
      let pos = string_pos "make_atom_expr" pos in
      let venv = venv_add venv v ty in
      let blocks, exp =
         match rvenv_lookup_type rvenv pos v with
            TyRunUnsafePointer ->
               make_unsafe_pointer_atom rvenv pos a (fun rvenv a_base a_off a_bounds ->
                     let rvenv = rvenv_add_unsafe_pointer rvenv v a_base a_bounds in
                     let blocks, exp = make_expr tenv venv rvenv blocks e in
                        blocks, LetAtom (v, a_off, exp))
          | TyRunSafePointer ->
               let rvenv = rvenv_add_safe_pointer rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetAtom (v, a, exp)
          | TyRunClosure ->
               make_closure_atom rvenv pos a (fun rvenv v_frame v_fun ->
                     let rvenv = rvenv_add_closure rvenv v v_fun in
                     let blocks, exp = make_expr tenv venv rvenv blocks e in
                        blocks, LetAtom (v, AtomVar v_frame, exp))
          | TyRunScalar ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetAtom (v, a, exp)
      in
      let fir = Fc_ir_type.LetAtom (v, ty, a, fir_nop) in
         blocks, CommentFIR (fir, exp)

   (*
    * Unary operation.
    *)
   and make_unop_expr tenv venv rvenv blocks pos v ty op a e =
      let pos = string_pos "make_unop_expr" pos in
      let venv = venv_add venv v ty in
      let _, ty' = tenv_expand tenv pos (type_of_atom venv pos a) in
      let blocks, exp =
         match op, ty' with
            Fc_ir_type.UMinusOp, TyAny
          | Fc_ir_type.UMinusOp, TyInt ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetUnop (v, UMinusIntOp, OpAtom a, exp)
          | Fc_ir_type.UMinusOp, TyFloat ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetUnop (v, UMinusFloatOp, OpAtom a, exp)
          | Fc_ir_type.UToChar, TyAny
          | Fc_ir_type.UToChar, TyInt ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetUnop (v, UCharOfIntOp, OpAtom a, exp)
          | Fc_ir_type.UToChar, TyFloat ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
               let v' = new_symbol v in
                  blocks, LetUnop (v', UIntOfFloatOp, OpAtom a,
                                   LetUnop (v, UCharOfIntOp, OpAtom (AtomVar v'), exp))
          | Fc_ir_type.UToInt, TyChar
          | Fc_ir_type.UToInt, TyInt
          | Fc_ir_type.UToInt, TyAny ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetAtom (v, a, exp)
          | Fc_ir_type.UToInt, TyFloat ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetUnop (v, UIntOfFloatOp, OpAtom a, exp)
          | Fc_ir_type.UToFloat, TyChar
          | Fc_ir_type.UToFloat, TyInt
          | Fc_ir_type.UToFloat, TyAny ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetUnop (v, UFloatOfIntOp, OpAtom a, exp)
          | Fc_ir_type.UToFloat, TyFloat ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetAtom (v, a, exp)
          | Fc_ir_type.UToFunPointer, _
          | Fc_ir_type.UToFun, _ ->
               make_closure_atom rvenv pos a (fun rvenv v_frame v_fun ->
                     let rvenv = rvenv_add_closure rvenv v v_fun in
                     let blocks, exp = make_expr tenv venv rvenv blocks e in
                        blocks, LetAtom (v, AtomVar v_frame, exp))
          | Fc_ir_type.UNotOp, _ ->
               raise (IRException (pos, InternalError "UNotOp not supported"))
          | _ ->
               raise (IRException (pos, TypeError (ty, ty')))
      in
      let fir = Fc_ir_type.LetUnop (v, ty, op, a, fir_nop) in
         blocks, CommentFIR (fir, exp)

   (*
    * Binary operation.
    *)
   and make_binop_expr tenv venv rvenv blocks pos v ty op a1 a2 e =
      let pos = string_pos "make_binop_expr" pos in
      let _, ty1 = tenv_expand tenv (int_pos 3 pos) (type_of_atom venv (int_pos 1 pos) a1) in
      let _, ty2 = tenv_expand tenv (int_pos 4 pos) (type_of_atom venv (int_pos 2 pos) a2) in
      let venv = venv_add venv v ty in
      let blocks, exp =
         match op, ty1, ty2 with
            Fc_ir_type.PlusOp, TyInt, TyInt ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, PlusIntOp, OpAtom a1, OpAtom a2, exp)
          | Fc_ir_type.PlusOp, TyFloat, TyFloat ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, PlusFloatOp, OpAtom a1, OpAtom a2, exp)
          | Fc_ir_type.PlusOp, TyArray (ty, _), TyInt ->
               make_unsafe_pointer_atom rvenv (int_pos 5 pos) a1 (fun rvenv a_base a_off a_bounds ->
                     let rvenv = rvenv_add_unsafe_pointer rvenv v a_base a_bounds in
                     let size = sizeof tenv (int_pos 6 pos) ty in
                     let blocks, exp = make_expr tenv venv rvenv blocks e in
                        blocks, LetAddr (v, AddrSafeMul (a_off, a2, size), exp))
          | Fc_ir_type.MinusOp, TyInt, TyInt ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, MinusIntOp, OpAtom a1, OpAtom a2, exp)
          | Fc_ir_type.MinusOp, TyFloat, TyFloat ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, MinusFloatOp, OpAtom a1, OpAtom a2, exp)
          | Fc_ir_type.MinusOp, TyArray (ty, _), TyInt ->
               make_unsafe_pointer_atom rvenv pos a1 (fun rvenv a_base a_off a_bounds ->
                     let rvenv = rvenv_add_unsafe_pointer rvenv v a_base a_bounds in
                     let size = sizeof tenv pos ty in
                     let blocks, exp = make_expr tenv venv rvenv blocks e in
                        blocks, LetAddr (v, AddrSafeMul (a_off, a2, -size), exp))
          | Fc_ir_type.MinusOp, TyArray (ty, _), TyArray _ ->
               make_unsafe_pointer_atom rvenv pos a1 (fun rvenv a1_base a1_off a1_bounds ->
                     make_unsafe_pointer_atom rvenv pos a2 (fun rvenv a2_base a2_off a2_bounds ->
                           let rvenv = rvenv_add_scalar rvenv v in
                           let size = sizeof tenv pos ty in
                           let v' = new_symbol v in
                           let blocks, exp = make_expr tenv venv rvenv blocks e in
                              blocks, LetBinop (v', MinusIntOp, OpAtom a1_off, OpAtom a2_off,
                                                LetBinop (v, DivIntOp, OpAtom (AtomVar v'), OpAtom (AtomInt size), exp))))
          | Fc_ir_type.MulOp, TyInt, TyInt ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, MulIntOp, OpAtom a1, OpAtom a2, exp)
          | Fc_ir_type.MulOp, TyFloat, TyFloat ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, MulFloatOp, OpAtom a1, OpAtom a2, exp)
          | Fc_ir_type.DivOp, TyInt, TyInt ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, DivIntOp, OpAtom a1, OpAtom a2, exp)
          | Fc_ir_type.DivOp, TyFloat, TyFloat ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, DivFloatOp, OpAtom a1, OpAtom a2, exp)
          | Fc_ir_type.RemOp, TyInt, TyInt ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, RemIntOp, OpAtom a1, OpAtom a2, exp)
          | Fc_ir_type.BAndOp, TyInt, TyInt ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, BAndIntOp, OpAtom a1, OpAtom a2, exp)
          | Fc_ir_type.BOrOp, TyInt, TyInt ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, BOrIntOp, OpAtom a1, OpAtom a2, exp)
          | Fc_ir_type.XorOp, TyInt, TyInt ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, XorIntOp, OpAtom a1, OpAtom a2, exp)
          | Fc_ir_type.LslOp, TyInt, TyInt ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, LslIntOp, OpAtom a1, OpAtom a2, exp)
          | Fc_ir_type.LsrOp, TyInt, TyInt ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, LsrIntOp, OpAtom a1, OpAtom a2, exp)
          | Fc_ir_type.AsrOp, TyInt, TyInt ->
               let rvenv = rvenv_add_scalar rvenv v in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  blocks, LetBinop (v, AsrIntOp, OpAtom a1, OpAtom a2, exp)
          | _ ->
               raise (IRException (pos, TypeError (ty1, ty2)))
      in
      let fir = Fc_ir_type.LetBinop (v, ty, op, a1, a2, fir_nop) in
         blocks, CommentFIR (fir, exp)

   (*
    * Tailcall for a known function.
    * Add the extra pointer arguments.
    *)
   and make_tailcall_known tenv venv rvenv blocks pos f args' =
      let pos = string_pos "make_tailcall_known" pos in
      let _, fvars' = rvenv_lookup_fun rvenv pos f in
      let rec collect f_vars args =
         match f_vars, args with
            v :: fvars, a :: args ->
               (match rvenv_lookup_var rvenv (int_pos 1 pos) v with
                   RunUnsafePointer (AtomVar v_base, AtomVar v_bounds) ->
                      make_unsafe_pointer_atom rvenv (int_pos 2 pos) a (fun rvenv a_base a_off a_bounds ->
                            let blocks, exp = collect fvars args in
                               blocks, LetAtom (v_base, a_base,
                                                LetAtom (v, a_off,
                                                         LetAtom (v_bounds, a_bounds,
                                                                  exp))))
                 | RunUnsafePointer _ ->
                      raise (IRException (pos, InternalError "function arguments should be vars"))
                 | RunClosure v_fun ->
                      make_closure_atom rvenv (int_pos 3 pos) a (fun rvenv v_frame' v_fun' ->
                            let blocks, exp = collect fvars args in
                               blocks, LetAtom (v, AtomVar v_frame',
                                                LetAtom (v_fun, AtomVar v_fun',
                                                         exp)))
                 | RunScalar
                 | RunSafePointer ->
                      let blocks, exp = collect fvars args in
                      let _ =
                         match a with
                            AtomVar v ->
                               (match rvenv_lookup_var rvenv pos v with
                                  RunUnsafePointer _ ->
                                      raise (IRException (pos, InternalError "argument of wrong type"))
                                 | _ ->
                                      ())
                          | _ ->
                               ()
                      in
                         blocks, LetAtom (v, a, exp))
          | [], [] ->
               blocks, Jump f
          | _ ->
               raise (IRException (pos, ArityMismatch (List.length fvars', List.length args')))
      in
         collect fvars' args'

   (*
    * Tailcall for a closure function.
    * Pass the arguments in the standard places.
    * All pointers are unsafe.
    *)
   and make_tailcall_unknown tenv venv rvenv blocks pos v_frame v_fun args =
      let pos = string_pos "make_tailcall_unknown" pos in
      let rec collect vars vars' args =
         match vars, args with
            v1 :: v2 :: v3 :: vars, a :: args ->
               (match a with
                   AtomChar _
                 | AtomInt _
                 | AtomFloat _ ->
                      let exp = collect (v2 :: v3 :: vars) (v1 :: vars') args in
                         LetAtom (v1, a, exp)
                 | AtomNil _ ->
                      let exp = collect vars (v1 :: v2 :: v3 :: vars') args in
                         LetAtom (v1, AtomInt 0,
                                  LetAtom (v2, AtomInt 0,
                                           LetAtom (v3, AtomInt 0,
                                                    exp)))
                 | AtomVar v ->
                      match rvenv_lookup_var rvenv pos v with
                         RunUnsafePointer (a_base, a_bounds) ->
                            let exp = collect vars (v1 :: v2 :: v3 :: vars') args in
                               LetAtom (v1, a_base,
                                        LetAtom (v2, AtomVar v,
                                                 LetAtom (v3, a_bounds,
                                                          exp)))
                       | RunSafePointer ->
                            let exp = collect vars (v1 :: v2 :: v3 :: vars') args in
                               LetAtom (v1, AtomVar v,
                                        LetAtom (v2, AtomInt 0,
                                                 LetMem (v3, AddrSafeOff (AtomVar v1, AtomInt (-4)),
                                                         exp)))
                       | RunClosure v_fun ->
                            let exp = collect (v3 :: vars) (v1 :: v2 :: vars') args in
                               LetAtom (v1, AtomVar v,
                                        LetAtom (v2, AtomVar v_fun,
                                                 exp))
                       | RunScalar ->
                            let exp = collect (v2 :: v3 :: vars) (v1 :: vars') args in
                               LetAtom (v1, AtomVar v, exp))
          | _, [] ->
               RJump (vars', OpAtom (AtomVar v_fun))
          | _ ->
               raise (IRException (pos, InternalError "not enough args"))
      in

      (* Get the standard args, assume at most three times as many args *)
      let exp =
         match Frame.get_stdargs (List.length args * 3 + 1) with
            v1 :: vars ->
               LetAtom (v1, AtomVar v_frame, collect vars [v1] args)
          | [] ->
               raise (Invalid_argument "make_tailcall_unknown")
      in
         blocks, exp

   (*
    * General tailcall.
    *)
   and make_tailcall_expr tenv venv rvenv blocks pos f args =
      let pos = string_pos "make_tailcall_expr" pos in
      let blocks, exp =
         match rvenv_lookup_closure rvenv f with
            Some v_fun ->
               make_tailcall_unknown tenv venv rvenv blocks pos f v_fun args
          | None ->
               make_tailcall_known tenv venv rvenv blocks pos f args
      in
      let fir = Fc_ir_type.TailCall (f, args) in
         blocks, CommentFIR (fir, exp)

   (*
    * Closure have to get their arguments in the
    * standard places.
    *)
   and make_closure_expr tenv venv rvenv blocks pos v ty f args' e =
      let pos = string_pos "make_closure_expr" pos in
      let blocks, exp =
         match args' with
            [frame] ->
               let f', _ = rvenv_lookup_fun rvenv pos f in
               let v_fun = new_symbol_string (Symbol.to_string v ^ "_fun") in
               let rvenv = rvenv_add_closure rvenv v v_fun in
               let venv = venv_add venv v ty in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
               let a_frame =
                  match frame with
                     AtomVar v ->
                        (match rvenv_lookup_var rvenv pos v with
                            RunUnsafePointer (a_base, _) ->
                               a_base
                          | RunSafePointer ->
                               frame
                          | RunClosure _
                          | RunScalar ->
                               raise (IRException (pos, InternalError "Closure expects a frame")))
                   | _ ->
                        raise (IRException (pos, InternalError "Closure expects a frame"))
               in
                  blocks, LetAtom (v, a_frame,
                                   LetAtom (v_fun, AtomVar f',
                                            exp))
          | _ ->
               raise (IRException (pos, InternalError "closure has the wrong number of args"))
      in
      let fir = Fc_ir_type.LetClosure (v, ty, f, args', fir_nop) in
         blocks, CommentFIR (fir, exp)

   (*
    * ExtCall computes normal C pointers before calling the C function.
    *)
   and make_extcall_expr tenv venv rvenv blocks pos v f ty args e =
      let pos = string_pos "make_extcall_expr" pos in
      let rec collect args' = function
         a :: args ->
            (match a with
                AtomChar _
              | AtomInt _
              | AtomFloat _
              | AtomNil _ ->
                   collect (a :: args') args
              | AtomVar v ->
                   match rvenv_lookup_var rvenv pos v with
                      RunUnsafePointer (a_base, _) ->
                         let v' = new_symbol v in
                         let blocks, exp = collect (AtomVar v' :: args') args in
                            blocks, LetAddr (v', AddrSafeOff (a_base, AtomVar v), exp)
                    | RunClosure _ ->
                         raise (IRException (pos, InternalError "can't pass a closure to an external function"))
                    | RunSafePointer
                    | RunScalar ->
                         collect (a :: args') args)
       | [] ->
            let _, ty_res = dest_fun_type2 tenv pos ty in
            let venv = venv_add venv v ty_res in
            let rvenv = rvenv_add_scalar rvenv v in
            let blocks, exp = make_expr tenv venv rvenv blocks e in
               blocks, LetExtCall (v, f, List.rev args', exp)
      in
      let blocks, exp = collect [] args in
      let fir = Fc_ir_type.LetExtCall (v, f, ty, args, fir_nop) in
         blocks, CommentFIR (fir, exp)

   (*
    * IntCall calls the function with two extra arguments:
    * the data segment, and the "exit" continuation that
    * returns the value to external C.
    *)
   and make_intcall_expr tenv venv rvenv blocks pos ty f args =
      let pos = string_pos "make_intcall_expr" pos in
      let fir = Fc_ir_type.IntCall (ty, f, args) in
      let args = AtomVar Frame.data_seg :: AtomVar Frame.exit_label :: args in
      let blocks, exp = make_tailcall_expr tenv venv rvenv blocks pos f args in
         blocks, CommentFIR (fir, exp)

   (*
    * Conditional.
    * We have to create two new blocks for the expressions.
    *)
   and make_if_expr tenv venv rvenv blocks pos op a1 a2 e1 e2 =
      (* Add e1, e2 to the block list *)
      let pos = string_pos "make_if_expr" pos in
      let fir = Fc_ir_type.IfThenElse (op, a1, a2, fir_nop, fir_nop) in
      let label1 = new_symbol_string "true_case" in
      let label2 = new_symbol_string "false_case" in
      let blocks, e1 = make_expr tenv venv rvenv blocks e1 in
      let blocks, e2 = make_expr tenv venv rvenv blocks e2 in
      let block1 =
         { block_global = false;
           block_label = label1;
           block_code = e1
         }
      in
      let block2 =
         { block_global = false;
           block_label = label2;
           block_code = e2
         }
      in
      let blocks = block1 :: block2 :: blocks in

      (* Now compute the code *)
      let _, ty1 = tenv_expand tenv pos (type_of_atom venv pos a1) in
      let _, ty2 = tenv_expand tenv pos (type_of_atom venv pos a2) in
      let op, l1, l2, is_pointer_op, l3 =
         match op, ty1, ty2 with
            EqOp, TyInt, TyInt ->
               EQIntOp, label1, label2, false, label2
          | NEqOp, TyInt, TyInt ->
               EQIntOp, label2, label1, false, label1
          | LeOp, TyInt, TyInt ->
               LEIntOp, label1, label2, false, label2
          | LtOp, TyInt, TyInt ->
               LTIntOp, label1, label2, false, label2
          | GtOp, TyInt, TyInt ->
               LEIntOp, label2, label1, false, label2
          | GeOp, TyInt, TyInt ->
               LTIntOp, label2, label1, false, label2
          | EqOp, TyArray _, TyArray _ ->
               EQIntOp, label1, label2, true, label2
          | NEqOp, TyArray _, TyArray _ ->
               EQIntOp, label2, label1, true, label1
          | LeOp, TyArray _, TyArray _ ->
               LEIntOp, label1, label2, true, label2
          | LtOp, TyArray _, TyArray _ ->
               LTIntOp, label1, label2, true, label2
          | GtOp, TyArray _, TyArray _ ->
               LEIntOp, label2, label1, true, label2
          | GeOp, TyArray _, TyArray _ ->
               LTIntOp, label2, label1, true, label2
          | EqOp, TyFloat, TyFloat ->
               EQFloatOp, label1, label2, false, label2
          | NEqOp, TyFloat, TyFloat ->
               EQFloatOp, label2, label1, false, label1
          | LeOp, TyFloat, TyFloat ->
               LEFloatOp, label1, label2, false, label2
          | LtOp, TyFloat, TyFloat ->
               LTFloatOp, label1, label2, false, label2
          | GtOp, TyFloat, TyFloat ->
               LEFloatOp, label2, label1, false, label2
          | GeOp, TyFloat, TyFloat ->
               LTFloatOp, label2, label1, false, label2
          | _ ->
               raise (IRException (pos, TypeError (ty1, ty2)))
      in
      let blocks, exp =
         if is_pointer_op then
            make_unsafe_pointer_atom rvenv pos a1 (fun rvenv a1_base a1_off _ ->
                  make_unsafe_pointer_atom rvenv pos a2 (fun rvenv a2_base a2_off _ ->
                        let l4 = new_symbol_string "pointer_case" in
                        let block4 =
                           { block_global = false;
                             block_label = l4;
                             block_code = CJump (op, OpAtom a1_off, OpAtom a2_off, l1, l2)
                           }
                        in
                        let exp = CJump (EQIntOp, OpAtom a1_base, OpAtom a2_base, l4, l3) in
                           block4 :: blocks, exp))
         else
            blocks, CJump (op, OpAtom a1, OpAtom a2, l1, l2)
      in
         blocks, CommentFIR (fir, exp)

   (*
    * Closure, check that the indexes are valid.
    *)
   and load_closure tenv venv rvenv blocks pos v addr e =
      let pos = string_pos "load_closure" pos in
      let s = Symbol.to_string v in
      let v_fun = new_symbol_string (s ^ "_fun") in
      let rvenv = rvenv_add_closure rvenv v v_fun in
      let blocks, exp = make_expr tenv venv rvenv blocks e in

      (* Temp registers *)
      let r0 = new_symbol_string "closure_addr" in
      let r1 = new_symbol_string "closure_frame_index" in
      let r2 = new_symbol_string "closure_fun_index" in
      let r3 = new_symbol_string "closure_frame" in

      (* Abort block specifies what to do if closure is bad *)
      let abort_label = new_symbol_string "abort_label" in
      let abort_block =
         { block_global = false;
           block_label = abort_label;
           block_code = Call Frame.seg_fault_label
         }
      in

      (* Closure size matches, so load the correct values and continue *)
      let closure_is_valid_label = new_symbol_string "closure_is_valid_label" in
      let exp4 =
         LetBinop (v, PlusIntOp, OpAtom (AtomVar r3), OpAtom (AtomInt 8),
                  LetMem (v_fun, AddrSafeMulOff (AtomVar Frame.closure_base, AtomVar r2, 8, AtomInt 4),
                          exp))
      in
      let closure_is_valid_block =
         { block_global = false;
           block_label = closure_is_valid_label;
           block_code = exp4
         }
      in

      (* Closure index is in bounds *)
      let closure_in_bounds_label = new_symbol_string "closure_in_bounds_label" in
      let exp3 =
         CJump (EQIntOp, OpAddr (AddrSafeMul (AtomVar Frame.closure_base,
                                              AtomVar r2, 8)),
                OpAddr (AddrSafeOff (AtomVar r3, AtomInt 4)),
                closure_is_valid_label, abort_label)
      in
      let closure_in_bounds_block =
         { block_global = false;
           block_label = closure_in_bounds_label;
           block_code = exp3
         }
      in

      (* Frame pointer is valid; check the closure *)
      let frame_is_valid_label = new_symbol_string "frame_is_valid" in
      let exp2 =
         CJump (ULEIntOp, OpAtom (AtomVar r2), OpAtom (AtomVar Frame.closure_size),
                closure_in_bounds_label, abort_label)
      in
      let frame_is_valid_block =
         { block_global = false;
           block_label = frame_is_valid_label;
           block_code = exp2
         }
      in

      (* Frame pointer is in bounds *)
      let frame_in_bounds_label = new_symbol_string "frame_in_bounds" in
      let exp1 =
         LetMem (r3, AddrSafeMul (AtomVar Frame.pointer_base, AtomVar r1, 4),
                 CJump (EQIntOp, OpAtom (AtomVar r1), OpAddr (AddrSafe (AtomVar r3)),
                        frame_is_valid_label, abort_label))
      in
      let frame_in_bounds_block =
         { block_global = false;
           block_label = frame_in_bounds_label;
           block_code = exp1
         }
      in

      (* Load the two parts, and check them *)
      let exp =
         LetAddr (r0, addr,
         LetBinop (r1, XorIntOp, OpAddr (AddrSafe (AtomVar r0)), OpAtom (AtomInt Frame.hash_index),
         LetBinop (r2, XorIntOp, OpAddr (AddrSafeOff (AtomVar r0, AtomInt 4)), OpAtom (AtomInt Frame.hash_index),
         CJump (ULEIntOp, OpAtom (AtomVar r1), OpAtom (AtomVar Frame.pointer_size),
         frame_in_bounds_label, abort_label))))
      in

      (* Gather all the blocks *)
      let blocks =
         abort_block
         :: closure_is_valid_block
         :: closure_in_bounds_block
         :: frame_is_valid_block
         :: frame_in_bounds_block
         :: blocks
      in
         blocks, exp

   (*
    * Pointer, make sure the pointer is valid.
    *)
   and load_pointer tenv venv rvenv blocks pos v addr e =
      let pos = string_pos "load_pointer" pos in
      let s = Symbol.to_string v in
      let v_base = new_symbol_string (s ^ "_base") in
      let v_bounds = new_symbol_string (s ^ "_bounds") in
      let rvenv = rvenv_add_unsafe_pointer rvenv v (AtomVar v_base) (AtomVar v_bounds) in
      let blocks, exp = make_expr tenv venv rvenv blocks e in

      (* New registers *)
      let r0 = new_symbol_string "pointer_addr" in
      let r1 = new_symbol_string "pointer_block_index" in
      let r2 = new_symbol_string "pointer_offset" in
      let r3 = new_symbol_string "pointer_block" in

      (* Abort block specifies what to do if closure is bad *)
      let abort_label = new_symbol_string "abort_label" in
      let abort_block =
         { block_global = false;
           block_label = abort_label;
           block_code = Call Frame.seg_fault_label
         }
      in

      (* Frame pointer is valid; get the bounds *)
      let frame_is_valid_label = new_symbol_string "frame_is_valid" in
      let exp2 =
         LetBinop (v_base, PlusIntOp, OpAtom (AtomVar r3), OpAtom (AtomInt 8),
                  LetAtom (v, AtomVar r2,
                           LetMem (v_bounds, AddrSafeOff (AtomVar r3, AtomInt 4),
                                   exp)))
      in
      let frame_is_valid_block =
         { block_global = false;
           block_label = frame_is_valid_label;
           block_code = exp2
         }
      in

      (*
       * Base index is in bounds.
       * Get the pointer value.
       *)
      let frame_in_bounds_label = new_symbol_string "frame_in_bounds" in
      let exp1 =
         LetMem (r3, AddrSafeMul (AtomVar Frame.pointer_base, AtomVar r1, 4),
                 CJump (EQIntOp, OpAtom (AtomVar r1), OpAddr (AddrSafe (AtomVar r3)),
                        frame_is_valid_label, abort_label))
      in
      let frame_in_bounds_block =
         { block_global = false;
           block_label = frame_in_bounds_label;
           block_code = exp1
         }
      in

      (* Load the two parts, and check them *)
      let exp =
         LetAddr (r0, addr,
         LetBinop (r1, XorIntOp, OpAddr (AddrSafe (AtomVar r0)), OpAtom (AtomInt Frame.hash_index),
         LetMem (r2, AddrSafeOff (AtomVar r0, AtomInt 4),
         CJump (ULEIntOp, OpAtom (AtomVar r1), OpAtom (AtomVar Frame.pointer_size),
         frame_in_bounds_label, abort_label))))
      in

      (* Gather all the blocks *)
      let blocks =
         abort_block
         :: frame_is_valid_block
         :: frame_in_bounds_block
         :: blocks
      in
         blocks, exp

   (*
    * Store a closure to memory.
    *)
   and store_closure rvenv blocks pos addr a2 exp =
      let pos = string_pos "store_closure" pos in
      let r0 = new_symbol_string "store_closure_addr" in
         match a2 with
            AtomNil _ ->
               let exp =
                  LetAddr (r0, addr,
                  SetMem (AddrSafe (AtomVar r0), OpAtom (AtomInt Frame.hash_index),
                  SetMem (AddrSafeOff (AtomVar r0, AtomInt 4), OpAtom (AtomInt Frame.hash_index),
                  exp)))
               in
                  blocks, exp
          | AtomVar v ->
               let r1 = new_symbol_string "store_closure_frame_index" in
               let r2 = new_symbol_string "store_closure_fun_index" in
               let exp =
                  match rvenv_lookup_var rvenv pos v with
                     RunClosure v_fun ->
                        LetAddr (r0, addr,
                        LetBinop (r1, XorIntOp, OpAddr (AddrSafeOff (AtomVar v, AtomInt (-8))), OpAtom (AtomInt Frame.hash_index),
                        SetMem (AddrSafe (AtomVar r0), OpAtom (AtomVar r1),
                        LetBinop (r2, XorIntOp, OpAddr (AddrSafeOff (AtomVar v_fun, AtomInt (-4))), OpAtom (AtomInt Frame.hash_index),
                        SetMem (AddrSafeOff (AtomVar r0, AtomInt 4), OpAtom (AtomVar r2),
                        exp)))))
                   | _ ->
                        raise (IRException (pos, InternalError "store_closure: requires a closure"))
               in
                  blocks, exp
          | AtomChar _
          | AtomInt _
          | AtomFloat _ ->
               raise (IRException (pos, InternalError "store_closure: requires a closure"))

   (*
    * Store a pointer to memory.
    *)
   and store_pointer rvenv blocks pos addr a2 exp =
      let pos = string_pos "store_pointer" pos in
      let r0 = new_symbol_string "store_pointer_addr" in
         match a2 with
            AtomNil _ ->
               let exp =
                  LetAddr (r0, addr,
                  SetMem (AddrSafe (AtomVar r0), OpAtom (AtomInt Frame.hash_index),
                  SetMem (AddrSafeOff (AtomVar r0, AtomInt 4), OpAtom (AtomInt 0),
                  exp)))
               in
                  blocks, exp
          | AtomVar v ->
               let r1 = new_symbol_string "store_pointer_frame_index" in
               let r2 = new_symbol_string "store_pointer_offset" in
               let exp =
                  match rvenv_lookup_var rvenv pos v with
                     RunUnsafePointer (a_base, a_bounds) ->
                        LetAddr (r0, addr,
                        LetBinop (r1, XorIntOp, OpAddr (AddrSafeOff (a_base, AtomInt (-8))), OpAtom (AtomInt Frame.hash_index),
                        SetMem (AddrSafe (AtomVar r0), OpAtom (AtomVar r1),
                        SetMem (AddrSafeOff (AtomVar r0, AtomInt 4), OpAtom (AtomVar v),
                        exp))))
                   | RunSafePointer ->
                        LetAddr (r0, addr,
                        LetBinop (r1, XorIntOp, OpAddr (AddrSafeOff (AtomVar v, AtomInt (-8))), OpAtom (AtomInt Frame.hash_index),
                        SetMem (AddrSafe (AtomVar r0), OpAtom (AtomVar r1),
                        SetMem (AddrSafeOff (AtomVar r0, AtomInt 4), OpAtom (AtomInt 0),
                        exp))))
                   | _ ->
                        raise (IRException (pos, InternalError "store_pointer: requires a pointer"))
               in
                  blocks, exp
          | AtomChar _
          | AtomInt _
          | AtomFloat _ ->
               raise (IRException (pos, InternalError "store_pointer: requires a pointer, got a number"))

   (*
    * Load a value from memory.
    *)
   and make_subscript_expr tenv venv rvenv blocks pos v ty a1 a2 e =
      let pos = string_pos "make_subscript_expr" pos in
      let venv = venv_add venv v ty in
      let blocks, exp =
         make_unsafe_pointer_atom rvenv (int_pos 1 pos) a1 (fun rvenv a_base a_off a_bounds ->
               let size = sizeof tenv pos ty in
               let addr = AddrUnsafeMulOff (a_base, a2, size, a_off, a_bounds) in
                  if is_char_type2 tenv pos ty then
                     let rvenv = rvenv_add_scalar rvenv v in
                     let blocks, exp = make_expr tenv venv rvenv blocks e in
                        blocks, LetMemB (v, addr, exp)
                  else if is_fun_type2 tenv pos ty then
                     load_closure tenv venv rvenv blocks pos v addr e
                  else if is_pointer_type2 tenv pos ty then
                     load_pointer tenv venv rvenv blocks pos v addr e
                  else
                     let rvenv = rvenv_add_scalar rvenv v in
                     let blocks, exp = make_expr tenv venv rvenv blocks e in
                        blocks, LetMem (v, addr, exp))
      in
      let fir = LetSubscript (v, ty, a1, a2, fir_nop) in
         blocks, CommentFIR (fir, exp)

   (*
    * Store a value to an array.
    *)
   and make_set_subscript_expr tenv venv rvenv blocks pos a1 ty a2 a3 e =
      let pos = string_pos "make_set_subscript_expr" pos in
      let blocks, exp =
         make_unsafe_pointer_atom rvenv pos a1 (fun rvenv a_base a_off a_bounds ->
               let size = sizeof tenv pos ty in
               let addr = AddrUnsafeMulOff (a_base, a2, size, a_off, a_bounds) in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
                  if is_char_type2 tenv pos ty then
                     blocks, SetMemB (addr, OpAtom a3, exp)
                  else if is_fun_type2 tenv pos ty then
                     store_closure rvenv blocks pos addr a3 exp
                  else if is_pointer_type2 tenv pos ty then
                     store_pointer rvenv blocks pos addr a3 exp
                  else
                     blocks, SetMem (addr, OpAtom a3, exp))
      in
      let fir = Fc_ir_type.SetSubscript (a1, ty, a2, a3, fir_nop) in
         blocks, CommentFIR (fir, exp)

   (*
    * Get the address of an element in the array.
    *)
   and make_addr_of_subscript_expr tenv venv rvenv blocks pos v ty a1 a2 e =
      let pos = string_pos "make_addr_of_subscript_expr" pos in
      let blocks, exp =
         make_unsafe_pointer_atom rvenv pos a1 (fun rvenv a_base a_off a_bounds ->
               let venv = venv_add venv v ty in
               let rvenv = rvenv_add_unsafe_pointer rvenv v a_base a_bounds in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
               let size = sizeof tenv pos (type_of_deref2 tenv pos ty) in
               let exp = LetAddr (v, AddrSafeMul (a_off, a2, size), exp) in
                  blocks, exp)
      in
      let fir = Fc_ir_type.LetAddrOfSubscript (v, ty, a1, a2, fir_nop) in
         blocks, CommentFIR (fir, exp)

   (*
    * Projections.
    *)
   and make_project_expr tenv venv rvenv blocks pos v ty a l e =
      let pos = string_pos "make_project_expr" pos in
      let _, ty' = tenv_expand tenv pos (type_of_atom venv pos a) in
      let ty' = type_of_deref2 tenv pos ty' in
      let off = offsetof tenv pos ty' l in
      let venv = venv_add venv v ty in
      let blocks, exp =
         match a with
            AtomChar _
          | AtomInt _
          | AtomFloat _ ->
               raise (IRException (pos, InternalError "project requires a struct argument"))
          | AtomNil _ ->
               blocks, Call Frame.seg_fault_label
          | AtomVar v' ->
               let addr =
                  match rvenv_lookup_var rvenv pos v' with
                     RunUnsafePointer (a_base, a_bounds) ->
                        AddrUnsafeOff2 (a_base, AtomVar v', AtomInt off, a_bounds)
                   | RunScalar ->
                        raise (IRException (pos, InternalError "projection on a scalar"))
                   | RunClosure _ ->
                        raise (IRException (pos, InternalError "projection on a closure"))
                   | RunSafePointer ->
                        AddrSafeOff (AtomVar v', AtomInt off)
               in
                  if is_char_type2 tenv pos ty then
                     let rvenv = rvenv_add_scalar rvenv v in
                     let blocks, exp = make_expr tenv venv rvenv blocks e in
                        blocks, LetMemB (v, addr, exp)
                  else if is_fun_type2 tenv pos ty then
                     load_closure tenv venv rvenv blocks pos v addr e
                  else if is_pointer_type2 tenv pos ty then
                     load_pointer tenv venv rvenv blocks pos v addr e
                  else
                     let rvenv = rvenv_add_scalar rvenv v in
                     let blocks, exp = make_expr tenv venv rvenv blocks e in
                        blocks, LetMem (v, addr, exp)
      in
      let fir = Fc_ir_type.LetProject (v, ty, a, l, fir_nop) in
         blocks, CommentFIR (fir, exp)

   (*
    * Save a value to a struct.
    *)
   and make_set_project_expr tenv venv rvenv blocks pos a1 l ty a2 e =
      let pos = string_pos "make_set_project_expr" pos in
      let blocks, exp = make_expr tenv venv rvenv blocks e in
      let _, ty' = tenv_expand tenv pos (type_of_atom venv (int_pos 1 pos) a1) in
      let ty' = type_of_deref2 tenv pos ty' in
      let off = offsetof tenv (int_pos 2 pos) ty' l in
      let blocks, exp =
         match a1 with
            AtomChar _
          | AtomInt _
          | AtomFloat _ ->
               raise (IRException (pos, InternalError "project requires a struct argument"))
          | AtomNil _ ->
               blocks, Call Frame.seg_fault_label
          | AtomVar v ->
               let addr =
                  match rvenv_lookup_var rvenv (int_pos 3 pos) v with
                     RunUnsafePointer (a_base, a_bounds) ->
                        AddrUnsafeOff2 (a_base, AtomVar v, AtomInt off, a_bounds)
                   | RunScalar ->
                        raise (IRException (pos, InternalError "projection on a scalar"))
                   | RunClosure _ ->
                        raise (IRException (pos, InternalError "projection on a closure"))
                   | RunSafePointer ->
                        AddrSafeOff (AtomVar v, AtomInt off)
               in
               let pos = int_pos 4 pos in
                  if is_char_type2 tenv pos ty then
                     blocks, SetMemB (addr, OpAtom a2, exp)
                  else if is_fun_type2 tenv pos ty then
                     store_closure rvenv blocks pos addr a2 exp
                  else if is_pointer_type2 tenv pos ty then
                     store_pointer rvenv blocks pos addr a2 exp
                  else
                     blocks, SetMem (addr, OpAtom a2, exp)
      in
      let fir = Fc_ir_type.SetProject (a1, l, ty, a2, fir_nop) in
         blocks, CommentFIR (fir, exp)

   (*
    * Address of a field in a struct.
    *)
   and make_addr_of_project_expr tenv venv rvenv blocks pos v ty a l e =
      let pos = string_pos "make_addr_of_project_expr" pos in
      let _, ty' = tenv_expand tenv pos (type_of_atom venv pos a) in
      let ty' = type_of_deref2 tenv pos ty' in
      let off = offsetof tenv pos ty' l in
      let blocks, exp =
         match a with
            AtomChar _
          | AtomInt _
          | AtomFloat _ ->
               raise (IRException (pos, InternalError "project requires a struct argument"))
          | AtomNil _ ->
               blocks, Call Frame.seg_fault_label
          | AtomVar v' ->
               make_unsafe_pointer_atom rvenv pos a (fun rvenv a_base a_off a_bounds ->
                     let venv = venv_add venv v ty in
                     let rvenv = rvenv_add_unsafe_pointer rvenv v a_base a_bounds in
                     let blocks, exp = make_expr tenv venv rvenv blocks e in
                     let exp = LetBinop (v, PlusIntOp, OpAtom a_off, OpAtom (AtomInt off), exp) in
                        blocks, exp)
      in
      let fir = Fc_ir_type.LetAddrOfProject (v, ty, a, l, fir_nop) in
         blocks, CommentFIR (fir, exp)

   (*
    * Expand a memcpy.
    *)
   and make_memcpy_expr tenv venv rvenv blocks pos a1 ty a2 e =
      let pos = string_pos "make_memcpy_expr" pos in
      let blocks, exp = make_expr tenv venv rvenv blocks e in
      let bytes = sizeof tenv pos ty in
      let blocks, exp =
         make_unsafe_pointer_atom rvenv pos a1 (fun rvenv a1_base a1_off a1_bounds ->
               make_unsafe_pointer_atom rvenv pos a2 (fun rvenv a2_base a2_off a2_bounds ->
                     let r1 = new_symbol_string "memcpy_bounds1" in
                     let r2 = new_symbol_string "memcpy_bounds2" in

                     (* Abort if out-of-bounds *)
                     let abort_label = new_symbol_string "memcpy_abort" in
                     let exp4 = Call Frame.seg_fault_label in
                     let abort_block =
                        { block_global = false;
                          block_label = abort_label;
                          block_code = exp4
                        }
                     in

                     (* Bounds are ok *)
                     let bounds_ok_label = new_symbol_string "bounds_ok" in
                     let exp3 =
                        let r1 = new_symbol_string "data1_addr" in
                        let r2 = new_symbol_string "data2_addr" in
                        let r3 = new_symbol_string "data" in
                        let rec collect off =
                           if off = bytes then
                              exp
                           else
                              SetMem (AddrSafeOff (AtomVar r1, AtomInt off),
                                      OpAddr (AddrSafeOff (AtomVar r2, AtomInt off)),
                                      collect (off + 4))
                        in
                           LetAddr (r1, AddrSafeOff (a1_base, a1_off),
                                    LetAddr (r2, AddrSafeOff (a2_base, a2_off),
                                             collect 0))
                     in
                     let bounds_ok_block =
                        { block_global = false;
                          block_label = bounds_ok_label;
                          block_code = exp3
                        }
                     in

                     (* Bounds1 is ok *)
                     let bounds1_ok_label = new_symbol_string "bounds1_ok" in
                     let exp2 =
                        LetBinop (r2, PlusIntOp, OpAtom a2_off, OpAtom (AtomInt bytes),
                                  CJump (ULEIntOp, OpAtom (AtomVar r2), OpAtom a2_bounds,
                                         bounds_ok_label, abort_label))
                     in
                     let bounds1_ok_block =
                        { block_global = false;
                          block_label = bounds1_ok_label;
                          block_code = exp2
                        }
                     in

                     let exp1 =
                        LetBinop (r1, PlusIntOp, OpAtom a1_off, OpAtom (AtomInt bytes),
                                  CJump (ULEIntOp, OpAtom (AtomVar r1), OpAtom a1_bounds,
                                         bounds1_ok_label, abort_label))
                     in
                     let blocks =
                        abort_block
                        :: bounds_ok_block
                        :: bounds1_ok_block
                        :: blocks
                     in
                        blocks, exp1))
      in
      let fir = Memcpy (a1, ty, a2, fir_nop) in
         blocks, CommentFIR (fir, exp)

   (*
    * Allocation.
    * We assume thkat there is space (that it has been
    * check with Reserve).
    *)
   and make_alloc_expr tenv venv rvenv blocks pos v ty e =
      let pos = string_pos "make_alloc_expr" pos in
      let venv = venv_add venv v (TyArray (ty, None)) in
      let size = sizeof tenv pos ty in
      let bytes = size + 8 in
      let words = size / 4 in

      (* Temporaries *)
      let r0 = new_symbol_string "alloc_block" in
      let r1 = new_symbol_string "alloc_index" in

      (* Add the the environment *)
      let v_base, rvenv =
         match rvenv_lookup_type rvenv (int_pos 1 pos) v with
            TyRunUnsafePointer ->
               let s = Symbol.to_string v in
               let v_base = new_symbol_string (s ^ "_base") in
               let rvenv = rvenv_add_unsafe_pointer rvenv v (AtomVar v_base) (AtomInt size) in
                  v_base, rvenv
          | TyRunSafePointer ->
               let rvenv = rvenv_add_safe_pointer rvenv v in
                  v, rvenv
          | TyRunClosure
          | TyRunScalar ->
               raise (IRException (pos, InternalError "can't allocate except to a pointer"))
      in

      (* Allocate the block *)
      let blocks, exp = make_expr tenv venv rvenv blocks e in
      let exp =
         if v = v_base then
            exp
         else
            LetAtom (v, AtomInt 0, exp)
      in
      let exp =
         (* Allocate a pointer table entry *)
         LetBinop (Frame.pointer_free, MinusIntOp, OpAtom (AtomVar Frame.pointer_free), OpAtom (AtomInt 1),

         (* Assign the entry in the pointer table *)
         LetAtom (r1, AtomVar Frame.pointer_next,
         LetMem (Frame.pointer_next, AddrSafe (AtomVar r1),
         LetAtom (r0, AtomVar Frame.mem_next,
         LetBinop (Frame.mem_next, PlusIntOp, OpAtom (AtomVar r0), OpAtom (AtomInt bytes),
         SetMem (AddrSafe (AtomVar r1), OpAtom (AtomVar r0),
         LetBinop (r1, MinusIntOp, OpAtom (AtomVar r1), OpAtom (AtomVar Frame.pointer_base),
         LetBinop (r1, LsrIntOp, OpAtom (AtomVar r1), OpAtom (AtomInt 2),

         (* Initialize the memory block *)
         SetMem (AddrSafe (AtomVar r0), OpAtom (AtomVar r1),
         SetMem (AddrSafeOff (AtomVar r0, AtomInt 4), OpAtom (AtomInt size),
         LetBinop (v_base, PlusIntOp, OpAtom (AtomVar r0), OpAtom (AtomInt 8), exp)))))))))))
      in

      let fir = Fc_ir_type.LetAlloc (v, ty, fir_nop) in
         blocks, CommentFIR (fir, exp)

   (*
    * Reserve some memory.
    *)
   let make_reserve_expr rvenv blocks pos vars count size e =
      (*
       * Calculate which vars contain pointers.
       *)
      let pos = string_pos "make_reserve_expr" pos in
      let rec collect_vars = function
         v :: vars ->
            (match rvenv_lookup_var rvenv pos v with
                RunUnsafePointer (AtomVar v_base, _) ->
                   v_base :: collect_vars vars
              | RunUnsafePointer _ ->
                   raise (IRException (pos, InternalError "argument has the wrong type"))
              | RunSafePointer ->
                   v :: collect_vars vars
              | RunClosure _ ->
                   v :: collect_vars vars
              | RunScalar ->
                   collect_vars vars)
       | [] ->
            []
      in
      let vars = collect_vars vars in

      (*
       * GC block keeps track of the pointer variables.
       *)
      let gc_label = new_symbol_string "gc" in
      let alloc_ok_label = new_symbol_string "alloc_ok" in
      let gc_block =
         { block_global = false;
           block_label = gc_label;
           block_code = Reserve (vars, alloc_ok_label)
         }
      in

      (*
       * If allocation succeeds.
       *)
      let alloc_block =
         { block_global = false;
           block_label = alloc_ok_label;
           block_code = e
         }
      in

      (*
       * Check that there is enough heap space.
       *)
      let pointer_ok_label = new_symbol_string "pointer_ok" in
      let r0 = new_symbol_string "mem_size" in
      let exp1 =
         LetBinop (r0, MinusIntOp, OpAtom (AtomVar Frame.mem_limit), OpAtom (AtomVar Frame.mem_next),
                   CJump (ULEIntOp, OpAtom (AtomInt size), OpAtom (AtomVar r0),
                          alloc_ok_label, gc_label))
      in
      let pointer_ok_block =
         { block_global = false;
           block_label = pointer_ok_label;
           block_code = exp1
         }
      in

      (*
       * Check that there are enough pointer entries.
       *)
      let exp =
         CJump (ULEIntOp, OpAtom (AtomInt count), OpAtom (AtomVar Frame.pointer_free),
                pointer_ok_label, gc_label)
      in

      let blocks =
         gc_block
         :: pointer_ok_block
         :: alloc_block
         :: blocks
      in
         blocks, exp

   (************************************************************************
    * UTILS
    ************************************************************************)

   (*
    * Check max allocation in a function body.
    *)
   let rec check_alloc_expr tenv e =
      let pos = string_pos "check_alloc_expr" (atom_pos e) in
         match e with
            Fc_ir_type.LetAtom (v, ty, a, e) ->
               check_alloc_expr tenv e
          | Fc_ir_type.LetUnop (v, ty, op, a, e) ->
               check_alloc_expr tenv e
          | Fc_ir_type.LetBinop (v, ty, op, a1, a2, e) ->
               check_alloc_expr tenv e
          | Fc_ir_type.TailCall (f, args) ->
               0, 0
          | Fc_ir_type.LetClosure (v, ty, f, args, e) ->
               check_alloc_expr tenv e
          | Fc_ir_type.LetExtCall (v, f, ty, args, e) ->
               check_alloc_expr tenv e
          | Fc_ir_type.IntCall (ty, f, args) ->
               0, 0
          | Fc_ir_type.IfThenElse (op, a1, a2, e1, e2) ->
               let count1, size1 = check_alloc_expr tenv e1 in
               let count2, size2 = check_alloc_expr tenv e2 in
                  max count1 count2, max size1 size2
          | Fc_ir_type.SetSubscript (a1, ty, a2, a3, e) ->
               check_alloc_expr tenv e
          | Fc_ir_type.LetSubscript (v, ty, a1, a2, e) ->
               check_alloc_expr tenv e
          | Fc_ir_type.LetAddrOfSubscript (v, ty, a1, a2, e) ->
               check_alloc_expr tenv e
          | Fc_ir_type.SetProject (a1, l, ty, a2, e) ->
               check_alloc_expr tenv e
          | Fc_ir_type.LetProject (v, ty, a, l, e) ->
               check_alloc_expr tenv e
          | Fc_ir_type.LetAddrOfProject (v, ty, a, l, e) ->
               check_alloc_expr tenv e
          | Fc_ir_type.Memcpy (a1, ty, a2, e) ->
               check_alloc_expr tenv e
          | Fc_ir_type.LetAlloc (v, ty, e) ->
               let count1, size1 = check_alloc_expr tenv e in
               let size2 = sizeof tenv pos ty in
                  succ count1, size1 + size2 + 8
          | Fc_ir_type.SetVar _
          | Fc_ir_type.LetTypes _
          | Fc_ir_type.LetFuns _
          | Fc_ir_type.LetFuns2 _
          | Fc_ir_type.LetString _
          | Fc_ir_type.LetAddrOfVar _
          | Fc_ir_type.LetCopy _
          | Fc_ir_type.LetApply _
          | Fc_ir_type.Return _ ->
               raise (IRException (pos, IRLevel 2))

   (*
    * Get args from standard places.
    *)
   let rec get_realvars rvenv pos = function
      v :: vars ->
         (match rvenv_lookup_var rvenv pos v with
             RunUnsafePointer (AtomVar v_base, AtomVar v_bounds) ->
                v_base :: v :: v_bounds :: get_realvars rvenv pos vars
           | RunUnsafePointer _ ->
                raise (IRException (pos, InternalError "bogus function arguments"))
           | RunClosure v_fun ->
                v :: v_fun :: get_realvars rvenv pos vars
           | RunSafePointer
           | RunScalar ->
                v :: get_realvars rvenv pos vars)
    | [] ->
         []

   let make_stdargs_block tenv rvenv pos vars f f_ty closures blocks =
      let vars = get_realvars rvenv pos vars in
      let args = Frame.get_stdargs (List.length vars) in
      let code =
         List.fold_left2 (fun exp v1 v2 ->
               LetAtom (v1, AtomVar v2, exp)) (Jump f) vars args
      in
      let label, _ = rvenv_lookup_fun rvenv pos f in
      let block =
         { block_global = false;
           block_label = label;
           block_code = code
         }
      in
      let closure =
         { closure_size = sizeof_frame tenv pos f_ty;
           closure_label = label
         }
      in
         closure :: closures, block :: blocks

   let make_extargs_block tenv rvenv pos vars f f' f_ty closures blocks =
      let vars = get_realvars rvenv pos vars in
      let args = Frame.get_extargs (List.length vars) in
      let code =
         List.fold_left2 (fun exp v1 v2 ->
               LetAtom (v1, AtomVar v2, exp)) (Jump f') vars args
      in
      let block =
         { block_global = true;
           block_label = f;
           block_code = code
         }
      in
         closures, block :: blocks

   (************************************************************************
    * PROGRAM CONVERSION
    ************************************************************************)

   (*
    * Build the code for the program.
    *)
   let make_prog
       { prog_data_type = data_type;
         prog_init_sym = init_sym;
         prog_types = types;
         prog_strings = strings;
         prog_funs = funs
       } =

      (* Add the types *)
      let tenv =
         List.fold_left (fun tenv (v, ty) ->
               Fc_ir_env.tenv_add tenv v ty) empty_tenv types
      in

      (* Add the strings and funs *)
      let venv =
         List.fold_left (fun venv (v, _) ->
               venv_add venv v (TyArray (TyChar, None))) empty_venv strings
      in
      let venv =
         List.fold_left (fun venv (f, _, ty, _, _) ->
               venv_add venv f ty) venv funs
      in

      (* Add all the strings as safe pointers *)
      let rtenv =
         List.fold_left (fun rtenv (v, _) ->
               let ty_var = new_symbol v in
               let rtenv = rtenv_add_ty_var rtenv ty_var (SubstBool false) in
                  rtenv_add_var rtenv v (TyInferPointer ty_var)) rtenv_empty strings
      in

      (* Add all the functions to the type env *)
      let rtenv =
         List.fold_left (fun rtenv (f, gflag, f_ty, vars, e) ->
               let pos = string_pos "rtype_fun" (atom_pos e) in
               let ty_vars, _ = dest_fun_type1 tenv pos f_ty in
               let sflag = gflag = FunContClass in
               let _, rtenv =
                  List.fold_left2 (fun (sflag', rtenv) v ty ->
                        let rtenv, ty = bool_rtype_of_type sflag' tenv rtenv pos ty in
                           sflag, rtenv_add_var rtenv v ty) (false, rtenv) vars ty_vars
               in
                  rtenv_add_fun rtenv f vars) rtenv funs
      in

      (* Perform run-time type inference *)
      let rtenv =
         List.fold_left (fun rtenv (f, _, f_ty, vars, e) ->
               let pos = string_pos "rtype_fun" (atom_pos e) in
               let ty_vars, _ = dest_fun_type1 tenv pos f_ty in
               let venv =
                  List.fold_left2 (fun venv v ty ->
                        venv_add venv v ty) venv vars ty_vars
               in
                  rtype_expr tenv venv rtenv e) rtenv funs
      in

      (* Get the variable environment *)
      let pos = string_pos "make_prog" (atom_pos (TailCall (init_sym, []))) in
      let tenv = make_tenv tenv pos in
      let rvenv = rvenv_of_rtenv rtenv in

      (* Add all the strings *)
      let strings, rvenv =
         List.fold_left (fun (strings, rvenv) (v, s') ->
               match rvenv_lookup_type rvenv pos v with
                  TyRunUnsafePointer ->
                     let s = Symbol.to_string v in
                     let v_base = new_symbol_string (s ^ "_base") in
                     let a_bounds = AtomInt (succ (String.length s)) in
                     let rvenv = rvenv_add_unsafe_pointer rvenv v (AtomVar v_base) a_bounds in
                        (v, Some v_base, s') :: strings, rvenv
                | TyRunSafePointer ->
                     let rvenv = rvenv_add_safe_pointer rvenv v in
                        (v, None, s') :: strings, rvenv
                | TyRunClosure ->
                     raise (IRException (pos, InternalError "string can't be a closure"))
                | TyRunScalar ->
                     raise (IRException (pos, InternalError "string can't be a scalar"))) (**)
            ([], rvenv) strings
      in

      (* Add all the arguments to the funs *)
      let rvenv =
         List.fold_left (fun rvenv (_, _, _, vars, e) ->
               let pos = string_pos "rvenv_add_args" (atom_pos e) in
               List.fold_left (fun rvenv v ->
                     match rvenv_lookup_type rvenv pos v with
                        TyRunUnsafePointer ->
                           let s = Symbol.to_string v in
                           let v_base = new_symbol_string (s ^ "_base") in
                           let v_bounds = new_symbol_string (s ^ "_bounds") in
                              rvenv_add_unsafe_pointer rvenv v (AtomVar v_base) (AtomVar v_bounds)
                      | TyRunClosure ->
                           let s = Symbol.to_string v in
                           let v_fun = new_symbol_string (s ^ "_fun") in
                              rvenv_add_closure rvenv v v_fun
                      | TyRunSafePointer ->
                           rvenv_add_safe_pointer rvenv v
                      | TyRunScalar ->
                           rvenv_add_scalar rvenv v) rvenv vars) rvenv funs
      in

      (* Add well-known variables *)
      let rvenv = rvenv_add_type rvenv Frame.data_seg TyRunSafePointer in
      let rvenv = rvenv_add_safe_pointer rvenv Frame.data_seg in
      let rvenv = rvenv_add_closure rvenv Frame.exit_label Frame.exit_fun_label in

      (* Produce code for each function body *)
      let closures, blocks =
         List.fold_left (fun (closures, blocks) (f, gflag, f_ty, vars, e) ->
               let pos = string_pos "make_reserve_expr" (atom_pos e) in
               let ty_vars, _ = dest_fun_type2 tenv pos f_ty in
               let venv =
                  List.fold_left2 (fun venv v ty ->
                        venv_add venv v ty) venv vars ty_vars
               in
               let blocks, exp = make_expr tenv venv rvenv blocks e in
               let count, size = check_alloc_expr tenv e in
               let blocks, exp =
                  if count <> 0 then
                     make_reserve_expr rvenv blocks pos vars count size exp
                  else
                     blocks, exp
               in
               let f' =
                  if gflag = FunGlobalClass then
                     new_symbol f
                  else
                     f
               in
               let block =
                  { block_global = false;
                    block_label = f';
                    block_code = exp
                  }
               in
               let blocks = block :: blocks in
                  match gflag with
                     FunContClass ->
                        (* Make a block that copies args from the standard places *)
                        make_stdargs_block tenv rvenv pos vars f f_ty closures blocks
                   | FunGlobalClass ->
                        (* Make a block that copies args from the C places *)
                        make_extargs_block tenv rvenv pos vars f f' f_ty closures blocks
                   | FunLocalClass
                   | FunDeclClass ->
                        closures, blocks) ([], []) funs
      in

      (* Get size of data segment *)
      let data_size = sizeof tenv pos data_type in
         { aal_data_size = data_size;
           aal_init_sym = init_sym;
           aal_strings = strings;
           aal_blocks = blocks;
           aal_closures = closures
         }
end

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
