(*
 * Generate x86 code.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2001 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Symbol
open Fc_ir_type
open Fc_ir_exn_type
open Aal_type
open X86_inst_type
open X86_frame.X86Frame

module X86Codegen =
struct
   type prog = X86_frame.X86Frame.prog
   type block = X86_frame.X86Frame.block
   type spset = X86_frame.X86Frame.spset

   (************************************************************************
    * UTILITIES
    ************************************************************************)

   (*
    * Powers of 2.
    *)
   let is_power_of_two i =
      let rec search k =
         if k > i || k = (1 lsl 30) then
            false
         else
            k = i || search (k lsl 1)
      in
         search 1

   let log2 i =
      let rec search j =
         let k = 1 lsl j in
            if k = i then
               j
            else
               search (succ j)
      in
         search 0

   (*
    * Get the address operand.
    * My need to generate temporaries.
    *)
   let code_addr spset addr cont =
      match addr with
         AddrSafe a ->
            let r = new_symbol_string "addr_safe" in
            let inst1 = code_load r (code_atom spset a) in
               inst1 :: cont (MemReg r)
       | AddrSafeOff (a, AtomInt 0) ->
            let r = new_symbol_string "addr_safe" in
            let inst1 = code_load r (code_atom spset a) in
               inst1 :: cont (MemReg r)
       | AddrSafeOff (a, AtomInt i) ->
            let r = new_symbol_string "addr_safe" in
            let inst1 = code_load r (code_atom spset a) in
               inst1 :: cont (MemRegOff (r, i))
       | AddrSafeOff (a1, a2) ->
            let r1 = new_symbol_string "addr_safe_base" in
            let r2 = new_symbol_string "addr_safe_off" in
            let inst1 = code_load r1 (code_atom spset a1) in
            let inst2 = code_load r2 (code_atom spset a2) in
               inst1 :: inst2 :: cont (MemRegRegOffMul (r1, r2, 0, 1))
       | AddrUnsafeOff (a1, AtomInt i, a3) ->
            let r1 = new_symbol_string "addr_unsafe_base" in
            let r3 = new_symbol_string "addr_unsafe_bounds" in
            let inst1 = code_load r3 (code_atom spset a3) in
            let inst2 =
               Inst { inst_dst = [];
                      inst_src = [r3];
                      inst_code = CMP (Register r3, ImmediateNumber i)
               }
            in
            let inst3 =
               Inst { inst_dst = [];
                      inst_src = [];
                      inst_code = JCC (ULE, ImmediateLabel seg_fault_label)
               }
            in
            let inst4 = code_load r1 (code_atom spset a1) in
               inst1 :: inst2 :: inst3 :: inst4 :: cont (MemRegOff (r1, i))
       | AddrUnsafeOff (a1, a2, a3) ->
            let r1 = new_symbol_string "addr_unsafe_base" in
            let r2 = new_symbol_string "addr_unsafe_off" in
            let r3 = new_symbol_string "addr_unsafe_bounds" in
            let inst1 = code_load r2 (code_atom spset a2) in
            let inst2 = code_load r3 (code_atom spset a3) in
            let inst3 =
               Inst { inst_dst = [];
                      inst_src = [r2; r3];
                      inst_code = CMP (Register r3, Register r2)
               }
            in
            let inst4 =
               Inst { inst_dst = [];
                      inst_src = [];
                      inst_code = JCC (ULE, ImmediateLabel seg_fault_label)
               }
            in
            let inst5 = code_load r1 (code_atom spset a1) in
               inst1 :: inst2 :: inst3 :: inst4 :: inst5 :: cont (MemRegRegOffMul (r1, r2, 0, 1))
       | AddrSafeOff2 (a1, AtomInt i, AtomInt j) ->
            let r1 = new_symbol_string "addr_safe_off2_base" in
            let inst1 = code_load r1 (code_atom spset a1) in
               inst1 :: cont (MemRegOff (r1, i + j))
       | AddrSafeOff2 (a1, AtomInt i, a3) ->
            let r1 = new_symbol_string "addr_safe_off2_base" in
            let r3 = new_symbol_string "addr_safe_off2_off2" in
            let inst1 = code_load r1 (code_atom spset a1) in
            let inst2 = code_load r3 (code_atom spset a3) in
               inst1 :: inst2 :: cont (MemRegRegOffMul (r1, r3, i, 1))
       | AddrSafeOff2 (a1, a2, AtomInt i) ->
            let r1 = new_symbol_string "addr_safe_off2_base" in
            let r2 = new_symbol_string "addr_safe_off2_off1" in
            let inst1 = code_load r1 (code_atom spset a1) in
            let inst2 = code_load r2 (code_atom spset a2) in
               inst1 :: inst2 :: cont (MemRegRegOffMul (r1, r2, i, 1))
       | AddrSafeOff2 (a1, a2, a3) ->
            let r1 = new_symbol_string "addr_safe_off2_base" in
            let inst1 = code_load r1 (code_atom spset a1) in
            let a2 = code_atom spset a2 in
            let a3 = code_atom spset a3 in
            let inst2 =
               Inst { inst_dst = [r1];
                      inst_src = r1 :: src_regs a2;
                      inst_code = ADD (Register r1, a2)
               }
            in
            let inst3 =
               Inst { inst_dst = [r1];
                      inst_src = r1 :: src_regs a3;
                      inst_code = ADD (Register r1, a3)
               }
            in
               inst1 :: inst2 :: inst3 :: cont (MemReg r1)
       | AddrUnsafeOff2 (a1, AtomInt i, AtomInt j, AtomInt k) ->
            let r1 = new_symbol_string "addr_unsafe_off_base" in
            let inst2 = code_load r1 (code_atom spset a1) in
            let insts = inst2 :: cont (MemRegOff (r1, i + j)) in
               if k <= i + j then
                  let inst1 =
                     Inst { inst_dst = [];
                            inst_src = [];
                            inst_code = JMP (ImmediateLabel seg_fault_label)
                     }
                  in
                     inst1 :: insts
               else
                  insts
       | AddrUnsafeOff2 (a1, AtomInt i, AtomInt j, a4) ->
            let a4 = code_atom spset a4 in
            let inst1 =
               Inst { inst_dst = [];
                      inst_src = src_regs a4;
                      inst_code = CMP (a4, ImmediateNumber (i + j))
               }
            in
            let inst2 =
               Inst { inst_dst = [];
                      inst_src = [];
                      inst_code = JCC (ULE, ImmediateLabel seg_fault_label)
               }
            in
            let r1 = new_symbol_string "addr_unsafe_off_base" in
            let inst3 = code_load r1 (code_atom spset a1) in
               inst1 :: inst2 :: inst3 :: cont (MemRegOff (r1, i + j))
       | AddrUnsafeOff2 (a1, a2, a3, a4) ->
            let r2 = new_symbol_string "addr_safe_off_off1" in
            let a1 = code_atom spset a1 in
            let a3 = code_atom spset a3 in
            let a4 = code_atom spset a4 in
            let inst1 = code_load r2 (code_atom spset a2) in
            let inst2 =
               Inst { inst_dst = [r2];
                      inst_src = r2 :: src_regs a3;
                      inst_code = ADD (Register r2, a3)
               }
            in
            let inst3 =
               Inst { inst_dst = [];
                      inst_src = r2 :: src_regs a4;
                      inst_code = CMP (Register r2, a4)
               }
            in
            let inst4 =
               Inst { inst_dst = [];
                      inst_src = [];
                      inst_code = JCC (UGE, ImmediateLabel seg_fault_label)
               }
            in
            let inst5 =
               Inst { inst_dst = [r2];
                      inst_src = r2 :: src_regs a1;
                      inst_code = ADD (Register r2, a1)
               }
            in
               inst1 :: inst2 :: inst3 :: inst4 :: inst5 :: cont (MemReg r2)
       | AddrSafeMul (a1, a2, i) when is_power_of_two i ->
            (match i with
                1 | 2 | 4 | 8 ->
                   let r1 = new_symbol_string "addr_safe_mul_base" in
                   let r2 = new_symbol_string "addr_safe_mul_index" in
                   let inst1 = code_load r1 (code_atom spset a1) in
                   let inst2 = code_load r2 (code_atom spset a2) in
                      inst1 :: inst2 :: cont (MemRegRegOffMul (r1, r2, 0, i))
              | _ ->
                   let r2 = new_symbol_string "addr_safe_mul_index" in
                   let a1 = code_atom spset a1 in
                   let inst1 = code_load r2 (code_atom spset a2) in
                   let inst2 =
                      Inst { inst_dst = [r2];
                             inst_src = [r2];
                             inst_code = SHL (Register r2, ImmediateNumber (log2 i))
                      }
                   in
                   let inst3 =
                      Inst { inst_dst = [r2];
                             inst_src = r2 :: src_regs a1;
                             inst_code = ADD (Register r2, a1)
                      }
                   in
                      inst1 :: inst2 :: inst3 :: cont (MemReg r2))
       | AddrSafeMul (a1, a2, i) ->
            let r2 = new_symbol_string "addr_safe_mul_index" in
            let a1 = code_atom spset a1 in
            let inst1 = code_load r2 (code_atom spset a2) in
            let inst2 =
               Inst { inst_dst = [r2];
                      inst_src = [r2];
                      inst_code = IMUL (Register r2, ImmediateNumber i)
               }
            in
            let inst3 =
               Inst { inst_dst = [r2];
                      inst_src = r2 :: src_regs a1;
                      inst_code = ADD (Register r2, a1)
               }
            in
               inst1 :: inst2 :: inst3 :: cont (MemReg r2)
       | AddrSafeMulOff (a1, a2, i, AtomInt j) when is_power_of_two i ->
            (match i with
                1 | 2 | 4 | 8 ->
                   let r1 = new_symbol_string "addr_safe_mul_base" in
                   let r2 = new_symbol_string "addr_safe_mul_index" in
                   let inst1 = code_load r1 (code_atom spset a1) in
                   let inst2 = code_load r2 (code_atom spset a2) in
                      inst1 :: inst2 :: cont (MemRegRegOffMul (r1, r2, j, i))
              | _ ->
                   let r2 = new_symbol_string "addr_safe_mul_index" in
                   let a1 = code_atom spset a1 in
                   let inst1 = code_load r2 (code_atom spset a2) in
                   let inst2 =
                      Inst { inst_dst = [r2];
                             inst_src = [r2];
                             inst_code = SHL (Register r2, ImmediateNumber (log2 i))
                      }
                   in
                   let inst3 =
                      Inst { inst_dst = [r2];
                             inst_src = r2 :: src_regs a1;
                             inst_code = ADD (Register r2, a1)
                      }
                   in
                      inst1 :: inst2 :: inst3 :: cont (MemRegOff (r2, j)))
       | AddrSafeMulOff (a1, a2, i, a3) ->
            let r2 = new_symbol_string "addr_safe_mul_index" in
            let a1 = code_atom spset a1 in
            let a3 = code_atom spset a3 in
            let inst1 = code_load r2 (code_atom spset a2) in
            let inst2 =
               Inst { inst_dst = [r2];
                      inst_src = [r2];
                      inst_code = IMUL (Register r2, ImmediateNumber i)
               }
            in
            let inst3 =
               Inst { inst_dst = [r2];
                      inst_src = r2 :: src_regs a1;
                      inst_code = ADD (Register r2, a1)
               }
            in
            let inst4 =
               Inst { inst_dst = [r2];
                      inst_src = r2 :: src_regs a3;
                      inst_code = ADD (Register r2, a3)
               }
            in
               inst1 :: inst2 :: inst3 :: inst4 :: cont (MemReg r2)
       | AddrUnsafeMulOff (a1, a2, i, a3, a4) when is_power_of_two i ->
            (match i with
                1 | 2 | 4 | 8 ->
                   let r2 = new_symbol_string "addr_unsafe_mul_off_index" in
                   let r3 = new_symbol_string "addr_unsafe_mul_off_off" in
                   let a1 = code_atom spset a1 in
                   let a4 = code_atom spset a4 in
                   let inst1 = code_load r2 (code_atom spset a2) in
                   let inst2 = code_load r3 (code_atom spset a3) in
                   let inst3 =
                      Inst { inst_dst = [r2];
                             inst_src = [r2; r3];
                             inst_code = LEA (Register r2, MemRegRegOffMul (r3, r2, 0, i))
                      }
                   in
                   let inst4 =
                      Inst { inst_dst = [];
                             inst_src = r2 :: src_regs a4;
                             inst_code = CMP (Register r2, a4)
                      }
                   in
                   let inst5 =
                      Inst { inst_dst = [];
                             inst_src = [];
                             inst_code = JCC (UGE, ImmediateLabel seg_fault_label)
                      }
                   in
                   let inst6 =
                      Inst { inst_dst = [r2];
                             inst_src = r2 :: src_regs a1;
                             inst_code = ADD (Register r2, a1)
                      }
                   in
                      inst1 :: inst2 :: inst3 :: inst4 :: inst5 :: inst6 :: cont (MemReg r2)
              | _ ->
                   let r2 = new_symbol_string "addr_unsafe_mul_off_index" in
                   let a1 = code_atom spset a1 in
                   let a3 = code_atom spset a3 in
                   let a4 = code_atom spset a4 in
                   let inst1 = code_load r2 (code_atom spset a2) in
                   let inst2 =
                      Inst { inst_dst = [r2];
                             inst_src = [r2];
                             inst_code = SHL (Register r2, ImmediateNumber (log2 i))
                      }
                   in
                   let inst3 =
                      Inst { inst_dst = [r2];
                             inst_src = r2 :: src_regs a3;
                             inst_code = ADD (Register r2, a3)
                      }
                   in
                   let inst4 =
                      Inst { inst_dst = [];
                             inst_src = r2 :: src_regs a4;
                             inst_code = CMP (Register r2, a4)
                      }
                   in
                   let inst5 =
                      Inst { inst_dst = [];
                             inst_src = [];
                             inst_code = JCC (UGE, ImmediateLabel seg_fault_label)
                      }
                   in
                   let inst6 =
                      Inst { inst_dst = [r2];
                             inst_src = r2 :: src_regs a1;
                             inst_code = ADD (Register r2, a1)
                      }
                   in
                      inst1 :: inst2 :: inst3 :: inst4 :: inst5 :: inst6 :: cont (MemReg r2))
       | AddrUnsafeMulOff (a1, a2, i, a3, a4) ->
            let r2 = new_symbol_string "addr_unsafe_mul_off_index" in
            let a1 = code_atom spset a1 in
            let a3 = code_atom spset a3 in
            let a4 = code_atom spset a4 in
            let inst1 = code_load r2 (code_atom spset a2) in
            let inst2 =
               Inst { inst_dst = [r2];
                      inst_src = [r2];
                      inst_code = IMUL (Register r2, ImmediateNumber i)
               }
            in
            let inst3 =
               Inst { inst_dst = [r2];
                      inst_src = r2 :: src_regs a3;
                      inst_code = ADD (Register r2, a3)
               }
            in
            let inst4 =
               Inst { inst_dst = [];
                      inst_src = r2 :: src_regs a4;
                      inst_code = CMP (Register r2, a4)
               }
            in
            let inst5 =
               Inst { inst_dst = [];
                      inst_src = [];
                      inst_code = JCC (UGE, ImmediateLabel seg_fault_label)
               }
            in
            let inst6 =
               Inst { inst_dst = [r2];
                      inst_src = r2 :: src_regs a1;
                      inst_code = ADD (Register r2, a1)
               }
            in
               inst1 :: inst2 :: inst3 :: inst4 :: inst5 :: inst6 :: cont (MemReg r2)

   (*
    * General operand conversion.
    *)
   let code_operand spset a cont =
      match a with
         OpAtom a ->
            cont (code_atom spset a)
       | OpAddr a ->
            code_addr spset a cont

   (************************************************************************
    * CODEGEN
    ************************************************************************)

   (*
    * Add an AAL comment.
    *)
   let code_aal code =
      Inst { inst_dst = [];
             inst_src = [];
             inst_code = CommentAAL code
      }

   (*
    * Code converter.
    *)
   let rec code_expr spset e =
      match e with
         LetAtom (v, a, e) ->
            code_atom_expr spset v a e
       | LetUnop (v, op, a, e) ->
            code_unop_expr spset v op a e
       | LetBinop (v, op, a1, a2, e) ->
            code_binop_expr spset v op a1 a2 e
       | Jump l ->
            code_jump_expr spset l
       | RJump (vars, v) ->
            code_rjump_expr spset vars v
       | CJump (op, a1, a2, e1, e2) ->
            code_cjump_expr spset op a1 a2 e1 e2
       | Call a ->
            code_call_expr spset a
       | Reserve (vars, v) ->
            code_reserve_expr spset vars v
       | LetAddr (v, a, e) ->
            code_addr_expr spset v a e
       | LetMem (v, addr, e) ->
            code_mem_expr spset v addr e
       | SetMem (addr, a, e) ->
            code_set_mem_expr spset addr a e
       | LetMemB (v, addr, e) ->
            code_mem_byte_expr spset v addr e
       | SetMemB (addr, a, e) ->
            code_set_mem_byte_expr spset addr a e
       | LetExtCall (v, f, args, e) ->
            code_extcall_expr spset v f args e
       | CommentNOP ->
            []
       | Aal_type.CommentFIR (exp, e) ->
            let inst0 =
               Inst { inst_dst = []; inst_src = []; inst_code = CommentFIR exp }
            in
               inst0 :: code_expr spset e

   (*
    * Code for an atom.
    *)
   and code_atom_expr spset v a e =
      let inst0 = code_aal (LetAtom (v, a, CommentNOP)) in
      let r = new_symbol_string "atom" in
      let v = code_var spset v in
      let a = code_atom spset a in
      let inst1 = code_load r a in
      let inst2 = code_store v r in
         inst0 :: inst1 :: inst2 :: code_expr spset e

   and code_addr_expr spset v a e =
      let inst0 = code_aal (LetAddr (v, a, CommentNOP)) in
         code_addr spset a (fun a ->
               let r = new_symbol_string "addr" in
               let v = code_var spset v in
               let inst1 =
                  Inst { inst_dst = [r];
                         inst_src = src_regs a;
                         inst_code = LEA (Register r, a)
                  }
               in
               let inst2 = code_store v r in
                  inst0 :: inst1 :: inst2 :: code_expr spset e)

   (*
    * Unary operation.
    *)
   and code_unop_expr spset v op a e =
      let inst0 = code_aal (LetUnop (v, op, a, CommentNOP)) in
      let r = new_symbol_string "unop" in
      let v = code_var spset v in
         code_operand spset a (fun a ->
               let inst1 = code_load r a in
               let inst2 =
                  match op with
                     UMinusIntOp ->
                        Inst { inst_dst = [r];
                               inst_src = [r];
                               inst_code = NEG (Register r)
                        }
                   | UCharOfIntOp ->
                        Inst { inst_dst = [r];
                               inst_src = [r];
                               inst_code = AND (Register r, ImmediateNumber 255)
                        }
                   | UMinusFloatOp
                   | UIntOfFloatOp
                   | UFloatOfIntOp ->
                        raise (Invalid_argument "unop: floats not implemented")
               in
               let inst3 = code_store v r in
                  inst0 :: inst1 :: inst2 :: inst3 :: code_expr spset e)

   (*
    * Binary operation.
    *)
   and code_binop_expr spset v op a1 a2 e =
      let inst0 = code_aal (LetBinop (v, op, a1, a2, CommentNOP)) in
      let v = code_var spset v in
         code_operand spset a1 (fun a1 ->
               code_operand spset a2 (fun a2 ->

                     (*
                      * Generate code for normal arithmetic.
                      *)
                     let std_int_op a1 a2 opcode =
                        let r = new_symbol_string "binop" in
                        let inst1 = code_load r a1 in
                        let inst2 =
                           Inst { inst_dst = [r];
                                  inst_src = r :: src_regs a2;
                                  inst_code = opcode (Register r) a2
                           }
                        in
                        let inst3 = code_store v r in
                           inst0 :: inst1 :: inst2 :: inst3 :: code_expr spset e
                     in
                     let std_shift_op a1 a2 opcode =
                        let r = new_symbol_string "shift" in
                        let inst1 = code_load r a1 in
                        let inst2 = code_load ecx a2 in
                        let inst3 =
                           Inst { inst_dst = [r];
                                  inst_src = [ecx; r];
                                  inst_code = opcode (Register r) (Register ecx)
                           }
                        in
                        let inst4 = code_store v r in
                           inst0 :: inst1 :: inst2 :: inst3 :: inst4 :: code_expr spset e
                     in
                        (*
                         * Generate code for any binop.
                         *)
                        match op with
                           PlusIntOp ->
                              std_int_op a1 a2 (fun a1 a2 -> ADD (a1, a2))
                         | MinusIntOp ->
                              std_int_op a1 a2 (fun a1 a2 -> SUB (a1, a2))
                         | MulIntOp ->
                              (match a2 with
                                  ImmediateNumber i when is_power_of_two i ->
                                     let i = log2 i in
                                        std_shift_op a1 (ImmediateNumber i) (fun a1 a2 -> SHL (a1, a2))
                                | _ ->
                                     std_int_op a1 a2 (fun a1 a2 -> IMUL (a1, a2)))
                         | DivIntOp ->
                              (match a2 with
                                  ImmediateNumber i when is_power_of_two i ->
                                     let i = log2 i in
                                        std_shift_op a1 (ImmediateNumber i) (fun a1 a2 -> SAR (a1, a2))
                                | _ ->
                                     let r = new_symbol_string "div" in
                                     let inst1 = code_load r a2 in
                                     let inst2 = code_load eax a1 in
                                     let inst3 =
                                        Inst { inst_dst = [eax; edx];
                                               inst_src = [eax; edx; r];
                                               inst_code = IDIV (Register r)
                                        }
                                     in
                                     let inst4 = code_store v eax in
                                        inst1 :: inst2 :: inst3 :: inst4 :: code_expr spset e)
                         | RemIntOp ->
                              (match a2 with
                                  ImmediateNumber i when is_power_of_two i ->
                                     let i = pred i in
                                        std_int_op a1 (ImmediateNumber i) (fun a1 a2 -> AND (a1, a2))
                                | _ ->
                                     let r = new_symbol_string "div" in
                                     let inst1 = code_load r a2 in
                                     let inst2 = code_load eax a1 in
                                     let inst3 =
                                        Inst { inst_dst = [eax; edx];
                                               inst_src = [eax; edx; r];
                                               inst_code = IDIV (Register r)
                                        }
                                     in
                                     let inst4 = code_store v edx in
                                        inst1 :: inst2 :: inst3 :: inst4 :: code_expr spset e)
                         | BAndIntOp ->
                              std_int_op a1 a2 (fun a1 a2 -> AND (a1, a2))
                         | BOrIntOp ->
                              std_int_op a1 a2 (fun a1 a2 -> OR (a1, a2))
                         | XorIntOp ->
                              std_int_op a1 a2 (fun a1 a2 -> XOR (a1, a2))
                         | LslIntOp ->
                              std_shift_op a1 a2 (fun a1 a2 -> SHL (a1, a2))
                         | LsrIntOp ->
                              std_shift_op a1 a2 (fun a1 a2 -> SHR (a1, a2))
                         | AsrIntOp ->
                              std_shift_op a1 a2 (fun a1 a2 -> SAR (a1, a2))
                         | PlusFloatOp
                         | MinusFloatOp
                         | MulFloatOp
                         | DivFloatOp ->
                              raise (Invalid_argument "float operations not implemented")))

   (*
    * Jump instruction.
    *)
   and code_jump_expr spset l =
      [code_aal (Jump l);
       Inst { inst_dst = [];
              inst_src = [];
              inst_code = JMP (ImmediateLabel l)
       }]

   and code_rjump_expr spset srcs a =
      let inst0 = code_aal (RJump (srcs, a)) in
         code_operand spset a (fun a ->
               [inst0;
                Inst { inst_dst = [];
                       inst_src = srcs @ src_regs a;
                       inst_code = JMP a
                }])

   and code_cjump_expr spset op a1 a2 l1 l2 =
      let r = new_symbol_string "cjump" in
      let inst0 = code_aal (CJump (op, a1, a2, l1, l2)) in
         code_operand spset a1 (fun a1 ->
               code_operand spset a2 (fun a2 ->
                     (* Reverse the comparison operator *)
                     let cc =
                        match op with
                           Aal_type.LTIntOp -> LE
                         | Aal_type.LEIntOp -> LT
                         | Aal_type.EQIntOp -> NEQ
                         | Aal_type.ULTIntOp -> ULE
                         | Aal_type.ULEIntOp -> ULT
                         | Aal_type.LTFloatOp
                         | Aal_type.LEFloatOp
                         | Aal_type.EQFloatOp ->
                              raise (Invalid_argument "float operations not implemented")
                     in
                     let inst1 = code_load r a2 in
                     let inst2 =
                        Inst { inst_dst = [];
                               inst_src = r :: src_regs a1;
                               inst_code = CMP (Register r, a1)
                        }
                     in
                     let inst3 =
                        Inst { inst_dst = [];
                               inst_src = [];
                               inst_code = JCC (cc, ImmediateLabel l2)
                        }
                     in
                     let inst4 =
                        Inst { inst_dst = [];
                               inst_src = [];
                               inst_code = JMP (ImmediateLabel l1)
                        }
                     in
                        [inst0; inst1; inst2; inst3; inst4]))

   (*
    * Call.
    *)
   and code_call_expr spset l =
      [code_aal (Call l);
       Inst { inst_dst = [];
              inst_src = [];
              inst_code = CALL (ImmediateLabel l)
       }]

   (*
    * Memory allocation.
    * Don't have to do anything in this pass.
    *)
   and code_reserve_expr spset vars v =
      let args = List.map (code_var spset) vars in
      let srcs =
         List.fold_left (fun srcs a ->
               src_regs a @ srcs) [] args
      in
         [code_aal (Reserve (vars, v));
          Inst { inst_dst = [];
                 inst_src = srcs;
                 inst_code = RES (args, ImmediateLabel v)
          }]

   (*
    * Memory operations.
    *)
   and code_mem_expr spset v a e =
      let inst0 = code_aal (LetMem (v, a, CommentNOP)) in
         code_addr spset a (fun a ->
               let v = code_var spset v in
               let r = new_symbol_string "let_mem" in
               let inst1 = code_load r a in
               let inst2 = code_store v r in
                  inst0 :: inst1 :: inst2 :: code_expr spset e)

   and code_set_mem_expr spset addr a e =
      let inst0 = code_aal (SetMem (addr, a, CommentNOP)) in
         code_operand spset a (fun a2 ->
               let r = new_symbol_string "set_mem" in
               let inst1 = code_load r a2 in
                  code_addr spset addr (fun a1 ->
                        let inst2 = code_store a1 r in
                           inst0 :: inst1 :: inst2 :: code_expr spset e))

   and code_mem_byte_expr spset v a e =
      let inst0 = code_aal (LetMemB (v, a, CommentNOP)) in
         code_addr spset a (fun a ->
               let v = code_var spset v in
               let r = new_symbol_string "let_mem" in
               let inst1 = code_load_byte r a in
               let inst2 = code_store v r in
                  inst0 :: inst1 :: inst2 :: code_expr spset e)

   and code_set_mem_byte_expr spset addr a e =
      let inst0 = code_aal (SetMemB (addr, a, CommentNOP)) in
         code_operand spset a (fun a2 ->
               let r = new_symbol_string "set_mem" in
               let inst1 = code_load r a2 in
                  code_addr spset addr (fun a1 ->
                        let inst2 = code_store_byte a1 r in
                           inst0 :: inst1 :: inst2 :: code_expr spset e))

   (*
    * External call.
    *)
   and code_extcall_expr spset v f args e =
      let inst0 = code_aal (LetExtCall (v, f, args, CommentNOP)) in
      let code = call_external spset v f args in
         inst0 :: code @ code_expr spset e

   (************************************************************************
    * OUTER FUNCTIONS
    ************************************************************************)

   (*
    * Collect defs, uses, moves, and jumps in the block.
    *)
   let rec collect_jumps = function
      Inst { inst_code = code } :: exps ->
         (* Get the rest of the program *)
         let jumps = collect_jumps exps in

            (* Add jumps *)
            (match code with
                JMP dst ->
                   dst_labels dst jumps
              | JCC (_, dst) ->
                   dst_labels dst jumps
              | _ ->
                   jumps)

    | Move _ :: exps ->
         collect_jumps exps

    | [] ->
         []

   (*
    * Block generator.
    *)
   let code_block spset
       { Aal_type.block_global = gflag;
         Aal_type.block_label = label;
         Aal_type.block_code = exp
       } =
      let exp = code_expr spset exp in
      let exp =
         if gflag then
            callee_header @ exp
         else
            exp
      in
      let jumps = collect_jumps exp in
         { block_global = gflag;
           block_label = label;
           block_code = exp;
           block_jumps = jumps
         }

   (*
    * Generic code generator.
    *)
   let compile_blocks spset blocks =
      List.map (code_block spset) blocks

   let compile spset
       { aal_data_size = data_size;
         aal_init_sym = init_sym;
         aal_strings = strings;
         aal_blocks = blocks;
         aal_closures = closures
       } =
      let blocks = compile_blocks spset blocks in
         { asm_data_size = data_size;
           asm_reg_size = 0;
           asm_init_sym = init_sym;
           asm_strings = strings;
           asm_blocks = blocks;
           asm_closures = closures
         }

   (*
    * Import printer.
    *)
   let vars_blocks = X86_util.X86Util.vars_blocks
   let subst_blocks = X86_util.X86Util.subst_blocks
   let print_blocks = X86_util.X86Util.print_blocks
end

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
