(*
 * X86 frame.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2001 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Symbol
open Fc_ir_type
open X86_inst_type

type label = symbol
type spset = SymbolSet.t

module X86Frame =
struct
   (************************************************************************
    * TYPES
    ************************************************************************)

   (*
    * Type of instructions.
    *)
   type inst = X86_inst_type.inst
   type operand = X86_inst_type.operand

   (*
    * Spill set gives translation for each of
    * the vars.
    *)
   type spset =
      { spset_index : int;
        spset_table : operand SymbolTable.t
      }

   (*
    * Internally an instruction is one of the three types.
    *)
   type inst_info =
      { inst_code : inst;
        inst_src : var list;
        inst_dst : var list
      }

   type move_info =
      { move_code : inst;
        move_src : var;
        move_dst : var
      }

   type t =
      Inst of inst_info
    | Move of move_info

   type code = t list

   (*
    * Type of assembly blocks.
    *)
   type block =
      { block_global : bool;
        block_label : label;
        block_code : code;
        block_jumps : label list
      }

   (*
    * External interface.
    *)
   type prog =
      { asm_data_size : int;               (* Size of the data segment *)
        asm_reg_size : int;                (* Number of spilled registers *)
        asm_init_sym : var;
        asm_strings : (var * var option * string) list;
        asm_blocks : block list;
        asm_closures : Aal_type.closure list
      }

   (************************************************************************
    * MACHINE-DEPENDENT SECTION
    ************************************************************************)

   (*
    * Names of the standard args.
    *)
   let eax = Symbol.add "eax"
   let ebx = Symbol.add "ebx"
   let ecx = Symbol.add "ecx"
   let edx = Symbol.add "edx"
   let esi = Symbol.add "esi"
   let edi = Symbol.add "edi"
   let esp = Symbol.add "esp"
   let ebp = Symbol.add "ebp"

   (*
    * General-purpose registers.
    *)
   let registers = [eax; ebx; ecx; edx; esi; edi]
   let registers_all = [eax; ebx; ecx; edx; esi; edi; esp; ebp]
   let callee_save = [ebx; ecx; edx; esi; edi; ebp]

   (*
    * Get the standard argument list.  Make up the rest on
    * the fly.
    *)
   let frame_args = ref [eax; ebx; ecx; edx]

   let get_stdargs argc =
      let args = !frame_args in
      let len = List.length args in
         if argc > len then
            let rec collect args i =
               if i = 0 then
                  args
               else
                  let arg = new_symbol_string "arg" in
                     collect (arg :: args) (pred i)
            in
            let args' = collect [] (argc - len) in
            let args = args @ args' in
               frame_args := args;
               args
         else if argc = len then
            args
         else
            fst (List_util.split argc args)

   (*
    * External args.  Args are loaded from the stack.
    *)
   let frame_eargs = ref []

   let get_extargs argc =
      let args = !frame_eargs in
      let len = List.length args in
         if argc > len then
            let rec collect args i =
               if i = 0 then
                  args
               else
                  let arg = new_symbol_string "ext_arg" in
                     collect (arg :: args) (pred i)
            in
            let args' = collect [] (argc - len) in
            let args = args @ args' in
               frame_eargs := args;
               args
         else if argc = len then
            args
         else
            fst (List_util.split argc args)

   (*
    * Name of the register that holds the data segment.
    *)
   let data_seg = Fc_ir_esc.data_sym
   let glob_regs = Symbol.add "glob_regs"

   (* Name of the "exit" function that returns a value to C *)
   let exit_label = Symbol.add "exit_frame"
   let exit_fun_label = Symbol.add "exit"
   let seg_fault_label = Symbol.add "seg_fault"

   (* Hash value for pointer entries *)
   let hash_index = 0x73ac95a2

   (* Memory space *)
   let mem_base = Symbol.add "mem_base"
   let mem_next = Symbol.add "mem_next"
   let mem_limit = Symbol.add "mem_limit"

   (* Pointer table descriptors *)
   let pointer_base = Symbol.add "pointer_base"
   let pointer_next = Symbol.add "pointer_next"
   let pointer_free = Symbol.add "pointer_free"
   let pointer_size = Symbol.add "pointer_size"

   (* Closure table descriptors *)
   let closure_base = Symbol.add "closure_base"
   let closure_size = Symbol.add "closure_size"

   (************************************************************************
    * SPILL SET
    ************************************************************************)

   (*
    * Empty spill set.
    *)
   let spset_empty =
      { spset_index = 0;
        spset_table = SymbolTable.create ()
      }

   (*
    * Size of spill set.
    *)
   let spset_size { spset_index = index } =
      index

   (*
    * Add a var to the spill set.
    *)
   let spset_spill spset v =
      let { spset_index = index;
            spset_table = table
          } = spset
      in
         if SymbolTable.mem table v then
            spset
         else
            { spset_index = index + 4;
              spset_table = SymbolTable.add table v (MemRegOff (ebp, index))
            }

   (*
    * Add a var.
    *)
   let spset_add_arg spset v arg =
      { spset with spset_table = SymbolTable.add spset.spset_table v arg }

   (*
    * Add a var that represents a label.
    *)
   let spset_label spset v =
      { spset with spset_table = SymbolTable.add spset.spset_table v (ImmediateLabel v) }

   let spset_clabel spset v =
      { spset with spset_table = SymbolTable.add spset.spset_table v (ImmediateCLabel v) }

   let spset_int spset v i =
      { spset with spset_table = SymbolTable.add spset.spset_table v (ImmediateNumber i) }

   (*
    * Combine two vars.
    *)
   let spset_add spset v1 v2 =
      { spset with spset_table = SymbolTable.add spset.spset_table v1 (Register v2) }

   (*
    * Translate a var.
    *)
   let rec spset_lookup spset v =
      try
         let src = SymbolTable.find spset.spset_table v in
            match src with
               Register v ->
                  spset_lookup spset v
             | _ ->
                  src
      with
         Not_found ->
            Register v

   (*
    * Add all the normally spilled vars.
    *)
   let label_vars =
      [data_seg;
       mem_base;
       mem_next;
       mem_limit;
       pointer_base;
       pointer_next;
       pointer_free;
       pointer_size;
       closure_size]

   let clabel_vars =
      [closure_base;
       exit_label;
       exit_fun_label;
       seg_fault_label;
       glob_regs]

   let initial_spset =
      let spset = List.fold_left spset_label spset_empty label_vars in
      let spset = List.fold_left spset_clabel spset clabel_vars in
         spset

   let stdargs_offset = (succ (List.length callee_save)) * 4

   let spset_stdargs spset =
      fst (List.fold_left (fun (spset, index) v ->
                 let spset = spset_add_arg spset v (MemRegOff (esp, index)) in
                 let index = index + 4 in
                    spset, index) (spset, stdargs_offset) !frame_eargs)

   (************************************************************************
    * CODE GENERATION UTILITIES
    ************************************************************************)

   (*
    * Operand for a register.
    *)
   let code_var spset v =
      spset_lookup spset v

   (*
    * Operand for an atom.
    *)
   let code_atom spset = function
      AtomChar c ->
         ImmediateNumber (Char.code c)
    | AtomInt i ->
         ImmediateNumber i
    | AtomFloat _ ->
         raise (Failure "code_atom: floats not implemented")
    | AtomNil _ ->
         raise (Invalid_argument "code_atom")
    | AtomVar v ->
         code_var spset v

   (*
    * Registers in a src operand.
    *)
   let src_regs = function
      ImmediateNumber _
    | ImmediateLabel _
    | ImmediateCLabel _ ->
         []
    | Register r ->
         [r]
    | MemReg r ->
         [r]
    | MemRegOff (r, _) ->
         [r]
    | MemRegRegOffMul (r1, r2, _, _) ->
         [r1; r2]

   let dst_regs = function
      Register r ->
         [r]
    | ImmediateNumber _
    | ImmediateLabel _
    | ImmediateCLabel _
    | MemReg _
    | MemRegOff _
    | MemRegRegOffMul _ ->
         []

   let dst_labels dst labels =
      match dst with
         ImmediateLabel l ->
            l :: labels
       | ImmediateCLabel l ->
            l :: labels
       | ImmediateNumber _
       | Register _
       | MemReg _
       | MemRegOff _
       | MemRegRegOffMul _ ->
            labels

   (*
    * Load a value into a register.
    *)
   let code_load r1 src =
      match src with
         Register r2 ->
            Move { move_dst = r1;
                   move_src = r2;
                   move_code = MOV (Register r1, Register r2)
            }
       | _ ->
            Inst { inst_dst = [r1];
                   inst_src = src_regs src;
                   inst_code = MOV (Register r1, src)
            }

   (*
    * Store a value from a register.
    *)
   let code_store dst r2 =
      match dst with
         Register r1 ->
            Move { move_dst = r1;
                   move_src = r2;
                   move_code = MOV (Register r1, Register r2)
            }
       | _ ->
            Inst { inst_dst = [];
                   inst_src = r2 :: src_regs dst;
                   inst_code = MOV (dst, Register r2)
            }

   (*
    * Load a value into a register.
    *)
   let code_load_byte r1 src =
      match src with
         Register r2 ->
            Move { move_dst = r1;
                   move_src = r2;
                   move_code = MOV (Register r1, Register r2)
            }
       | _ ->
            Inst { inst_dst = [r1];
                   inst_src = src_regs src;
                   inst_code = MOVB (Register r1, src)
            }

   (*
    * Store a value from a register.
    *)
   let code_store_byte dst r2 =
      match dst with
         Register r1 ->
            Move { move_dst = r1;
                   move_src = r2;
                   move_code = MOV (Register r1, Register r2)
            }
       | _ ->
            Inst { inst_dst = [];
                   inst_src = r2 :: src_regs dst;
                   inst_code = MOVB (dst, Register r2)
            }

   (*
    * Generate an external call.
    * On x86 Linux, all arguments are pushed on the stack.
    * The return value is in eax.
    *)
   let call_external spset v label args =
      let rec push_args = function
         a :: args ->
            let insts = push_args args in
            let a = code_atom spset a in
            let inst =
               Inst { inst_dst = [esp];
                      inst_src = esp :: src_regs a;
                      inst_code = PUSH a
               }
            in
               inst :: insts
       | [] ->
            let v = code_var spset v in
            let inst1 =
               Inst { inst_dst = [eax; ebx; ecx; edx; esi; edi; ebp];
                      inst_src = [esp];
                      inst_code = CALL (ImmediateLabel label)
               }
            in
            let inst2 = code_store v eax in
            let inst3 =
               Inst { inst_dst = [esp];
                      inst_src = [esp];
                      inst_code = ADD (Register esp, ImmediateNumber (List.length args * 4))
               }
            in
               [inst1; inst2; inst3]
      in
         push_args args

   (*
    * Global header saves callee-save registers.
    *)
   let callee_header =
      let save_regs =
         List.map (fun r ->
               Inst { inst_dst = [esp];
                      inst_src = [r];
                      inst_code = PUSH (Register r)
               }) callee_save
      in
      let load_ebp =
         Inst { inst_dst = [ebp];
                inst_src = [];
                inst_code = LEA (Register ebp, ImmediateLabel glob_regs)
         }
      in
         save_regs @ [load_ebp]
end

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
