(*
 * Actual X86 frame.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2001 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Symbol
open Fc_ir_type
open X86_inst_type

(*
 * standard definition of vars.
 *)
type label = Symbol.symbol
type spset = SymbolSet.t

(*
 * Frame defines calling convention.
 *)
module X86Frame :
sig
   (************************************************************************
    * TYPES
    ************************************************************************)

   (*
    * Type of instructions.
    *)
   type inst = X86_inst_type.inst
   type operand = X86_inst_type.operand
   type spset

   (*
    * Internally an instruction is one of the three types.
    *)
   type inst_info =
      { inst_code : inst;
        inst_src : var list;
        inst_dst : var list
      }

   type move_info =
      { move_code : inst;
        move_src : var;
        move_dst : var
      }

   type t =
      Inst of inst_info
    | Move of move_info

   type code = t list

   (*
    * Type of assembly blocks.
    *)
   type block =
      { block_global : bool;
        block_label : label;
        block_code : code;
        block_jumps : label list
      }

   (*
    * External interface.
    *)
   type prog =
      { asm_data_size : int;               (* Size of the data segment *)
        asm_reg_size : int;                (* Number of spilled registers *)
        asm_init_sym : var;
        asm_strings : (var * var option * string) list;
        asm_blocks : block list;
        asm_closures : Aal_type.closure list
      }

   (************************************************************************
    * MACHINE-DEPENDENT SECTION
    ************************************************************************)

   (* List of general-purpose registers *)
   val registers : var list
   val registers_all : var list

   (* Get the standard argument list: pass the param count *)
   val get_stdargs : int -> var list
   val get_extargs : int -> var list

   (* Get the standard argument list for external functions *)
   val call_external : spset -> var -> label -> atom list -> code

   (* Callee-header for global blocks *)
   val callee_header : code

   (* Name of the register that holds the data segment *)
   val data_seg : var
   val glob_regs : label

   (* Name of the "exit" function that returns a value to C *)
   val exit_label : label
   val exit_fun_label : label
   val seg_fault_label : label

   (* Hash value for pointer entries *)
   val hash_index : int

   (* Memory space *)
   val mem_base : var
   val mem_next : var
   val mem_limit : var

   (* Pointer table descriptors *)
   val pointer_base : var
   val pointer_next : var
   val pointer_free : var
   val pointer_size : var

   (* Closure table descriptors *)
   val closure_base : var
   val closure_size : var

   (* Standard register names *)
   val eax : var
   val ebx : var
   val ecx : var
   val edx : var
   val esi : var
   val edi : var
   val esp : var
   val ebp : var

   (* Spill set *)
   val initial_spset : spset
   val spset_stdargs : spset -> spset
   val spset_size : spset -> int
   val spset_spill : spset -> var -> spset
   val spset_label : spset -> var -> spset
   val spset_clabel : spset -> var -> spset
   val spset_int : spset -> var -> int -> spset
   val spset_add : spset -> var -> var -> spset
   val spset_lookup : spset -> var -> operand

   (* Basic code generation *)
   val code_var : spset -> var -> src
   val code_atom : spset -> atom -> src
   val src_regs : src -> reg list
   val dst_regs : dst -> reg list
   val dst_labels : dst -> label list -> label list
   val code_load : reg -> src -> t
   val code_store : dst -> reg -> t
   val code_load_byte : reg -> src -> t
   val code_store_byte : dst -> reg -> t
end

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
