/*
 * Runtime for x86.
 */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <memory.h>
#include <string.h>

#define INITIAL_HEAP_SIZE       (1 << 20)
#define INITIAL_PTR_COUNT       (1 << 16)
#define HASH_INDEX              (0xf3ac95a2)

/*
 * Two memory areas: from/to space.
 */
static void *to_space;
static void *to_scan;
static void *to_next;

/*
 * FC program variables.
 */
static unsigned long mem_size;
static unsigned long mem_live;
void *__mem_base;
void *__mem_next;
void *__mem_limit;

void **__pointer_base;
void **__pointer_next;
unsigned __pointer_free;
unsigned __pointer_size;

/*
 * Data areas.
 */
extern char __strings[];
extern char __exit_frame[];
extern char __data_seg[];
extern void *__glob_regs[];

/*
 * Program functions.
 */
extern int _fc_init(void);
extern int _main(int argc, char *argv);

/*
 * Debugging.
 */
static int trace_gc = 1;

/*
 * Allocate a string.
 */
static int alloc_string(char *s)
{
    void *next;
    char *p;
    int i, index;

    /* Copy it */
    i = (strlen(s) + 4) & ~3;
    p = malloc(i + 8);
    index = __pointer_next - __pointer_base;
    ((int *) p)[0] = index;
    ((int *) p)[1] = i;
    strcpy(p + 8, s);

    /* Link it in */
    __pointer_free--;
    next = *__pointer_next;
    *__pointer_next = p;
    __pointer_next = next;

    /* Return it */
    return index;
}

/*
 * Copy the argv arguments.
 */
static char *copy_argv(int argc, char **argv)
{
    char *fc_argv;
    void *next;
    int i;

    /* Allocate the argv block */
    fc_argv = malloc(sizeof(void *) * (argc + 1) * 2);
    ((int *) fc_argv)[0] = __pointer_next - __pointer_base;
    ((int *) fc_argv)[1] = sizeof(void *) * argc * 2;

    /* Link it */
    __pointer_free--;
    next = *__pointer_next;
    *__pointer_next = fc_argv;
    __pointer_next = next;

    /* Copy the strings */
    fc_argv += 8;
    for(i = 0; i != argc; i++) {
        ((int *) fc_argv)[i * 2] = alloc_string(argv[i]) ^ HASH_INDEX;
        ((int *) fc_argv)[i * 2 + 1] = 0;
    }
    return fc_argv;
}    

/*
 * Main function initializes the memory areas, then jumps.
 */
int main(int argc, char **argv)
{
    int i, scount, size;
    char *strings, *fc_argv;
    void **p, *next;

    /* Print GC messages */
    trace_gc = getenv("NO_TRACE_GC") == 0;

    /* Allocate memory */
    mem_size = INITIAL_HEAP_SIZE;
    to_space = malloc(mem_size);
    __mem_base = malloc(mem_size);
    __pointer_base = (void **) malloc(INITIAL_PTR_COUNT * sizeof(void *));
    if(__mem_base == 0 || to_space == 0 || __pointer_base == 0) {
        fprintf(stderr, "%s: no memory\n", argv[0]);
        return 1;
    }

    /* Initialize the memory area */
    __mem_next = __mem_base;
    __mem_limit = __mem_base + mem_size;

    /* Link the pointer area */
    p = __pointer_base;
    for(i = 0; i != INITIAL_PTR_COUNT - 1; i++) {
        *p = (void *) (p + 1);
        p++;
    }
    *p = 0;

    /* Initialize the pointer area */
    __pointer_next = __pointer_base;
    __pointer_free = INITIAL_PTR_COUNT - 1;
    __pointer_size = INITIAL_PTR_COUNT;

    /* Link all the strings into the pointer area */
    strings = __strings;
    scount = *(int *)strings;
    strings += sizeof(int);
    __pointer_free -= scount + 2;

    /* Scan all the strings */
    p = __pointer_next;
    next = *p;
    *p = __exit_frame - 8;
    p = next;
    next = *p;
    *p = __data_seg - 8;
    p = next;
    while(scount--) {
        next = *p;
        *p = strings;
        p = next;
        size = ((int *)strings)[1];
        strings += size + 8;
    }
    __pointer_next = p;

    /* Copy argv */
    fc_argv = copy_argv(argc, argv);

    /* Call initialization */
    _fc_init();

    /* Now execute the program */
    i = _main(argc, fc_argv);
    fprintf(stderr, "FC: program exited with code %d\n", i);
    return i;
}

/*
 * Utilities.
 */
int _print_string(char *s)
{
    printf("%s", s);
    return 0;
}

int _print_int(int i)
{
    printf("%d", i);
    return 0;
}

int _atoi(char *s)
{
    return atoi(s);
}

/*
 * Seg fault handler.
 */
int __seg_fault()
{
    fprintf(stderr, "FC: Segmentation fault (no core dumped)\n");
    exit(1);
}

/************************************************************************
 * GARBAGE COLLECTION
 ************************************************************************/

/*
 * On the x86, a field is an unsigned int.
 */
typedef unsigned int fc_field;
typedef unsigned int fc_unsigned;

/*
 * Register indexes.
 */
#define EAX             0
#define EBX             1
#define ECX             2
#define EDX             3
#define ESI             4
#define EDI             5
#define EBP             6
#define MAX_REGS        7
#define CALLER          7

/*
 * Copy a block from the from_space to the to_space.
 * Update the pointer to the to-space.
 */
static void copy_block(void *blockp)
{
    fc_unsigned index, size;

    /* Ignore blocks outside the from space */
    if(blockp < __mem_base || blockp >= __mem_limit)
        return;

    /* Check that the block is valid */
    index = ((fc_field *)blockp)[0];
    if(index >= __pointer_size)
        return;
    if(__pointer_base[index] != blockp)
        return;

    /* Update the block pointer to the new location */
    __pointer_base[index] = to_next;
    
    /* Copy the entire block to the to_space */
    size = ((fc_field *)blockp)[1] + 8;
    memcpy(to_next, blockp, size);
    to_next += size;
}

/*
 * When a block is scanned, look through it
 * for valid pointers, and copy the blocks.
 */
static void scan_block()
{
    void *block2;
    fc_field *blockp;
    fc_unsigned size, index;

    /* Get the size and the data space */
    size = ((fc_field *)to_scan)[1];
    blockp = (fc_field *) (to_scan + 8);
    to_scan += size + 8;

    /* Now look for pointers */
    size >>= 2;
    while(size--) {
        /* Check for a valid index */
        index = *blockp++ ^ HASH_INDEX;
        if(index >= __pointer_size)
            continue;

        /* Check that the block is valid */
        block2 = __pointer_base[index];
        if(*(fc_field *)block2 != index)
            continue;

        /* Copy the block */
        copy_block(block2);
    }
}

/*
 * Now scan all the copied blocks.
 */
static void scan()
{
    while(to_scan != to_next)
        scan_block();
}

/*
 * Scan pointers left-to-right.
 */
static void scan_pointers()
{
    void **pointers, **from_pointers, **to_pointers, **pointer_limit;
    int i;

    /* If pointer table is getting full, resize it */
    if(__pointer_free < __pointer_size / 3) {
        if(trace_gc)
            fprintf(stderr, "\tResizing pointer table to %d entries\n", __pointer_size * 2);

        /* Allocate space and copy data */
        pointers = malloc(2 * sizeof(void *) * __pointer_size);
        if(pointers == 0) {
            fprintf(stderr, "Out of memory\n");
            exit(1);
        }

        /* Copy all pointers */
        __pointer_next = pointers;
        __pointer_free = -1;
        from_pointers = __pointer_base;
        to_pointers = pointers;
        pointer_limit = from_pointers + __pointer_size;
        for(i = __pointer_size; i; i--) {
            void *blockp = *from_pointers;
            if((blockp >= __mem_base && blockp < __mem_limit)
               || (blockp >= (void *) __pointer_base && blockp < (void *) pointer_limit)
               || blockp == 0) {
                *to_pointers = __pointer_next;
                __pointer_next = to_pointers;
                __pointer_free++;
            }
            else
                *to_pointers = blockp;
            from_pointers++;
            to_pointers++;
        }
        free(__pointer_base);
        __pointer_base = pointers;

        /* Add all the new pointers to the free list */
        to_pointers = pointers + __pointer_size;
        for(i = __pointer_size; i; i--) {
            *to_pointers = __pointer_next;
            __pointer_next = to_pointers;
            to_pointers++;
        }
        __pointer_free += __pointer_size;
        __pointer_size *= 2;
    }
    else {
        pointers = __pointer_base;
        for(i = __pointer_size; i; i--) {
            void *blockp = *pointers;
            if(blockp >= __mem_base && blockp < __mem_limit) {
                /* This is a free entry */
                *pointers = __pointer_next;
                __pointer_next = pointers;
                __pointer_free++;
            }
            pointers++;
        }
    }
}
    
/*
 * Garbage collector.
 * The argument is an array containing the registers.
 *
 * regs[0] = eax
 * regs[1] = ebx
 * regs[2] = ecx
 * regs[3] = edx
 * regs[4] = esi
 * regs[5] = edi
 * regs[6] = ebp
 * regs[7] = ret
 *
 * Get the register mask from the return address
 */
void gc(void **regs)
{
    int i, mask, resize_flag;
    void *from_space, *glob_regs, *blockp, **blockpp;
    int *maskp;

    /* Get the address of the register mask */
    maskp = regs[CALLER] + 2;
    maskp = (int *) (((long) maskp + 7) & ~7);

    /* If more than 50% used, increase the heap size */
    resize_flag = 0;
    if(mem_live > mem_size / 2) {
        free(to_space);
        mem_size *= 2;
        to_space = malloc(mem_size);
        resize_flag = 1;
        if(to_space == 0) {
            fprintf(stderr, "Out of memory\n");
            exit(1);
        }
    }

    /* Set up the memory areas */
    to_scan = to_space;
    to_next = to_space;

    /* Print out the register names that contain pointers */
    if(trace_gc) {
        fprintf(stderr, "Garbage collection from 0x%08x:\n", (unsigned int) regs[CALLER]);
        fprintf(stderr, "\tHeap size: %ld bytes\n", mem_size);
    }
    mask = *maskp++;
    for(i = 0; i != MAX_REGS; i++) {
        if(mask & (1 << i)) {
            blockp = to_next + 8;
            copy_block(regs[i] - 8);
            regs[i] = blockp;
        }
    }
    glob_regs = regs[EBP];
    while(1) {
        mask = *maskp++;
        if(mask == -1)
            break;
        blockp = to_next + 8;
        blockpp = (void **) (glob_regs + mask);
        copy_block(*blockpp - 8);
        *blockpp = blockp;
    }

    /* Now scan the to_space */
    scan();

    /* Clean up the pointer table */
    scan_pointers();

    /* Swap from/to space */
    from_space = __mem_base;
    __mem_base = to_space;
    __mem_next = to_next;
    __mem_limit = __mem_base + mem_size;
    to_space = from_space;

    /* Increase size of space? */
    if(resize_flag) {
        free(to_space);
        to_space = malloc(mem_size);
        if(to_space == 0) {
            fprintf(stderr, "Out of memory\n");
            exit(1);
        }
    }

    /* Print stats */
    mem_live = __mem_next - __mem_base;
    if(trace_gc)
        fprintf(stderr, "\tLive bytes: %ld\n", mem_live);
    if(mem_live > mem_size / 2) {
        if(trace_gc)
            fprintf(stderr, "\tRecursive GC call\n");
        gc(regs);
    }
}
