(*
 * Assembly printout.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2001 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Printf
open Symbol
open X86_inst_type
open X86_frame.X86Frame

(************************************************************************
 * PRINTING
 ************************************************************************)

(*
 * Print the AAL comments.
 *)
let print_asm_aal = ref false

(*
 * Registers are preceded by a % symbol.
 *)
let print_register out r =
   fprintf out "%%%s" (string_of_symbol r)

let print_symbol out l =
   let s = string_of_symbol l in
   let len = String.length s in
   let s =
      if len <> 0 then
         match s.[0] with
            '%' -> "_" ^ String.sub s 1 (pred len)
          | _ -> "__" ^ s
      else
         "_"
   in
      output_string out s

(*
 * Print an operand.
 *)
let print_operand out = function
   ImmediateNumber i -> fprintf out "$%d" i
 | ImmediateLabel l -> print_symbol out l
 | ImmediateCLabel l -> fprintf out "$%a" print_symbol l
 | Register r -> print_register out r
 | MemReg r -> fprintf out "(%a)" print_register r
 | MemRegOff (r, i) -> fprintf out "%d(%a)" i print_register r
 | MemRegRegOffMul (r1, r2, i1, i2) -> fprintf out "%d(%a,%a,%d)" i1 print_register r1 print_register r2 i2

let print_operand_cl out op =
   match op with
      Register r when r = ecx ->
         fprintf out "%%cl"
    | _ ->
         print_operand out op

(*
 * Print conditional jump.
 *)
let print_jcc out op =
   let s =
      match op with
         LT -> "jl"
       | LE -> "jle"
       | EQ -> "je"
       | NEQ -> "jne"
       | GT -> "jg"
       | GE -> "jge"
       | ULT -> "jb"
       | ULE -> "jbe"
       | UGT -> "ja"
       | UGE -> "jae"
   in
      fprintf out "%s" s

(*
 * Print the variable in a reserve.
 *)
let print_reserve out reg_count vars =
   let rec collect_regs = function
      arg :: args ->
         (match arg with
             Register r ->
                r :: collect_regs args
           | _ ->
                collect_regs args)
    | [] ->
         []
   in
   let rec collect_off = function
      arg :: args ->
         (match arg with
             MemReg ebp' ->
                if ebp' = ebp then
                   0 :: collect_off args
                else
                   raise (Invalid_argument "print_reserve: address is not relative to ebp")
           | MemRegOff (ebp', off) ->
                if ebp' = ebp then
                   off :: collect_off args
                else
                   raise (Invalid_argument "print_reserve: address is not relative to ebp")
           | Register _ ->
                collect_off args
           | _ ->
                raise (Invalid_argument "print_reserve"))
    | [] ->
         []
   in
   let regs = collect_regs vars in
   let offs = Sort.list (<) (collect_off vars) in

   (* Get the byte mask for the registers *)
   let rec reg_mask i mask = function
      r :: regs' ->
         let mask =
            if List.mem r regs then
               mask lor (1 lsl i)
            else
               mask
         in
            reg_mask (succ i) mask regs'
    | [] ->
         mask
   in
   let mask = reg_mask 0 0 registers_all in
      (* Print a mask of the register numbers that contain pointers *)
      fprintf out "\t/* For gc: registers that contain pointers */\n\t.align 8\n\t/* Regs:";
      List.iter (fun r -> fprintf out " %a" print_register r) regs;
      fprintf out " */\n\t.long %d\n" mask;

      (* Print the number of the global regs that contain pointers *)
      List.iter (fun off ->
            fprintf out "\t.long %d\n" off) offs;
      fprintf out "\t.long -1\n"

(*
 * Print an instruction.
 *)
let print_code out reg_count = function
   MOV (dst, src) ->
      fprintf out "\tmovl %a,%a\n" print_operand src print_operand dst
 | MOVB (dst, src) ->
      fprintf out "\tmovsbl %a,%a\n" print_operand src print_operand dst
 | NEG dst ->
      fprintf out "\tnegl %a\n" print_operand dst
 | LEA (dst, src) ->
      fprintf out "\tleal %a,%a\n" print_operand src print_operand dst
 | ADD (dst, src) ->
      fprintf out "\taddl %a,%a\n" print_operand src print_operand dst
 | SUB (dst, src) ->
      fprintf out "\tsubl %a,%a\n" print_operand src print_operand dst
 | MUL dst ->
      fprintf out "\tmull %a\n" print_operand dst
 | IMUL (dst, src) ->
      fprintf out "\timull %a,%a\n" print_operand src print_operand dst
 | DIV dst ->
      fprintf out "\tdivl %a\n" print_operand dst
 | IDIV dst ->
      fprintf out "\tcltd\n\tidivl %a\n" print_operand dst
 | AND (dst, src) ->
      fprintf out "\tandl %a,%a\n" print_operand src print_operand dst
 | OR (dst, src) ->
      fprintf out "\torl %a,%a\n" print_operand src print_operand dst
 | XOR (dst, src) ->
      fprintf out "\txorl %a,%a\n" print_operand src print_operand dst
 | SAL (dst, src) ->
      fprintf out "\tsall %a,%a\n" print_operand_cl src print_operand dst
 | SAR (dst, src) ->
      fprintf out "\tsarl %a,%a\n" print_operand_cl src print_operand dst
 | SHL (dst, src) ->
      fprintf out "\tshll %a,%a\n" print_operand_cl src print_operand dst
 | SHR (dst, src) ->
      fprintf out "\tshrl %a,%a\n" print_operand_cl src print_operand dst
 | PUSH src ->
      fprintf out "\tpushl %a\n" print_operand src
 | CALL src ->
      fprintf out "\tcall %a\n" print_operand src
 | RET ->
      fprintf out "\tret\n"
 | CMP (src1, src2) ->
      fprintf out "\tcmpl %a,%a\n" print_operand src2 print_operand src1
 | JMP dst ->
      (match dst with
          ImmediateLabel _ ->
             fprintf out "\tjmp %a\n" print_operand dst
        | _ ->
             fprintf out "\tjmp *%a\n" print_operand dst)
 | JCC (op, dst) ->
      (match dst with
          ImmediateLabel _ ->
             fprintf out "\t%a %a\n" print_jcc op print_operand dst
        | _ ->
             fprintf out "\t%a *%a\n" print_jcc op print_operand dst)
 | RES (vars, v) ->
      fprintf out "\tcall __gc\n\tjmp %a\n" print_operand v;
      print_reserve out reg_count vars
 | CommentFIR exp ->
      if !Aal_util.print_asm_fir then
         begin
            Format.set_formatter_out_channel out;
            Format.open_vbox 3;
            Format.print_string "/* FIR:";
            Format.print_space ();
            Fc_ir_util.print_expr exp;
            Format.close_box ();
            Format.print_space ();
            Format.print_string "*/";
            Format.print_newline ()
         end
 | CommentAAL exp ->
      if !print_asm_aal then
         begin
            Format.set_formatter_out_channel out;
            Format.open_vbox 3;
            Format.print_string "/* AAL:";
            Format.print_space ();
            Aal_util.print_expr exp;
            Format.close_box ();
            Format.print_space ();
            Format.print_string "*/";
            Format.print_newline ()
         end

(*
 * Print an instruction.
 *)
let print_inst out reg_count = function
   Inst { inst_dst = dst; inst_src = src; inst_code = code } ->
      print_code out reg_count code
 | Move { move_dst = dst; move_src = src; move_code = code } ->
      print_code out reg_count code

(*
 * Print a single block.
 *
 * BUG: every closure block must have a header
 * word that is the index into the closure table.
 * It would be better to generalize gflag so that it
 * would indicate which blocks represent closures.
 *)
let rec find_closure i label = function
   { Aal_type.closure_label = label' } :: closures ->
      if label = label' then
         Some i
      else
         find_closure (succ i) label closures
 | [] ->
      None

let print_block reg_count closures out
    { block_global = gflag;
      block_label = label;
      block_code = code
    } =
   if gflag then
      fprintf out ".global %a\n" print_symbol label;
   (match find_closure 1 label closures with
       Some i ->
          fprintf out "\t.align 4\n\t.long %d\n" i
     | None ->
          ());
   fprintf out "%a:\n" print_symbol label;
   List.iter (print_inst out reg_count) code

(*
 * Print the assembly code.
 *)
let print_prog out
    { asm_data_size = data_size;
      asm_reg_size = reg_size;
      asm_init_sym = init_sym;
      asm_strings = strings;
      asm_blocks = blocks;
      asm_closures = closures
    } =

   (* Print the strings *)
   fprintf out "/* String table */\n";
   fprintf out ".section .rodata\n\t.global __strings\n__strings:\n\t.long %d\n" (List.length strings);
   ignore (List.fold_left (fun i (v, v_base, s) ->
                 let v =
                    match v_base with
                       Some v_base -> v_base
                     | None -> v
                 in
                    fprintf out "\t.align 4
\t.long %d
\t.long %d
%a:\n\t.string \"%s\"\n"
                  i
                  ((String.length s + 4) land (lnot 3))
                  print_symbol v
                  (String.escaped s);
                    succ i)
           2 strings);

   (* Print the closures *)
   fprintf out "\n\n/* Closure table */\n";
   fprintf out "\t.text\n\t.align 4\n";
   fprintf out "__closure_size:\n\t.long %d\n" (succ (List.length closures));
   fprintf out "__closure_base:\n";
   fprintf out "\t.long 0\n\t.long __exit\n";
   List.iter (fun { Aal_type.closure_size = size; Aal_type.closure_label = label } ->
         fprintf out "\t.long %d\n" size;
         fprintf out "\t.long %a\n" print_symbol label) closures;

   (* Print the blocks *)
   fprintf out "\n\n/* Code */\n";
   fprintf out "\t.text\n\t.align 4\n";
   List.iter (print_block reg_size closures out) blocks;

   (* Allocate the data segment *)
   fprintf out "\n/* Data segment */\n\t.data\n\t.align 4\n";

   fprintf out "%a:\n\t.long __data_seg\n" print_symbol data_seg;
   fprintf out "\t.global __data_seg\n";
   fprintf out "\t.long 1\n";
   fprintf out "\t.long %d\n" data_size;
   fprintf out "__data_seg:\n";
   for i = 0 to pred (data_size / 4) do
      fprintf out "\t.long 0\n"
   done;

   (* Space for the global registers *)
   fprintf out "\n%a:\n" print_symbol glob_regs;
   for i = 0 to pred (reg_size / 4) do
      fprintf out "\t.long 0\n"
   done

(************************************************************************
 * LISTED VARS
 ************************************************************************)

(*
 * Get the vars in an operand.
 *)
let vars_operand vars = function
   ImmediateNumber _
 | ImmediateLabel _
 | ImmediateCLabel _ ->
      vars
 | Register r ->
      SymbolSet.add r vars
 | MemReg r ->
      SymbolSet.add r vars
 | MemRegOff (r, _) ->
      SymbolSet.add r vars
 | MemRegRegOffMul (r1, r2, _, _) ->
      SymbolSet.add r1 (SymbolSet.add r2 vars)

let vars_operand2 vars op1 op2 =
   vars_operand (vars_operand vars op1) op2

let vars_operand_mul vars op =
   let vars = SymbolSet.add eax vars in
   let vars = SymbolSet.add edx vars in
      vars_operand vars op

let vars_operands vars ops =
   List.fold_left vars_operand vars ops

(*
 * Get the vars in an instruction.
 *)
let vars_inst vars = function
   MOV  (dst, src) -> vars_operand2 vars dst src
 | MOVB (dst, src) -> vars_operand2 vars dst src
 | ADD  (dst, src) -> vars_operand2 vars dst src
 | LEA  (dst, src) -> vars_operand2 vars dst src
 | SUB  (dst, src) -> vars_operand2 vars dst src
 | IMUL (dst, src) -> vars_operand2 vars dst src
 | AND  (dst, src) -> vars_operand2 vars dst src
 | OR   (dst, src) -> vars_operand2 vars dst src
 | XOR  (dst, src) -> vars_operand2 vars dst src
 | SAL  (dst, src) -> vars_operand2 vars dst src
 | SAR  (dst, src) -> vars_operand2 vars dst src
 | SHL  (dst, src) -> vars_operand2 vars dst src
 | SHR  (dst, src) -> vars_operand2 vars dst src
 | CMP  (dst, src) -> vars_operand2 vars dst src
 | NEG  dst ->        vars_operand vars dst
 | MUL  dst ->        vars_operand_mul vars dst
 | DIV  dst ->        vars_operand_mul vars dst
 | IDIV dst ->        vars_operand_mul vars dst
 | PUSH src ->        vars_operand vars src
 | CALL src ->        vars_operand vars src
 | JMP  src ->        vars_operand vars src
 | JCC  (_, src) ->   vars_operand vars src
 | RET      ->        vars
 | RES  (src, _) ->   vars_operands vars src
 | CommentFIR _
 | CommentAAL _ ->    vars

(*
 * Get the vars in an instruction.
 *)
let vars_code vars inst =
   let code =
      match inst with
         Inst { inst_code = code } -> code
       | Move { move_code = code } -> code
   in
      vars_inst vars code

(*
 * Get the vars in a block.
 *)
let vars_block vars { block_code = code } =
   List.fold_left vars_code vars code

(*
 * Get the vars in the blocks.
 *)
let vars_blocks blocks =
   List.fold_left vars_block SymbolSet.empty blocks

(************************************************************************
 * OPERAND CALCULATIONS
 ************************************************************************)

(*
 * Get operands of code.
 *)
let get_operands_src src inst =
   Inst { inst_dst = [];
          inst_src = src_regs src;
          inst_code = inst
   }

let get_operands1 src inst =
   match src with
      Register r ->
         Inst { inst_dst = [r];
                inst_src = [r];
                inst_code = inst
         }
    | _ ->
         Inst { inst_dst = [];
                inst_src = src_regs src;
                inst_code = inst
         }

let get_operands_mul src inst =
   Inst { inst_dst = [eax; edx];
          inst_src = eax :: edx :: src_regs src;
          inst_code = inst
   }

let get_operands_list src inst =
   let src =
       List.fold_left (fun srcs src ->
             src_regs src @ srcs) [] src
   in
      Inst { inst_dst = [];
             inst_src = src;
             inst_code = inst
      }

let get_operands2 dst src inst =
   let dst, src =
      match dst with
         Register r ->
            [r], r :: src_regs src
       | _ ->
            [], src_regs dst @ src_regs src
   in
      Inst { inst_dst = dst;
             inst_src = src;
             inst_code = inst
      }

let get_operands_cmp src1 src2 inst =
   Inst { inst_dst = [];
          inst_src = src_regs src1 @ src_regs src2;
          inst_code = inst
   }

let get_operands_mov dst src inst =
   let dst, src =
      match dst with
         Register r ->
            [r], src_regs src
       | _ ->
            [], src_regs dst @ src_regs src
   in
      Inst { inst_dst = dst;
             inst_src = src;
             inst_code = inst
      }

let get_operands_call src inst =
   Inst { inst_dst = [eax; ebx; ecx; edx; esi; edi; ebp];
          inst_src = esp :: src_regs src;
          inst_code = inst
   }

let get_operands inst =
   match inst with
      MOV (Register dst, Register src) ->
         Move { move_dst = dst;
                move_src = src;
                move_code = inst
         }
    | MOV (dst, src) ->
         get_operands_mov dst src inst
    | MOVB (dst, src) ->
         get_operands_mov dst src inst
    | ADD (dst, src) ->
         get_operands2 dst src inst
    | LEA (dst, src) ->
         get_operands_mov dst src inst
    | SUB (dst, src) ->
         get_operands2 dst src inst
    | IMUL (dst, src) ->
         get_operands2 dst src inst
    | AND (dst, src) ->
         get_operands2 dst src inst
    | OR  (dst, src) ->
         get_operands2 dst src inst
    | XOR (dst, src) ->
         get_operands2 dst src inst
    | SAL (dst, src) ->
         get_operands2 dst src inst
    | SAR (dst, src) ->
         get_operands2 dst src inst
    | SHL (dst, src) ->
         get_operands2 dst src inst
    | SHR (dst, src) ->
         get_operands2 dst src inst
    | CMP (src1, src2) ->
         get_operands_cmp src1 src2 inst
    | NEG src ->
         get_operands1 src inst
    | MUL src ->
         get_operands_mul src inst
    | DIV src ->
         get_operands_mul src inst
    | IDIV src ->
         get_operands_mul src inst
    | JMP src ->
         get_operands_src src inst
    | JCC (cc, src) ->
         get_operands_src src inst
    | PUSH src ->
         get_operands_src src inst
    | CALL src ->
         get_operands_call src inst
    | RET ->
         Inst { inst_dst = [esp]; inst_src = [esp]; inst_code = inst }
    | RES (src, v) ->
         get_operands_list src inst
    | CommentFIR _
    | CommentAAL _ ->
         Inst { inst_dst = []; inst_src = []; inst_code = inst }

(************************************************************************
 * SUBSTITUTION
 ************************************************************************)

(*
 * Get the instruction in the code.
 *)
let code_of_inst = function
   Inst { inst_code = code } -> code
 | Move { move_code = code } -> code

(*
 * Adjust the dst/src of the instruction.
 *)
let subst_dst_src spset dst src =
   let rec collect_src srcs = function
      v :: src ->
         collect_src (src_regs (spset_lookup spset v) @ srcs) src
    | [] ->
         srcs
   in
   let rec collect_dst dsts srcs = function
      v :: dst ->
         (match spset_lookup spset v with
             Register v ->
                collect_dst (v :: dsts) srcs dst
           | v ->
                collect_dst dsts (src_regs v @ srcs) dst)
    | [] ->
         let srcs = collect_src srcs src in
            dsts, srcs
   in
      collect_dst [] [] dst

(*
 * Replace the code.
 *)
let replace_code spset inst code =
   match code with
      JMP (ImmediateLabel _) ->
         get_operands code
    | JMP _ ->
         let dst, src =
            match inst with
               Inst { inst_dst = dst;
                      inst_src = src
               } ->
                  dst, src
             | Move { move_dst = dst;
                      move_src = src
               } ->
                  [dst], [src]
         in
         let dst, src = subst_dst_src spset dst src in
            Inst { inst_dst = dst;
                   inst_src = src;
                   inst_code = code
            }
    | _ ->
         get_operands code

(*
 * Load a register.
 *)
let subst_register spset v cont =
   match spset_lookup spset v with
      Register r -> cont r
    | op ->
         let r = new_symbol_string "tmp" in
         let inst =
            Inst { inst_dst = [r];
                   inst_src = src_regs op;
                   inst_code = MOV (Register r, op)
            }
         in
            inst :: cont r

(*
 * Operand.
 *)
let subst_operand spset op cont =
   match op with
      ImmediateNumber _
    | ImmediateLabel _
    | ImmediateCLabel _ ->
         cont op
    | Register r ->
         cont (spset_lookup spset r)
    | MemReg r ->
         subst_register spset r (fun r ->
               cont (MemReg r))
    | MemRegOff (r, i) ->
         subst_register spset r (fun r ->
               cont (MemRegOff (r, i)))
    | MemRegRegOffMul (r1, r2, i1, i2) ->
         subst_register spset r1 (fun r1 ->
               subst_register spset r2 (fun r2 ->
                     cont (MemRegRegOffMul (r1, r2, i1, i2))))

(*
 * Check if two operands are the same.
 *)
let equal_operands spset op1 op2 =
   match op1, op2 with
      ImmediateNumber i1, ImmediateNumber i2 -> i1 = i2
    | ImmediateLabel l1, ImmediateLabel l2 -> l1 = l2
    | Register r1, Register r2 ->
         spset_lookup spset r1 = spset_lookup spset r2
    | MemReg r1, MemReg r2 ->
         spset_lookup spset r1 = spset_lookup spset r2
    | MemRegOff (r1, i1), MemRegOff (r2, i2) ->
         spset_lookup spset r1 = spset_lookup spset r2 && i1 = i2
    | MemRegRegOffMul (r11, r12, i11, i12), MemRegRegOffMul (r21, r22, i21, i22) ->
         (spset_lookup spset r11 = spset_lookup spset r21)
         && (spset_lookup spset r12 = spset_lookup spset r22)
         && (i11 = i21)
         && (i12 = i22)
    | _ ->
         false

(*
 * Replace the operands in a two-address dst/src instruction.
 *)
let subst_operand2 spset dst src cont =
   subst_operand spset dst (fun dst ->
         subst_operand spset src (fun src ->
               match dst, src with
                  Register _, _
                | _, Register _
                | _, ImmediateNumber _ ->
                     cont dst src
                | _ ->
                     let r = new_symbol_string "tmp" in
                     let inst =
                        Inst { inst_dst = [r];
                               inst_src = src_regs src;
                               inst_code = MOV (Register r, src)
                        }
                     in
                        inst :: cont dst (Register r)))

(*
 * Be careful about IMUL, the destination must always
 * be a register.
 *)
let subst_operand_mul spset dst src cont1 cont2 =
   subst_operand spset dst (fun dst ->
         subst_operand spset src (fun src ->
               match dst, src with
                  Register _, _ ->
                     cont1 dst src :: cont2
                | _ ->
                     let r = new_symbol_string "tmp" in
                     let inst1 = code_load r dst in
                     let inst2 = cont1 (Register r) src in
                     let inst3 = code_store dst r in
                        inst1 :: inst2 :: inst3 :: cont2))

(*
 * Be careful about LEA, it should not become a MOV.
 *)
let subst_lea spset dst src cont =
   subst_operand spset dst (fun dst ->
         subst_operand spset src (fun src ->
               match dst, src with
                  Register _, _
                | _, Register _ ->
                     cont dst src
                | _ ->
                     let r = new_symbol_string "tmp" in
                     let inst =
                        Inst { inst_dst = [r];
                               inst_src = src_regs src;
                               inst_code = LEA (Register r, src)
                        }
                     in
                        inst :: cont dst (MemReg r)))

(*
 * Reserve.
 *)
let spset_register spset r =
   match spset_lookup spset r with
      Register r ->
         r
    | _ ->
         raise (Invalid_argument "spset_register")

let rec subst_reserve spset = function
   src :: srcs ->
      (match src with
          ImmediateNumber _
        | ImmediateLabel _
        | ImmediateCLabel _ ->
             subst_reserve spset srcs
        | Register r ->
             spset_lookup spset r :: subst_reserve spset srcs
        | MemReg r ->
             MemReg (spset_register spset r) :: subst_reserve spset srcs
        | MemRegOff (r, i) ->
             MemRegOff (spset_register spset r, i) :: subst_reserve spset srcs
        | MemRegRegOffMul _ ->
             subst_reserve spset srcs)
 | [] ->
      []

(*
 * Instruction.
 *)
let subst_inst spset inst cont =
   match code_of_inst inst with
      MOV  (dst, src) ->
         if equal_operands spset dst src then
            cont
         else
            subst_operand2 spset dst src (fun dst src ->
                  replace_code spset inst (MOV (dst, src)) :: cont)
    | MOVB (dst, src) ->
         if equal_operands spset dst src then
            cont
         else
            subst_operand2 spset dst src (fun dst src ->
                  replace_code spset inst (MOV (dst, src)) :: cont)
    | ADD  (dst, src) ->
         subst_operand2 spset dst src (fun dst src ->
               replace_code spset inst (ADD (dst, src)) :: cont)
    | LEA  (dst, src) ->
         subst_lea spset dst src (fun dst src ->
               match src with
                  MemReg src ->
                     replace_code spset inst (MOV (dst, Register src)) :: cont
                | _ ->
                     replace_code spset inst (LEA (dst, src)) :: cont)
    | SUB  (dst, src) ->
         subst_operand2 spset dst src (fun dst src ->
               replace_code spset inst (SUB (dst, src)) :: cont)
    | IMUL (dst, src) ->
         subst_operand_mul spset dst src (fun dst src ->
               replace_code spset inst (IMUL (dst, src))) cont
    | AND  (dst, src) ->
         subst_operand2 spset dst src (fun dst src ->
               replace_code spset inst (AND (dst, src)) :: cont)
    | OR   (dst, src) ->
         subst_operand2 spset dst src (fun dst src ->
               replace_code spset inst (OR (dst, src)) :: cont)
    | XOR  (dst, src) ->
         subst_operand2 spset dst src (fun dst src ->
               replace_code spset inst (XOR (dst, src)) :: cont)
    | SAL  (dst, src) ->
         subst_operand2 spset dst src (fun dst src ->
               replace_code spset inst (SAL (dst, src)) :: cont)
    | SAR  (dst, src) ->
         subst_operand2 spset dst src (fun dst src ->
               replace_code spset inst (SAR (dst, src)) :: cont)
    | SHL  (dst, src) ->
         subst_operand2 spset dst src (fun dst src ->
               replace_code spset inst (SHL (dst, src)) :: cont)
    | SHR  (dst, src) ->
         subst_operand2 spset dst src (fun dst src ->
               replace_code spset inst (SHR (dst, src)) :: cont)
    | CMP  (dst, src) ->
         subst_operand2 spset dst src (fun dst src ->
               replace_code spset inst (CMP (dst, src)) :: cont)
    | NEG  dst ->
         subst_operand spset dst (fun dst ->
               replace_code spset inst (NEG dst) :: cont)
    | MUL  dst ->
         subst_operand spset dst (fun dst ->
               replace_code spset inst (MUL dst) :: cont)
    | DIV  dst ->
         subst_operand spset dst (fun dst ->
               replace_code spset inst (DIV dst) :: cont)
    | IDIV dst ->
         subst_operand spset dst (fun dst ->
               replace_code spset inst (IDIV dst) :: cont)
    | PUSH src ->
         subst_operand spset src (fun src ->
               replace_code spset inst (PUSH src) :: cont)
    | CALL src ->
         subst_operand spset src (fun src ->
               replace_code spset inst (CALL src) :: cont)
    | JMP  src ->
         subst_operand spset src (fun src ->
               replace_code spset inst (JMP src) :: cont)
    | JCC  (cc, src) ->
         subst_operand spset src (fun src ->
               replace_code spset inst (JCC (cc, src)) :: cont)
    | RET      ->
         inst :: cont
    | RES  (src, v) ->
         replace_code spset inst (RES (subst_reserve spset src, v)) :: cont
    | CommentFIR _
    | CommentAAL _ ->
         inst :: cont

let rec subst_code spset = function
   inst :: code ->
      let code = subst_code spset code in
         subst_inst spset inst code
 | [] ->
      []

(*
 * Rewrite regs according to the spset.
 *)
let subst_block spset
    { block_global = global;
      block_label = label;
      block_code = code;
      block_jumps = jumps
    } =
   { block_global = global;
     block_label = label;
     block_code = subst_code spset code;
     block_jumps = jumps
   }

(*
 * Wrap up in a module.
 *)
module X86Util =
struct
   type prog = X86_frame.X86Frame.prog
   type block = X86_frame.X86Frame.block
   type spset = X86_frame.X86Frame.spset

   let print_blocks out blocks =
      List.iter (print_block 0 [] out) blocks

   let print_prog = print_prog

   let vars_blocks = vars_blocks

   let subst_blocks spset blocks =
      List.map (subst_block spset) blocks
end

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
