(*
 * Generate the interference graph for the abstract
 * assembly instructions.
 *
 * ----------------------------------------------------------------
 *
 * Copyright (C) 1999 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * jyh@cs.caltech.edu
 *)
open Printf

open Debug
open Symbol
open Frame_type
open Codegen_type

(*
 * Debugging.
 *)
let debug_regalloc = ref false

(*
 * Build the dataflow graph.
 *)
module Dataflow (Frame : FrameSig) =
struct
   open Frame

   type block = Frame.block

   (*
    * Abbreviations for tables and graphs.
    *)
   module DGraph = Table_graph.DirectedGraph
   module DNTable = DGraph.Table

   module UGraph = Table_graph.UndirectedGraph
   module UNSet = UGraph.Set
   module UNTable = UGraph.Table

   (*
    * The interference graph is an indirected graph
    * where each node is a var.
    *)
   module IGraph = UGraph

   type igraph =
      { mutable igraph_graph     : IGraph.t;
        mutable igraph_node_vars : var IGraph.Table.t;
        mutable igraph_var_nodes : IGraph.node SymbolTable.t;
        mutable igraph_moves     : (IGraph.node * IGraph.node) list
      }

   (*
    * A basic block contains a sequence of assembly instructions,
    * it begins with a label, ends with a jump, and contains no
    * other labels or jumps.
    *
    * We use faster algorithms for computing defs/uses on
    * basic blocks.
    *)
   type block_aux =
      { block_global : bool;
        block_code : Frame.code;
        block_label : label;

        (* Defs and uses *)
        block_defs : var list;
        block_uses : SymbolSet.t;

        (* Dataflow *)
        mutable block_in : SymbolSet.t;
        mutable block_out : SymbolSet.t;

        (* Outward jumps *)
        block_jumps : label list
      }

   (************************************************************************
    * UTILITIES
    ************************************************************************)

   (*
    * Print a set.
    *)
   let print_set out s =
      SymbolSet.iter (fun v ->
            fprintf out " %s" (string_of_symbol v)) s

   let print_list out l =
      List.iter (fun v ->
            fprintf out " %s" (string_of_symbol v)) l

   let label_lookup table label =
      try SymbolTable.find table label with
         Not_found ->
            raise (Invalid_argument ("label_lookup: undefined label " ^ string_of_symbol label))

   let dntable_lookup table node =
      try DNTable.find table node with
         Not_found ->
            raise (Invalid_argument "dntable_lookup")

   (************************************************************************
    * CONTROL FLOW GRAPH                                                   *
    ************************************************************************)

   (*
    * Compute the defs in a basic block.
    * The defs are the union of all the defined values.
    *)
   let block_defs instrs =
      let rec collect defs = function
         inst :: rest ->
            let defs' =
               match inst with
                  Inst { inst_dst = defs } -> defs
                | Move { move_dst = dst } -> [dst]
            in
               collect (List.fold_left (fun s x -> SymbolSet.add x s) defs defs') rest
       | [] ->
            defs
      in
         collect SymbolSet.empty instrs

   (*
    * The uses are propagated backward, because a use is
    * not a use if it is preceded by a def to the same location.
    *)
   let rec block_uses = function
      inst :: rest ->
         let uses = block_uses rest in
         let defs', uses' =
            match inst with
               Inst { inst_dst = defs; inst_src = uses } -> defs, uses
             | Move { move_dst = dst; move_src = src } -> [dst], [src]
         in
         let uses = SymbolSet.subtract_list uses defs' in
            SymbolSet.add_list uses uses'
    | [] ->
         SymbolSet.empty

   (*
    * Build a block from the assembly instruction list.
    * The first instruction should be a label, and there
    * should be no jumps except for the last instruction.
    *)
   let basic_block { Frame.block_global = gflag;
                     Frame.block_label = label;
                     Frame.block_code = code;
                     Frame.block_jumps = jumps
       } =
      { block_global = gflag;
        block_code = code;
        block_label = label;
        block_defs = SymbolSet.to_list (block_defs code);
        block_uses = block_uses code;
        block_in = SymbolSet.empty;
        block_out = SymbolSet.empty;
        block_jumps = jumps
      }

   (*
    * Build the control flow graph for basic blocks.
    * The input is a list of basic blocks.
    *)
   let control_flow blocks =
      (* Build the list of basic blocks *)
      let blocks = List.map basic_block blocks in

      (* Build a graph, and add a node for each block *)
      let add_block (graph, label_table, node_table) block =
         let graph, node = DGraph.new_node graph in
         let label_table = SymbolTable.add label_table block.block_label node in
         let node_table = DNTable.add node_table node block in
            (graph, label_table, node_table)
      in
      let graph, label_table, node_table =
         List.fold_left add_block (DGraph.create (), SymbolTable.create (), DNTable.create ()) blocks
      in

      (* Add edges to all the blocks *)
      let new_edges graph block =
         let node1 = SymbolTable.find label_table block.block_label in
         let new_edge graph label =
            try
               let node2 = SymbolTable.find label_table label in
                  DGraph.new_edge graph node1 node2
            with
               Not_found ->
                  (*
                   * If the destination is not found, it
                   * is an external function, and we can just
                   * ignore it.
                   *)
                  graph
         in
            List.fold_left new_edge graph block.block_jumps
      in
      let graph = List.fold_left new_edges graph blocks in
         blocks, graph, label_table, node_table

   (************************************************************************
    * DATAFLOW GRAPH                                                       *
    ************************************************************************)

   (*
    * Compute the least-fixed point of the dataflow equations:
    *     in[n] = uses[n] + (out[n] - defs[n])
    *     out[n] = union s in succ[n]. in[n]
    *
    * The dataflow equations are monotone, and the size of the sets
    * is limited by the total number of vars in the program,
    * so we can calculate the least-fixed point by iteration
    * (and be sure of termination).  Since values are propagated
    * from outputs to inputs, we compute the fixpoint in
    * reverse depth-first order.
    *
    * The input includes the control-flow graph,
    * a table that maps nodes to basic blocks,
    * and a set of nodes.
    *)
   let data_flow graph table =
      (*
       * Compute the dataflow assignment for a single node.
       * Returns true iff the live sets have changed.
       *)
      let rec update updated = function
         node :: nodes ->
            let block = dntable_lookup table node in
            let succ = List.map (dntable_lookup table) (DGraph.succ graph node) in
            let live_out = List.fold_left (fun set block -> SymbolSet.union set block.block_in) SymbolSet.empty succ in
            let live_in = SymbolSet.union block.block_uses (SymbolSet.subtract_list live_out block.block_defs) in
               if !debug_regalloc then
                  begin
                     eprintf "Dataflow.update: %s\n" (string_of_symbol block.block_label);
                     eprintf "Old in: %a\n" print_set block.block_in;
                     eprintf "Old out: %a\n" print_set block.block_out;
                     eprintf "Defs: %a\n" print_list block.block_defs;
                     eprintf "Uses: %a\n" print_set block.block_uses;
                     eprintf "New in: %a\n" print_set live_in;
                     eprintf "New out: %a%t" print_set live_out eflush
                  end;

               if SymbolSet.equal block.block_in live_in && SymbolSet.equal block.block_out live_out then
                  update updated nodes
               else
                  begin
                     block.block_in <- live_in;
                     block.block_out <- live_out;
                     update true nodes
                  end
       | [] ->
            updated
      in

      (* Sort the nodes in reverse depth-first-order *)
      let nodes = List.rev (DGraph.sort graph) in

      (* Print messages during the updates *)
      let rec fixpoint iteration =
         if !debug_regalloc then
            eprintf "Dataflow iteration %d\n" iteration;
         if update false nodes then
            fixpoint (succ iteration)
         else if !debug_regalloc then
            eprintf "Dataflow complete%t" eflush
      in
         fixpoint 1

   (************************************************************************
    * INTERFERENCE GRAPH                                                   *
    ************************************************************************)

   (*
    * Get the node corresponding to the var.
    * If the var is not already listed, create a new node.
    *)
   let igraph_node_of_var igraph var =
      try SymbolTable.find igraph.igraph_var_nodes var with
         Not_found ->
            let graph, node = UGraph.new_node igraph.igraph_graph in
               igraph.igraph_graph <- graph;
               igraph.igraph_var_nodes <- SymbolTable.add igraph.igraph_var_nodes var node;
               igraph.igraph_node_vars <- UNTable.add igraph.igraph_node_vars node var;
               node

   let igraph_nodes_of_vars igraph (vars : SymbolSet.t) =
      SymbolSet.fold (fun nodes var ->
            UNSet.add (igraph_node_of_var igraph var) nodes) UNSet.empty vars

   (*
    * Add an interference edge between all the def and all vars
    * in the outgoing set.
    *)
   let add_interference_edge igraph (def : UGraph.node) (out : UNSet.t) =
      (* Check for self-cycles when adding edges *)
      let new_edge out =
         if not (UGraph.eq_node def out) then
            igraph.igraph_graph <- UGraph.new_edge igraph.igraph_graph def out
      in
         UNSet.iter new_edge out

   let add_interference_edges igraph (defs : UGraph.node list) (out : UNSet.t) =
      List.iter (fun def -> add_interference_edge igraph def out) defs

   (*
    * Add an interference edge for a move.
    *)
   let add_interference_move igraph (def : UGraph.node) (use : UGraph.node) (out : UNSet.t) =
      add_interference_edge igraph def (UNSet.remove use out);
      igraph.igraph_moves <- (def, use) :: igraph.igraph_moves

   (*
    * Update the interference graph for the instructions within
    * the block.  The live edges are compute from last to first
    * in the instruction order.
    *)
   let block_interference igraph { block_code = instrs; block_out = out } =
      let rec collect = function
         inst :: rest ->
            let live_out = collect rest in
               (match inst with
                   Inst { inst_src = uses; inst_dst = defs } ->
                      let uses = List.map (igraph_node_of_var igraph) uses in
                      let defs = List.map (igraph_node_of_var igraph) defs in
                         add_interference_edges igraph defs live_out;
                         UNSet.add_list (UNSet.subtract_list live_out defs) uses

                 | Move { move_src = src; move_dst = dst } ->
                      let src = igraph_node_of_var igraph src in
                      let dst = igraph_node_of_var igraph dst in
                         add_interference_move igraph dst src live_out;
                         UNSet.add src (UNSet.remove dst live_out))
       | [] ->
            igraph_nodes_of_vars igraph out
      in
         ignore (collect instrs)

   (*
    * Generate the interference graph from the live-sets
    * in the basic-blocks.
    *)
   let empty_igraph () =
      { igraph_graph = UGraph.create ();
        igraph_node_vars = UNTable.create ();
        igraph_var_nodes = SymbolTable.create ();
        igraph_moves = []
      }

   let interference_graph blocks =
      let igraph = empty_igraph () in
         List.iter (block_interference igraph) blocks;
         igraph

   (************************************************************************
    * MAIN                                                                 *
    ************************************************************************)

   (*
    * Create the interference graph.
    *)
   let create blocks =
      let blocks, graph, ltable, ntable = control_flow blocks in
      let _ = data_flow graph ntable in
         interference_graph blocks
end

(*
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
