(*
 * Common-subexpression elimination.
 *
 * Ling Li, Xin Yu
 * Mar 14, 2001, Caltech
 *
 * This implementation is pretty simple:
 * subexpressions are the same only if they
 * compute exactly the same expression.
 *
 * Note: all variables are different.
 * IR must be functional.
 *)

open Symbol
open Fc_ir_type
open Fc_ir_exn_type
open Fc_ir_exn

(************************************************************************
 * ENVIRONMENTS
 ************************************************************************)

(*
 * Define the type of variable environments.
 *)
type cse =
 | CseAtom of atom
 | CseUnop of unop * cse
 | CseBinop of binop * cse * cse
 | CseAddrOfSubscript of cse * cse
 | CseAddrOfProject of cse * var
 | CseClosure of var * cse list

(*
 * Comparison functions
 *)
let compare_atoms a1 a2 =
   match a1, a2 with
    | AtomChar c1, AtomChar c2 ->   Pervasives.compare c1 c2
    | AtomInt i1, AtomInt i2 ->     Pervasives.compare i1 i2
    | AtomFloat x1, AtomFloat x2 -> Pervasives.compare x1 x2
    | AtomNil ty1, AtomNil ty2 ->   Pervasives.compare ty1 ty2 (*???*)
    | AtomVar v1, AtomVar v2 ->     Symbol.compare v1 v2
   (* The order of uncomparable pairs --- use alphabet order *)
    | AtomChar _, _ -> -1
    | _, AtomChar _ ->  1
    | AtomFloat _, _ -> -1
    | _, AtomFloat _ ->  1
    | AtomInt _, _ -> -1
    | _, AtomInt _ ->  1
    | AtomNil _, _ -> -1
    | _, AtomNil _ ->  1
(*  | AtomVar _, _ -> -1  (* already included in other cases *)
    | _, AtomVar _ ->  1   *)

(*
 * Compare pairs of expressions with order
 *)
let rec compare_exp_pairs_order e11 e12 e21 e22 =
   let c = compare_exps e11 e21 in
      if c = 0 then
         compare_exps e12 e22
      else
         c

(*
 * Compare pairs of expressions without order
 * BE CAREFUL TO MAKE THE SET OF EXP-PAIR ORDERABLE.
 * That is, pair1 < pair2 & pair2 < pair3 ==> pair1 < pair3 ...
 *)
and compare_exp_pairs_noorder e11 e12 e21 e22 =
   let c = compare_exp_pairs_order e11 e12 e22 e21 in
      if c = 0 then
         c
      else
         compare_exp_pairs_order e11 e12 e21 e22

and compare_exps e1 e2 =
   match e1, e2 with
    | CseAtom a1, CseAtom a2 ->
         compare_atoms a1 a2
    | CseUnop (op1, e1), CseUnop (op2, e2) ->
         let i = Pervasives.compare op1 op2 in
            if i = 0 then
               compare_exps e1 e2
            else
               i
    | CseBinop (op1, e11, e12), CseBinop (op2, e21, e22) ->
         let i = Pervasives.compare op1 op2 in
            if i = 0 then
               match op1 with
                | PlusOp | MulOp | BAndOp | BOrOp | XorOp ->
                     compare_exp_pairs_noorder e11 e12 e21 e22
                | MinusOp | DivOp | RemOp | LslOp | LsrOp | AsrOp ->
                     compare_exp_pairs_order e11 e12 e21 e22
            else
               i
    | CseAddrOfSubscript (e11, e12), CseAddrOfSubscript (e21, e22) ->
         let i = compare_exps e11 e21 in
            if i = 0 then
               compare_exps e12 e22
            else
               i
    | CseAddrOfProject (e1, l1), CseAddrOfProject (e2, l2) ->
         let i = compare_exps e1 e2 in
            if i = 0 then
               Symbol.compare l1 l2
            else
               i
    | CseClosure (v1, args1), CseClosure (v2, args2) ->
         let i = Symbol.compare v1 v2 in
            if i = 0 then
               compare_exp_lists args1 args2
            else
               i
   (* The order of uncomparable pairs --- use alphabet order *)
    | CseAddrOfProject _, _ -> -1
    | _, CseAddrOfProject _ ->  1
    | CseAddrOfSubscript _, _ -> -1
    | _, CseAddrOfSubscript _ ->  1
    | CseAtom _, _ -> -1
    | _, CseAtom _ ->  1
    | CseBinop _, _ -> -1
    | _, CseBinop _ ->  1
    | CseClosure _, _ -> -1
    | _, CseClosure _ ->  1
(*  | CseUnop _, _ -> -1  (* already included in other cases *)
    | _, CseUnop _ ->  1   *)

and compare_exp_lists args1 args2 =
   match args1, args2 with
    | [], [] ->  0
    | [], _  -> -1
    | _ , [] ->  1
    | e1 :: args1, e2 :: args2 ->
         let i = compare_exps e1 e2 in
            if i = 0 then
               compare_exp_lists args1 args2
            else
               i

(*
 * A table with expression comparison.
 *)
module CseCompare =
struct
   type elt = cse
   let compare = compare_exps
end

module CseTable = Red_black_table.MakeTable (CseCompare)

(*
 * Variable environment now contains expressions
 * We can look up: var -> exp, or exp -> var
 *)
type venv = {
   cse_vars : cse SymbolTable.t;
   cse_exps : var CseTable.t
}

let venv_empty = {
   cse_vars = SymbolTable.create ();
   cse_exps = CseTable.create ()
}

(*
 * Lookup a variable.
 *)
let venv_lookup_var venv pos v =
   try
      SymbolTable.find venv.cse_vars v
   with
      Not_found -> raise (IRException (pos, UnboundVar v))

let rec expand_exp venv pos e =
   match e with
    | CseAtom (AtomVar v) ->
         venv_lookup_var venv pos v
    | CseAtom _ ->
         e
    | CseUnop (op, e) ->
         CseUnop (op, expand_exp venv pos e)
    | CseBinop (op, e1, e2) ->
         CseBinop (op, expand_exp venv pos e1, expand_exp venv pos e2)
    | CseAddrOfSubscript (e1, e2) ->
         CseAddrOfSubscript (expand_exp venv pos e1, expand_exp venv pos e2)
    | CseAddrOfProject (e, l) ->
         CseAddrOfProject (expand_exp venv pos e, l)
    | CseClosure (v, args) ->
         CseClosure (v, List.map (expand_exp venv pos) args)

(*
 * Lookup an expression (with expansion).
 *)
let venv_lookup_exp venv pos exp =
   let exp = expand_exp venv pos exp in
   let a =
      try
         Some (AtomVar (CseTable.find venv.cse_exps exp))
      with
         Not_found -> None
   in
      a, exp

(*
 * Add a abstract var to the venv.
 * This is a variable that can't be expanded to
 * an expression.
 *)
let venv_add_abs_var venv v = { venv with
      cse_vars = SymbolTable.add venv.cse_vars v (CseAtom (AtomVar v)) }

(*
 * Add a variable together with an expression.
 *)
let venv_add_var { cse_vars = cse_vars; cse_exps = cse_exps } v exp =
   {  cse_vars = SymbolTable.add cse_vars v exp;
      cse_exps = CseTable.add cse_exps exp v
   }

(************************************************************************
 * CSE
 ************************************************************************)

(*
 * Common-subexpression elimination.
 * We use a simple algorithm: no memory optimization,
 * only identical code.
 *)
let rec cse_expr venv e =
   let pos = string_pos "cse_expr" (atom_pos e) in
      match e with
         LetAtom (v, ty, a, e) ->
            cse_atom_expr venv pos v ty a e
       | LetUnop (v, ty, op, a, e) ->
            cse_unop_expr venv pos v ty op a e
       | LetBinop (v, ty, op, a1, a2, e) ->
            cse_binop_expr venv pos v ty op a1 a2 e
       | TailCall _
       | IntCall _ ->
            e
       | LetExtCall (v, f, ty, args, e) ->
            let venv = venv_add_abs_var venv v in
               LetExtCall (v, f, ty, args, cse_expr venv e)
       | IfThenElse (op, a1, a2, e1, e2) ->
            IfThenElse (op, a1, a2, cse_expr venv e1, cse_expr venv e2)
       | SetSubscript (a1, ty, a2, a3, e) ->
            SetSubscript (a1, ty, a2, a3, cse_expr venv e)
       | LetSubscript (v, ty, a1, a2, e) ->
            let venv = venv_add_abs_var venv v in
               LetSubscript (v, ty, a1, a2, cse_expr venv e)
       | LetAddrOfSubscript (v, ty, a1, a2, e) ->
            cse_addr_of_subscript_expr venv pos v ty a1 a2 e
       | SetProject (a1, l, ty, a2, e) ->
            SetProject (a1, l, ty, a2, cse_expr venv e)
       | LetProject (v, ty, a, l, e) ->
            let venv = venv_add_abs_var venv v in
               LetProject (v, ty, a, l, cse_expr venv e)
       | LetAddrOfProject (v, ty, a, l, e) ->
            cse_addr_of_project_expr venv pos v ty a l e
       | Memcpy (a1, ty, a2, e) ->
            Memcpy (a1, ty, a2, cse_expr venv e)
       | LetAlloc (v, ty, e) ->
            let venv = venv_add_abs_var venv v in
               LetAlloc (v, ty, cse_expr venv e)
       | LetClosure (v, ty, f, args, e) ->
            cse_closure_expr venv pos v ty f args e
       | SetVar _
       | LetTypes _
       | LetFuns _
       | LetFuns2 _
       | LetString _
       | LetAddrOfVar _
       | LetCopy _
       | LetApply _
       | Return _ ->
            let pos = string_pos "cse_expr" (atom_pos e) in
               raise (IRException (pos, IRLevel 2))

(*
 * Inline an atom.
 *)
and cse_atom_expr venv pos v ty a e =
   let pos = string_pos "cse_atom_expr" pos in
      match venv_lookup_exp venv pos (CseAtom a) with
       | Some a, exp ->
            (* replace the definition of v, if there is any *)
            let venv = venv_add_var venv v exp in
               LetAtom (v, ty, a, cse_expr venv e)
       | None, _ ->
            raise (IRException (pos, InternalError "cse_atom_expr"))

(*
 * Unary operation.
 *)
and cse_unop_expr venv pos v ty op a e =
   let pos = string_pos "cse_unop_expr" pos in
      match venv_lookup_exp venv pos (CseUnop (op, CseAtom a)) with
       | Some a, exp ->
            let venv = venv_add_var venv v exp in
               LetAtom (v, ty, a, cse_expr venv e)
       | None, exp ->
            let venv = venv_add_var venv v exp in
               LetUnop (v, ty, op, a, cse_expr venv e)

(*
 * Binary operation.
 *)
and cse_binop_expr venv pos v ty op a1 a2 e =
   let pos = string_pos "cse_binop_expr" pos in
   let exp = CseBinop (op, CseAtom a1, CseAtom a2) in
      match venv_lookup_exp venv pos exp with
       | Some a, exp ->
            let venv = venv_add_var venv v exp in
               LetAtom (v, ty, a, cse_expr venv e)
       | None, exp ->
            let venv = venv_add_var venv v exp in
               LetBinop (v, ty, op, a1, a2, cse_expr venv e)

(*
 * Address of subscript.
 *)
and cse_addr_of_subscript_expr venv pos v ty a1 a2 e =
   let pos = string_pos "cse_addr_of_subscript_expr" pos in
   let exp = CseAddrOfSubscript (CseAtom a1, CseAtom a2) in
      match venv_lookup_exp venv pos exp with
       | Some a, exp ->
            let venv = venv_add_var venv v exp in
               LetAtom (v, ty, a, cse_expr venv e)
       | None, exp ->
            let venv = venv_add_var venv v exp in
               LetAddrOfSubscript (v, ty, a1, a2, cse_expr venv e)

(*
 * Address of struct projection.
 *)
and cse_addr_of_project_expr venv pos v ty a l e =
   let pos = string_pos "cse_addr_of_project_expr" pos in
   let exp = CseAddrOfProject (CseAtom a, l) in
      match venv_lookup_exp venv pos exp with
       | Some a, exp ->
            let venv = venv_add_var venv v exp in
               LetAtom (v, ty, a, cse_expr venv e)
       | None, exp ->
            let venv = venv_add_var venv v exp in
               LetAddrOfProject (v, ty, a, l, cse_expr venv e)

(*
 * Closure.
 *)
and cse_closure_expr venv pos v ty f args e =
   let pos = string_pos "cse_closure_expr" pos in
   let exp = CseClosure (f, List.map (fun a -> CseAtom a) args) in
      match venv_lookup_exp venv pos exp with
       | Some a, exp ->
            let venv = venv_add_var venv v exp in
               LetAtom (v, ty, a, cse_expr venv e)
       | None, exp ->
            let venv = venv_add_var venv v exp in
               LetClosure (v, ty, f, args, cse_expr venv e)

(************************************************************************
 * GLOBAL FUNCTION
 ************************************************************************)

(*
 * Inline a prog.
 *)
let cse_prog prog =
   let { prog_funs = funs; prog_strings = strings } = prog in

   (* Add strings & functions to variable environment *)
   let venv =
      List.fold_left (fun venv (v, _) ->
            venv_add_abs_var venv v) venv_empty strings
   in
   let venv =
      List.fold_left (fun venv (f, _, _, _, _) ->
            venv_add_abs_var venv f) venv funs
   in

   (* Apply CSE to all functions *)
   let apply_CSE (f, gflag, ty, args, e) =
       let venv = List.fold_left venv_add_abs_var venv args in
          f, gflag, ty, args, cse_expr venv e
   in
      { prog with prog_funs = List.map apply_CSE funs }
