(*
 * Function inlining and constant folding.
 *
 * Ling Li, Xin Yu
 * Mar 14, 2001, Caltech
 *)

open Debug
open Symbol
open Fc_ir_type
open Fc_ir_exn_type
open Fc_ir_exn
open Fc_ir_env
open Fc_ir_standardize

(************************************************************************
 * ENVIRONMENT
 ************************************************************************)

(*
 * Default exception.
 *)
let raise_exn pos v =
   IRException (pos, v)

let tenv_lookup = tenv_lookup raise_exn
let tenv_expand = tenv_expand raise_exn

(*
 * Variable environment.
 * A variable can be an atom,
 * or it can have a type (with no value),
 * or it can be a function,
 * or it can be a closure.
 *)
type entry =
   VarAtom of atom
 | VarType of ty
 | VarFun of ty * var list * exp
 | VarClosure of ty * var * atom list

type venv = entry SymbolTable.t

let venv_empty = SymbolTable.create ()

let venv_add venv v ent = SymbolTable.add venv v ent

(*
 * Add a variable that can't be inlined.
 *)
let venv_add_atom pos venv v = function
 | AtomVar v' when v' = v ->
      venv
 | a ->
      venv_add venv v (VarAtom a)

let venv_add_type venv v ty =
   venv_add venv v (VarType ty)

let venv_add_fun venv v ty vars e =
   venv_add venv v (VarFun (ty, vars, e))

let venv_add_closure venv v ty f args =
   venv_add venv v (VarClosure (ty, f, args))

(*
 * Lookup a variable (and its type).
 *)
let venv_lookup venv pos v =
   try
      SymbolTable.find venv v
   with
      Not_found -> raise (IRException (pos, UnboundVar v))

let venv_lookup_opt venv v =
   try
      Some (SymbolTable.find venv v)
   with
      Not_found -> None

let rec venv_lookup_type venv pos v =
   match venv_lookup venv pos v with
    | VarAtom a -> venv_lookup_atom venv pos a
    | VarType ty -> ty
    | VarFun (ty, _, _) -> ty
    | VarClosure (ty, _, _) -> ty

and venv_lookup_atom venv pos a =
   match a with
    | AtomChar _ -> TyChar
    | AtomInt _ -> TyInt
    | AtomFloat _ -> TyFloat
    | AtomNil ty -> ty
    | AtomVar v -> venv_lookup_type venv pos v

(************************************************************************
 * UTILITIES
 ************************************************************************)

let is_power_of_two i =
   let rec search k =
      if k > i || k = (1 lsl 30) then
         false
      else
         k = i || search (k lsl 1)
   in
      search 1

let log2 i =
   let rec search j =
      let k = 1 lsl j in
         if k = i then
            j
         else
            search (succ j)
   in
      search 0

let dest_fun_type tenv pos ty =
   match tenv_expand tenv pos ty with
      TyFun (ty_vars, ty_res) -> ty_vars, ty_res
    | ty -> raise (IRException (pos, NotAFunction ty))

(************************************************************************
 * INLINING
 ************************************************************************)

(*
 * Inline an atom.
 * If the atom is a var,
 * you should look up its value from
 * the variable environment.
 *)
let inline_atom venv pos a =
   match a with
    | AtomChar _
    | AtomInt _
    | AtomFloat _
    | AtomNil _ ->
         a
    | AtomVar v ->
         match venv_lookup venv pos v with
          | VarAtom a -> a
          | VarType _
          | VarFun _
          | VarClosure _ ->
               a

(*
 * Function inlining,
 * Constant folding.
 *)
let rec inline_expr tenv venv dsym e =
   let pos = string_pos "inline_expr" (atom_pos e) in
      match e with
         LetAtom (v, ty, a, e) ->
            inline_atom_expr tenv venv dsym pos v ty a e
       | LetUnop (v, ty, op, a, e) ->
            inline_unop_expr tenv venv dsym pos v ty op a e
       | LetBinop (v, ty, op, a1, a2, e) ->
            inline_binop_expr tenv venv dsym pos v ty op a1 a2 e
       | TailCall (f, args) ->
            inline_tailcall_expr tenv venv dsym pos f args
       | LetExtCall (v, f, ty, args, e) ->
            inline_extcall_expr tenv venv dsym pos v f ty args e
       | IntCall (ty, f, args) ->
            inline_intcall_expr tenv venv dsym pos ty f args
       | IfThenElse (op, a1, a2, e1, e2) ->
            inline_if_expr tenv venv dsym pos op a1 a2 e1 e2
       | SetSubscript (a1, ty, a2, a3, e) ->
            let a1 = inline_atom venv pos a1 in
            let a2 = inline_atom venv pos a2 in
            let a3 = inline_atom venv pos a3 in
               SetSubscript (a1, ty, a2, a3, inline_expr tenv venv dsym e)
       | LetSubscript (v, ty, a1, a2, e) ->
            let a1 = inline_atom venv pos a1 in
            let a2 = inline_atom venv pos a2 in
            let venv = venv_add_type venv v ty in
               LetSubscript (v, ty, a1, a2, inline_expr tenv venv dsym e)
       | LetAddrOfSubscript (v, ty, a1, a2, e) ->
            let a1 = inline_atom venv pos a1 in
            let a2 = inline_atom venv pos a2 in
            let venv = venv_add_type venv v ty in
               LetAddrOfSubscript (v, ty, a1, a2, inline_expr tenv venv dsym e)
       | SetProject (a1, l, ty, a2, e) ->
            let a1 = inline_atom venv pos a1 in
            let a2 = inline_atom venv pos a2 in
               SetProject (a1, l, ty, a2, inline_expr tenv venv dsym e)
       | LetProject (v, ty, a, l, e) ->
            let a = inline_atom venv pos a in
            let venv = venv_add_type venv v ty in
               LetProject (v, ty, a, l, inline_expr tenv venv dsym e)
       | LetAddrOfProject (v, ty, a, l, e) ->
            let a = inline_atom venv pos a in
            let venv = venv_add_type venv v ty in
               LetAddrOfProject (v, ty, a, l, inline_expr tenv venv dsym e)
       | Memcpy (a1, ty, a2, e) ->
            let a1 = inline_atom venv pos a1 in
            let a2 = inline_atom venv pos a2 in
               Memcpy (a1, ty, a2, inline_expr tenv venv dsym e)
       | LetAlloc (v, ty, e) ->
            let venv = venv_add_type venv v (TyArray (ty, None)) in
               LetAlloc (v, ty, inline_expr tenv venv dsym e)
       | LetClosure (v, ty, f, args, e) ->
            inline_closure_expr tenv venv dsym pos v ty f args e
       | SetVar _
       | LetTypes _
       | LetFuns _
       | LetFuns2 _
       | LetString _
       | LetAddrOfVar _
       | LetCopy _
       | LetApply _
       | Return _ ->
            let pos = string_pos "inline_expr" (atom_pos e) in
               raise (IRException (pos, IRLevel 2))

(*
 * Inline an atom.
 * This always gets inlined because it has no effect.
 *)
and inline_atom_expr tenv venv dsym pos v ty a e =
   let pos = string_pos "inline_atom_expr" pos in
   let a = inline_atom venv pos a in
   let venv = venv_add_atom pos venv v a in
      inline_expr tenv venv dsym e

(*
 * Inline a unop.
 * Inline only if the value is a constant.
 *)
and inline_unop_expr tenv venv dsym pos v ty op a e = (*??? need more work*)
   let pos = string_pos "inline_unop_expr" pos in
   let a = inline_atom venv pos a in
      match op with
       | UMinusOp -> (
            match a with
             | AtomInt i ->
                  let venv = venv_add_atom pos venv v (AtomInt (-i)) in
                     inline_expr tenv venv dsym e
             | AtomFloat x ->
                  let venv = venv_add_atom pos venv v (AtomFloat (-. x)) in
                     inline_expr tenv venv dsym e
             | _ ->
                  let venv = venv_add_type venv v ty in
                     LetUnop (v, ty, op, a, inline_expr tenv venv dsym e))
       | UToChar -> (
            match a with
             | AtomChar _ ->
                  let venv = venv_add_atom pos venv v a in
                     inline_expr tenv venv dsym e
             | AtomInt i ->
                  let venv = venv_add_atom pos venv v (AtomChar (Char.chr (i land 255))) in
                     inline_expr tenv venv dsym e
             | AtomFloat x ->
                  let venv = venv_add_atom pos venv v (AtomChar (Char.chr ((int_of_float x) land 255))) in
                     inline_expr tenv venv dsym e
             | AtomNil _ ->
                  let venv = venv_add_atom pos venv v (AtomChar '\000') in
                     inline_expr tenv venv dsym e
             | AtomVar v' ->
                  let ty = venv_lookup_type venv pos v' in
                     match tenv_expand tenv pos ty with
                      | TyChar ->
                           let venv = venv_add_atom pos venv v (AtomVar v') in
                              inline_expr tenv venv dsym e
                      | _ ->
                           let venv = venv_add_type venv v ty in
                              LetUnop (v, ty, op, a, inline_expr tenv venv dsym e))
       | UToInt -> (
            match a with
             | AtomChar c ->
                  let venv = venv_add_atom pos venv v (AtomInt (Char.code c)) in
                     inline_expr tenv venv dsym e
             | AtomInt _ ->
                  let venv = venv_add_atom pos venv v a in
                     inline_expr tenv venv dsym e
             | AtomFloat x ->
                  let venv = venv_add_atom pos venv v (AtomInt (int_of_float x)) in
                     inline_expr tenv venv dsym e
             | AtomNil _ ->
                  let venv = venv_add_atom pos venv v (AtomInt 0) in
                     inline_expr tenv venv dsym e
             | AtomVar v' ->
                  let ty = venv_lookup_type venv pos v' in
                     match tenv_expand tenv pos ty with
                      | TyInt ->
                           let venv = venv_add_atom pos venv v a in
                              inline_expr tenv venv dsym e
                      | _ ->
                           let venv = venv_add_type venv v ty in
                              LetUnop (v, ty, op, a, inline_expr tenv venv dsym e))
       | UToFloat -> (
            match a with
             | AtomChar c ->
                  let venv = venv_add_atom pos venv v (AtomFloat (float_of_int (Char.code c))) in
                     inline_expr tenv venv dsym e
             | AtomInt i ->
                  let venv = venv_add_atom pos venv v (AtomFloat (float_of_int i)) in
                     inline_expr tenv venv dsym e
             | AtomFloat x ->
                  let venv = venv_add_atom pos venv v a in
                     inline_expr tenv venv dsym e
             | AtomNil _ ->
                  let venv = venv_add_atom pos venv v (AtomFloat 0.0) in
                     inline_expr tenv venv dsym e
             | AtomVar v' ->
                  let ty = venv_lookup_type venv pos v' in
                     match tenv_expand tenv pos ty with
                      | TyFloat ->
                           let venv = venv_add_atom pos venv v a in
                              inline_expr tenv venv dsym e
                      | _ ->
                           let venv = venv_add_type venv v ty in
                              LetUnop (v, ty, op, a, inline_expr tenv venv dsym e))
       | UToFun | UToFunPointer ->
            let venv = venv_add_type venv v ty in
               LetUnop (v, ty, op, a, inline_expr tenv venv dsym e)
       | UNotOp -> raise (IRException (pos, IRLevel 2))

(*
 * Binop.
 *)
and fnop = (+.)
and inline_binop tenv venv dsym pos v ty op a1 a2 op_int op_float e =
   match a1, a2 with
    | AtomInt i1, AtomInt i2 ->
         let venv = venv_add_atom pos venv v (AtomInt (op_int i1 i2)) in
            inline_expr tenv venv dsym e
    | AtomFloat x1, AtomFloat x2 ->
         let venv = venv_add_atom pos venv v (AtomFloat (op_float x1 x2)) in
            inline_expr tenv venv dsym e
    | _ ->
         let venv = venv_add_type venv v ty in
            LetBinop (v, ty, op, a1, a2, inline_expr tenv venv dsym e)

and inline_binop_power2 op a1 a2 alt_op =
   match a2 with
    | AtomInt i2 ->
         if is_power_of_two i2 then
            alt_op, a1, (AtomInt (log2 i2))
         else
            op, a1, a2
    | _ ->  op, a1, a2

and inline_binop_simplify op a1 a2 =
   let inline_remove_default a2 def_a2 def_result =
      match a2 with
       | AtomInt i2 ->
            if i2 = def_a2 then
               Some def_result
            else
               None
       | _ -> None
   in

   let inline_binop_remove0 a1 a2 = inline_remove_default a2 0 a1 in
   let inline_binop_remove00 a1 a2 =
      match inline_binop_remove0 a1 a2 with
       | None ->
            inline_binop_remove0 a2 a1
       | r -> r
   in
   let inline_binop_mul0 a1 a2 =
      match inline_remove_default a1 0 (AtomInt 0) with
       | None ->
            inline_remove_default a2 0 (AtomInt 0)
       | r -> r
   in
      match op with
       | PlusOp
       | BOrOp
       | XorOp   -> inline_binop_remove00 a1 a2
       | MinusOp
       | LslOp
       | LsrOp
       | AsrOp   -> inline_binop_remove0  a1 a2
       | MulOp
       | BAndOp  -> inline_binop_mul0     a1 a2
       | DivOp
       | RemOp   -> None

and inline_binop_expr tenv venv dsym pos v ty op a1 a2 e =
   let pos = string_pos "inline_binop_expr" pos in
   let a1 = inline_atom venv pos a1 in
   let a2 = inline_atom venv pos a2 in
   let op, a1, a2 =
      match op with
       | MulOp ->
            let op, a1, a2 = inline_binop_power2 MulOp a1 a2 LslOp in
               if op = MulOp then
                  inline_binop_power2 MulOp a2 a1 LslOp
               else op, a1, a2
       | DivOp -> inline_binop_power2 DivOp a1 a2 AsrOp
       | _ -> op, a1, a2
   in
   match inline_binop_simplify op a1 a2 with
    | Some a ->
         let venv = venv_add_atom pos venv v a in
            inline_expr tenv venv dsym e
    | None ->
      match op with
       | PlusOp  -> inline_binop tenv venv dsym pos v ty op a1 a2 (+)    (+.) e
       | MinusOp -> inline_binop tenv venv dsym pos v ty op a1 a2 (-)    (-.) e
       | MulOp   -> inline_binop tenv venv dsym pos v ty op a1 a2 ( * )( *. ) e
       | DivOp   -> inline_binop tenv venv dsym pos v ty op a1 a2 (/)    (/.) e
       | RemOp   -> inline_binop tenv venv dsym pos v ty op a1 a2 (mod)  fnop e
       | BAndOp  -> inline_binop tenv venv dsym pos v ty op a1 a2 (land) fnop e
       | BOrOp   -> inline_binop tenv venv dsym pos v ty op a1 a2 (lor)  fnop e
       | LslOp   -> inline_binop tenv venv dsym pos v ty op a1 a2 (lsl)  fnop e
       | LsrOp   -> inline_binop tenv venv dsym pos v ty op a1 a2 (lsr)  fnop e
       | AsrOp   -> inline_binop tenv venv dsym pos v ty op a1 a2 (asr)  fnop e
       | XorOp   -> inline_binop tenv venv dsym pos v ty op a1 a2 (lxor) fnop e

(*
 * Function call.
 * Inline the function if possible.
 *)
and inline_tailcall_expr tenv venv dsym pos f args =
   let pos = string_pos "inline_tailcall_expr" pos in
   let args = List.map (inline_atom venv pos) args in
      match venv_lookup venv pos f with
       | VarFun (ty, vars, e) ->
            (* Inline this function *)
            let venv = venv_add_type venv f ty in
               List.fold_left2 (fun e v a ->
                     let ty = venv_lookup_atom venv pos a in
                        LetAtom (v, ty, a, e)) e vars args
       | VarClosure (ty, f', args') ->
            TailCall (f', args' @ args)
       | VarAtom (AtomVar f) ->
            TailCall (f, args)
       | VarAtom _
       | VarType _ ->
            TailCall (f, args)

(*
 * Closure.
 * Save the closure in the environment,
 *)
and inline_closure_expr tenv venv dsym pos v ty f args e =
   let pos = string_pos "inline_closure_expr" pos in
   let args = List.map (inline_atom venv pos) args in
      match venv_lookup venv pos f with
       | VarClosure (_, f', args') ->
            inline_closure_expr tenv venv dsym pos v ty f' (args' @ args) e
       | VarFun _
       | VarAtom _
       | VarType _ ->
            let venv = venv_add_closure venv v ty f args in
               LetClosure (v, ty, f, args, inline_expr tenv venv dsym e)

(*
 * External call.
 * Unfold it if the call is to an internal function.
 *)
and inline_extcall_expr tenv venv dsym pos v f ty args e =
   let pos = string_pos "inline_extcall_expr" pos in
   let args = List.map (inline_atom venv pos) args in
      match e, venv_lookup_opt venv f with
       | TailCall (cont, _), Some (VarFun (_, _, IntCall (_, f', _))) ->
            TailCall (f', AtomVar dsym :: AtomVar cont :: args)
       | _ ->
            let _, ty_res = dest_fun_type tenv pos ty in
            let venv = venv_add_type venv v ty_res in
               LetExtCall (v, f, ty, args, inline_expr tenv venv dsym e)

(*
 * Internal call is also like a tailcall.
 *)
and inline_intcall_expr tenv venv dsym pos ty f args =
   let pos = string_pos "inline_intcall_expr" pos in
   let args = List.map (inline_atom venv pos) args in
      IntCall (ty, f, args)

(*
 * Conditional.
 * Remove the conditional if the test can be determined.
 *)
and inline_if tenv venv dsym pos op op_int op_float a1 a2 e1 e2 =
   let pos = string_pos "inline_if" pos in
   let e1 = inline_expr tenv venv dsym e1 in
   let e2 = inline_expr tenv venv dsym e2 in
      match a1, a2 with
       | AtomInt i1, AtomInt i2 ->
            if op_int i1 i2 then e1
            else e2
       | AtomFloat x1, AtomFloat x2 ->
            if op_float x1 x2 then e1
            else e2
       | _ ->
            IfThenElse (op, a1, a2, e1, e2)

and inline_if_expr tenv venv dsym pos op a1 a2 e1 e2 =
   let pos = string_pos "inline_if_expr" pos in
   let a1 = inline_atom venv pos a1 in
   let a2 = inline_atom venv pos a2 in
      match op with
       | EqOp  -> inline_if tenv venv dsym pos op (=)  (=)  a1 a2 e1 e2
       | NEqOp -> inline_if tenv venv dsym pos op (<>) (<>) a1 a2 e1 e2
       | LeOp  -> inline_if tenv venv dsym pos op (<=) (<=) a1 a2 e1 e2
       | LtOp  -> inline_if tenv venv dsym pos op (<)  (<)  a1 a2 e1 e2
       | GtOp  -> inline_if tenv venv dsym pos op (>)  (>)  a1 a2 e1 e2
       | GeOp  -> inline_if tenv venv dsym pos op (>=) (>=) a1 a2 e1 e2

(************************************************************************
 * GLOBAL FUNCTIONS
 ************************************************************************)

(*
 * How many instruction is "small"?
 *)
let inline_size = 50

(*
 * Compute the 'weighted' sum of an expression
 *)
let rec expr_size e = match e with
 | LetAtom (_, _, _, e) ->
      expr_size e
 | LetUnop (_, _, _, _,  e) ->
      1 + (expr_size e)
 | LetBinop (_, _, _, _, _, e) ->
      2 + (expr_size e)
 | TailCall _ ->
      0
 | IntCall _
 | LetExtCall _ ->
      8
 | IfThenElse (_, _, _, e1, e2) ->
      max (expr_size e1) (expr_size e2)
 | SetSubscript (_, _, _, _, e) ->
      4 + (expr_size e)
 | LetSubscript (_, _, _, _, e) ->
      2 + (expr_size e)
 | LetAddrOfSubscript (_, _, _, _, e) ->
      3 + (expr_size e)
 | SetProject (_, _, _, _, e) ->
      4 + (expr_size e)
 | LetProject (_, _, _, _, e) ->
      2 + (expr_size e)
 | LetAddrOfProject (_, _, _, _, e) ->
      3 + (expr_size e)
 | Memcpy (_, _, _, e) ->
      4 + (expr_size e)
 | LetAlloc (_, _, e) ->
      expr_size e
 | LetClosure (_, _, _, _, e) ->
      3 + (expr_size e)
 | SetVar _
 | LetTypes _
 | LetFuns _
 | LetFuns2 _
 | LetString _
 | LetAddrOfVar _
 | LetCopy _
 | LetApply _
 | Return _ ->
      let pos = string_pos "expr_size" (atom_pos e) in
         raise (IRException (pos, IRLevel 2))

(*
 * Inline fun (add it as VarFun) if it is small.
 *)
let inline_fun inline venv (f, _, ty, vars, e) =
   if inline && expr_size e < inline_size then
      venv_add_fun venv f ty vars e
   else
      venv_add_type venv f ty

(*
 * Inline a prog.
 *)
let inline_prog fun_flag
    { prog_data_type = data_type;
      prog_init_sym = init_sym;
      prog_types = types;
      prog_strings = strings;
      prog_funs = funs
    } =
   let dsym = new_symbol_string "data_seg" in

   (* Add types to the environment *)
   let tenv =
      List.fold_left (fun tenv (v, ty) ->
            tenv_add tenv v ty) empty_tenv types
   in

   (* Add strings to the environment *)
   let venv =
      List.fold_left (fun venv (v, _) ->
            venv_add_type venv v (TyArray (TyChar, None))) venv_empty strings
   in

   (* Add functions to the environment *)
   let venv = List.fold_left (inline_fun fun_flag) venv funs in

   (* Inline all the function bodies *)
   let funs =
      List.map (fun (f, gflag, ty, vars, e) ->
            let pos = string_pos "inline_prog" (atom_pos e) in
            let ty_vars, _ = dest_fun_type tenv pos ty in
            let venv = List.fold_left2 venv_add_type venv vars ty_vars in
            let e = inline_expr tenv venv dsym e in
               f, gflag, ty, vars, e) funs
   in

   (* Reconstruct the program *)
   let prog =
      { prog_data_type = data_type;
        prog_init_sym = init_sym;
        prog_types = types;
        prog_strings = strings;
        prog_funs = funs
      }
   in
      standardize_prog prog
