(*
 * Closure conversion.
 *
 * Ling Li, Xin Yu
 * Mar 13, 2001, Caltech
 *)

open Symbol
open Fc_ast_util
open Fc_ir_type
open Fc_ir_exn_type
open Fc_ir_exn
open Fc_ir_env
open Fc_ir_standardize

let debug_closure = ref false

(************************************************************************
 * ENVIRONMENTS
 ************************************************************************)

(*
 * Default exception.
 *)
let raise_exn pos v =
   IRException (pos, v)

let tenv_expand = tenv_expand raise_exn
let venv_lookup = venv_lookup raise_exn

(*
 * Equations to calculate free variables and
 * Table to keep funtion equations
 *)
type vset = SymbolSet.t

type free_eqn =
 | EqnNil
 | EqnAddFun of free_eqn * var
 | EqnAdd of free_eqn * vset
 | EqnSub of free_eqn * vset
 | EqnMerge of free_eqn * free_eqn

(*
 * Utilities
 *)
let set_empty           = SymbolSet.empty
let set_mem set v       = SymbolSet.mem v set
let set_add set v       = SymbolSet.add v set
let set_remove set v    = SymbolSet.remove v set
let set_union set set'  = SymbolSet.union set set'
let set_diff  set set'  = SymbolSet.fold set_remove set set'
let set_equal set set'  = SymbolSet.equal set set'
let set_to_list set     = SymbolSet.to_list set
let set_of_list l       = SymbolSet.of_list l
let set_add_list set vl = SymbolSet.add_list set vl
let set_init v          = set_add set_empty v

let tab_empty ()        = SymbolTable.create ()
let tab_mem tab v       = SymbolTable.mem tab v
let tab_add_key tab v   = SymbolTable.add_empty tab v
let tab_add tab v data  = SymbolTable.add tab v data
let tab_fold f env tab  = SymbolTable.fold f env tab
let tab_iter f tab      = SymbolTable.iter f tab
let tab_find tab v      = SymbolTable.find tab v

(* For external call: no free variables *)
let vtab_lookup tab v =
   try tab_find tab v with
      Not_found -> []

let eqn_empty = EqnNil

let eqn_add ftab eqn v =
   if tab_mem ftab v then
      EqnAddFun (eqn, v)
   else
      match eqn with
       | EqnAdd (eqn, add) ->
            EqnAdd (eqn, set_add add v)
       | _ ->
            EqnAdd (eqn, set_init v)

let eqn_sub eqn v =
   match eqn with
    | EqnSub (eqn, sub) ->
         EqnSub (eqn, set_add sub v)
    | EqnNil ->
         raise (Invalid_argument "fc_ir_closure.eqn_sub")
    | _ ->
         EqnSub (eqn, set_init v)

let eqn_sub_list eqn vl =
   match eqn with
    | EqnSub (eqn, sub) ->
         EqnSub (eqn, set_add_list sub vl)
    | EqnNil ->
         raise (Invalid_argument "fc_ir_closure.eqn_sub_list")
    | _ ->
         EqnSub (eqn, set_of_list vl)

let eqn_merge eqn1 eqn2 =
   EqnMerge (eqn1, eqn2)

(* add an atom to equation *)
let eqn_atom ftab eqn = function
 | AtomChar _ | AtomInt _ | AtomFloat _ | AtomNil _ ->
      eqn
 | AtomVar v ->
      eqn_add ftab eqn v

let dest_fun_type tenv pos ty =
   match tenv_expand tenv pos ty with
      TyFun (ty_vars, ty_res) -> ty_vars, ty_res
    | ty -> raise (IRException (pos, NotAFunction ty))

let make_args = List.map (fun v -> AtomVar v)

(*
 * Collect free variables.
 *)
let rec free_expr ftab dsym = function
 | LetTypes (types, e) ->
      free_types_expr ftab dsym types e
 | LetFuns2 (funs, e) ->
      free_funs_expr ftab dsym funs e
 | LetAtom (v, _, a, e) ->
      free_atom_expr ftab dsym v a e
 | LetUnop (v, _, _, a, e) ->
      free_unop_expr ftab dsym v a e
 | LetBinop (v, _, _, a1, a2, e) ->
      free_binop_expr ftab dsym v a1 a2 e
 | TailCall (f, args) ->
      free_tailcall_expr ftab dsym f args
 | LetExtCall (v, f, _, args, e) ->
      free_extcall_expr ftab dsym v f args e
 | LetString (v, s, e) ->
      free_string_expr ftab dsym v s e
 | IfThenElse (_, a1, a2, e1, e2) ->
      free_if_expr ftab dsym a1 a2 e1 e2
 | SetVar (v, _, a, e) ->
      free_set_var_expr ftab dsym v a e
 | SetSubscript (a1, _, a2, a3, e) ->
      free_set_subscript_expr ftab dsym a1 a2 a3 e
 | LetSubscript (v, _, a1, a2, e) ->
      free_subscript_expr ftab dsym v a1 a2 e
 | LetAddrOfSubscript (v, _, a1, a2, e) ->
      free_addr_of_subscript_expr ftab dsym v a1 a2 e
 | SetProject (a1, l, _, a2, e) ->
      free_set_project_expr ftab dsym a1 l a2 e
 | LetProject (v, _, a, l, e) ->
      free_project_expr ftab dsym v a l e
 | LetAddrOfProject (v, _, a, l, e) ->
      free_addr_of_project_expr ftab dsym v a l e
 | Memcpy (a1, _, a2, e) ->
      free_memcpy_expr ftab dsym a1 a2 e
 | LetAlloc (v, _, e) ->
      free_alloc_expr ftab dsym v e
 |(LetAddrOfVar _ | LetCopy _ | LetFuns _
 | LetApply _ | Return _ | LetClosure _ | IntCall _) as e ->
      let pos = string_pos "free_expr" (atom_pos e) in
         raise (IRException (pos, IRLevel 1))

and free_types_expr ftab dsym types e =
   free_expr ftab dsym e

and free_funs_expr ftab dsym funs e =
   (* Add empty entries for these funtions *)
   let add_fun ftab (f, _, _, _, _) = tab_add_key ftab f in

   (* Create the dataflow equation for one function *)
   let free_fun ftab (f, gflag, _, args, body) =
      let ftab, eqn = free_expr ftab dsym body in
      let eqn = eqn_sub_list eqn args in
      let eqn =
         match gflag with
          | FunGlobalClass ->
               eqn_add ftab eqn dsym
          | FunDeclClass
          | FunContClass
          | FunLocalClass ->
               eqn
      in
         tab_add ftab f eqn
   in

   let ftab = List.fold_left add_fun ftab funs in
   let ftab = List.fold_left free_fun ftab funs in
      free_expr ftab dsym e

and free_atom_expr ftab dsym v a e =
   free_unop_expr ftab dsym v a e

and free_unop_expr ftab dsym v a e =
   let ftab, eqn = free_expr ftab dsym e in
   let eqn = eqn_sub eqn v in
   let eqn = eqn_atom ftab eqn a in
      ftab, eqn

and free_binop_expr ftab dsym v a1 a2 e =
   let ftab, eqn = free_expr ftab dsym e in
   let eqn = eqn_sub eqn v in
   let eqn = eqn_atom ftab eqn a1 in
   let eqn = eqn_atom ftab eqn a2 in
      ftab, eqn

and free_tailcall_expr ftab dsym f args =
   let eqn = List.fold_left (eqn_atom ftab) eqn_empty args in
   let eqn = eqn_add ftab eqn f in
      ftab, eqn

and free_extcall_expr ftab dsym v f args e =
   let ftab, eqn = free_expr ftab dsym e in
   let eqn = eqn_sub eqn v in
   let eqn = List.fold_left (eqn_atom ftab) eqn args in
      ftab, eqn

and free_string_expr ftab dsym v s e =
   free_alloc_expr ftab dsym v e

and free_if_expr ftab dsym a1 a2 e1 e2 =
   let ftab, eqn1 = free_expr ftab dsym e1 in
   let ftab, eqn2 = free_expr ftab dsym e2 in
   let eqn = eqn_merge eqn1 eqn2 in
   let eqn = eqn_atom ftab eqn a1 in
   let eqn = eqn_atom ftab eqn a2 in
      ftab, eqn

and free_set_var_expr ftab dsym v a e =
   let ftab, eqn = free_expr ftab dsym e in
   let eqn = eqn_sub eqn v in
   let eqn = eqn_atom ftab eqn a in
      ftab, eqn

and free_set_subscript_expr ftab dsym a1 a2 a3 e =
   let ftab, eqn = free_expr ftab dsym e in
   let eqn = eqn_atom ftab eqn a1 in
   let eqn = eqn_atom ftab eqn a2 in
   let eqn = eqn_atom ftab eqn a3 in
      ftab, eqn

and free_subscript_expr ftab dsym v a1 a2 e =
   free_binop_expr ftab dsym v a1 a2 e

and free_addr_of_subscript_expr ftab dsym v a1 a2 e =
   free_binop_expr ftab dsym v a1 a2 e

and free_set_project_expr ftab dsym a1 l a2 e =
   free_memcpy_expr ftab dsym a1 a2 e

and free_project_expr ftab dsym v a l e =
   free_unop_expr ftab dsym v a e

and free_addr_of_project_expr ftab dsym v a l e =
   free_unop_expr ftab dsym v a e

and free_memcpy_expr ftab dsym a1 a2 e =
   let ftab, eqn = free_expr ftab dsym e in
   let eqn = eqn_atom ftab eqn a1 in
   let eqn = eqn_atom ftab eqn a2 in
      ftab, eqn

and free_alloc_expr ftab dsym v e =
   let ftab, eqn = free_expr ftab dsym e in
   let eqn = eqn_sub eqn v in
      ftab, eqn

(*
 * Besides the equation, we now need dependent list and free variable set
 * to solve the equations. Here is the converter (ftab -> vtab, ftab).
 * vtab contains free variables, while the new ftab contains eqn and dep.
 *)
let ftab_to_vtab ftab =
   let rec fun_depends = function
    | EqnNil -> set_empty
    | EqnAdd (eqn, _) -> fun_depends eqn
    | EqnSub (eqn, _) -> fun_depends eqn
    | EqnAddFun (eqn, f) ->
         let dep = fun_depends eqn in
            set_add dep f
    | EqnMerge (eqn1, eqn2) ->
         let dep1 = fun_depends eqn1 in
         let dep2 = fun_depends eqn2 in
            set_union dep1 dep2
   in
   let convert (vtab, ftab) f eqn =
      let eqn = List.hd eqn in (* the table stores a list of eqns *)
      let dep = set_to_list (fun_depends eqn) in
      let vtab = tab_add vtab f set_empty in
      let ftab = tab_add ftab f (eqn, dep) in
         vtab, ftab
   in
      tab_fold convert (tab_empty (), tab_empty ()) ftab

(*
 * Evaluate a free vars expression.
 *)
let rec eval_eqn vtab = function
 | EqnNil -> set_empty
 | EqnAddFun (eqn, f) ->
      let fv = eval_eqn vtab eqn in
      let fv' = tab_find vtab f in
         set_union fv fv'
 | EqnAdd (eqn, add) ->
      let fv = eval_eqn vtab eqn in
         set_union fv add
 | EqnSub (eqn, sub) ->
      let fv = eval_eqn vtab eqn in
         set_diff fv sub
 | EqnMerge (eqn1, eqn2) ->
      let fv1 = eval_eqn vtab eqn1 in
      let fv2 = eval_eqn vtab eqn2 in
         set_union fv1 fv2

(*
 * Evaluate a new vtab
 *)
let eval_vtab vtab ftab is_change =
   let eval_vtab_item (vtab', chg') f fv =
      let eqn, dep = tab_find ftab f in
      let fv = List.hd fv in
      let fv', chg' =
         if is_change dep then
            (* Update this set *)
            let fv' = eval_eqn vtab eqn in
               if set_equal fv fv' then
                  fv, chg'
               else
                  fv', set_add chg' f
         else
            fv, chg'
      in
         tab_add vtab' f fv', chg'
   in
      tab_fold eval_vtab_item (tab_empty (), set_empty) vtab

(*
 * Solve all the equations and get the free variables
 *)
let solve_vtab (vtab, ftab) =
   (* first time calc *)
   let vtab, change = eval_vtab vtab ftab (fun _ -> true) in
   (* Iterate until no changes *)
   let rec iterate (vtab, change) =
      if set_equal change set_empty then
         vtab
      else
         iterate (eval_vtab vtab ftab (List.exists (set_mem change)))
   in
   let vtab = iterate (vtab, change) in
   let to_list vtab f fv =
      tab_add vtab f (set_to_list (List.hd fv))
   in
      tab_fold to_list (tab_empty ()) vtab

(*
 * tenv, venv below are constructed as tenv, venv in ESC stage.
 * (not with free vars)
 *)

(*
 * Atom, Var: change funtion -> closure (may change venv)
 *)
let close_var tenv venv vtab pos v cont =
   try
      let v' = new_symbol v in
      let ty = venv_lookup venv pos v in
      let fv = tab_find vtab v in
      let args = make_args fv in
      let venv = venv_add venv v' ty in
         LetClosure (v', ty, v, args, cont venv v')
   with
      Not_found -> cont venv v

let close_atom tenv venv vtab pos a cont =
   match a with
    | AtomChar _ | AtomInt _ | AtomFloat _ | AtomNil _ ->
         cont venv a
    | AtomVar v ->
         close_var tenv venv vtab pos v (fun venv v ->
               cont venv (AtomVar v))

let close_atoms tenv venv vtab pos al cont =
   let rec loop venv bl = function
    | a :: al ->
         close_atom tenv venv vtab pos a (fun venv b ->
               loop venv (b :: bl) al)
    | [] ->
         cont venv (List.rev bl)
   in
      loop venv [] al

(*
 * Close the expression
 *)
let rec close_expr tenv venv vtab e =
   let pos = atom_pos e in
      match e with
         LetTypes (types, e) ->
            close_types_expr tenv venv vtab pos types e
       | LetFuns2 (funs, e) ->
            close_funs_expr tenv venv vtab pos funs e
       | LetAtom (v, ty, a, e) ->
            close_atom_expr tenv venv vtab pos v ty a e
       | LetUnop (v, ty, op, a, e) ->
            close_unop_expr tenv venv vtab pos v ty op a e
       | LetBinop (v, ty, op, a1, a2, e) ->
            close_binop_expr tenv venv vtab pos v ty op a1 a2 e
       | TailCall (f, args) ->
            close_tailcall_expr tenv venv vtab pos f args
       | LetExtCall (v, f, ty, args, e) ->
            close_extcall_expr tenv venv vtab pos v f ty args e
       | LetString (v, s, e) ->
            close_string_expr tenv venv vtab pos v s e
       | IfThenElse (op, a1, a2, e1, e2) ->
            close_if_expr tenv venv vtab pos op a1 a2 e1 e2
       | SetVar (v, ty, a, e) ->
            close_set_var_expr tenv venv vtab pos v ty a e
       | SetSubscript (a1, ty, a2, a3, e) ->
            close_set_subscript_expr tenv venv vtab pos ty a1 a2 a3 e
       | LetSubscript (v, ty, a1, a2, e) ->
            close_subscript_expr tenv venv vtab pos v ty a1 a2 e
       | LetAddrOfSubscript (v, ty, a1, a2, e) ->
            close_addr_of_subscript_expr tenv venv vtab pos v ty a1 a2 e
       | SetProject (a1, l, ty, a2, e) ->
            close_set_project_expr tenv venv vtab pos ty a1 l a2 e
       | LetProject (v, ty, a, l, e) ->
            close_project_expr tenv venv vtab pos v ty a l e
       | LetAddrOfProject (v, ty, a, l, e) ->
            close_addr_of_project_expr tenv venv vtab pos v ty a l e
       | Memcpy (a1, ty, a2, e) ->
            close_memcpy_expr tenv venv vtab pos a1 ty a2 e
       | LetAlloc (v, ty, e) ->
            close_alloc_expr tenv venv vtab pos v ty e
       | LetFuns _
       | LetAddrOfVar _
       | LetCopy _
       | LetApply _
       | Return _
       | LetClosure _
       | IntCall _ ->
            raise (IRException (pos, IRLevel 1))

and close_types_expr tenv venv vtab pos types e =
   let pos = string_pos "close_types_expr" pos in
   let tenv =
      List.fold_left (fun tenv (v, ty) -> tenv_add tenv v ty) tenv types
   in
      LetTypes (types, close_expr tenv venv vtab e)

and close_funs_expr tenv venv vtab pos funs e =
   let pos = string_pos "close_funs_expr" pos in
   let venv =
      List.fold_left (fun venv (f, _, ty, _, _) ->
            venv_add venv f ty) venv funs
   in
   (* Insert free variables to each function *)
   let insert_fv (f, gflag, ty, vars, e) =
      let ty_vars, ty_res = dest_fun_type tenv (int_pos 1 pos) ty in
      let venv = List.fold_left2 venv_add venv vars ty_vars in
      let e = close_expr tenv venv vtab e in
      let fv = tab_find vtab f in
      let ty_fv = List.map (venv_lookup venv (int_pos 2 pos)) fv in
      let ty = TyFun (ty_fv @ ty_vars, ty_res) in
         f, gflag, ty, fv @ vars, e
   in
   let funs = List.map insert_fv funs in
      LetFuns2 (funs, close_expr tenv venv vtab e)

and close_atom_expr tenv venv vtab pos v ty a e =
   let pos = string_pos "close_atom_expr" pos in
      close_atom tenv venv vtab pos a (fun venv a ->
            let venv = venv_add venv v ty in
               LetAtom (v, ty, a, close_expr tenv venv vtab e))

and close_unop_expr tenv venv vtab pos v ty op a e =
   let pos = string_pos "close_unop_expr" pos in
      close_atom tenv venv vtab pos a (fun venv a ->
            let venv = venv_add venv v ty in
               LetUnop (v, ty, op, a, close_expr tenv venv vtab e))

and close_binop_expr tenv venv vtab pos v ty op a1 a2 e =
   let pos = string_pos "close_binop_expr" pos in
      close_atom tenv venv vtab pos a1 (fun venv a1 ->
      close_atom tenv venv vtab pos a2 (fun venv a2 ->
            let venv = venv_add venv v ty in
               LetBinop (v, ty, op, a1, a2, close_expr tenv venv vtab e)))

and close_tailcall_expr tenv venv vtab pos f args =
   let pos = string_pos "close_tailcall_expr" pos in
      close_atoms tenv venv vtab pos args (fun venv args ->
            let fv = vtab_lookup vtab f in
            let args = (make_args fv) @ args in
               TailCall (f, args))

and close_extcall_expr tenv venv vtab pos v f ty args e =
   let pos = string_pos "close_extcall_expr" pos in
   let _, ty_res = dest_fun_type tenv pos ty in
   let venv = venv_add venv v ty_res in
      close_atoms tenv venv vtab pos args (fun venv args ->
            LetExtCall (v, f, ty, args, close_expr tenv venv vtab e))

and close_string_expr tenv venv vtab pos v s e =
   let venv = venv_add venv v (TyArray (TyChar, None)) in
      LetString (v, s, close_expr tenv venv vtab e)

and close_if_expr tenv venv vtab pos op a1 a2 e1 e2 =
   let pos = string_pos "close_if_expr" pos in
      close_atom tenv venv vtab pos a1 (fun venv a1 ->
      close_atom tenv venv vtab pos a2 (fun venv a2 ->
            let e1 = close_expr tenv venv vtab e1 in
            let e2 = close_expr tenv venv vtab e2 in
               IfThenElse (op, a1, a2, e1, e2)))

and close_set_var_expr tenv venv vtab pos v ty a e =
   let pos = string_pos "close_set_var_expr" pos in
      close_atom tenv venv vtab pos a (fun venv a ->
            LetAtom (v, ty, a, close_expr tenv venv vtab e))

and close_set_subscript_expr tenv venv vtab pos ty a1 a2 a3 e =
   let pos = string_pos "close_set_subscript_expr" pos in
      close_atom tenv venv vtab pos a1 (fun venv a1 ->
      close_atom tenv venv vtab pos a2 (fun venv a2 ->
      close_atom tenv venv vtab pos a3 (fun venv a3 ->
            SetSubscript (a1, ty, a2, a3, close_expr tenv venv vtab e))))

and close_subscript_expr tenv venv vtab pos v ty a1 a2 e =
   let pos = string_pos "close_subscript_expr" pos in
      close_atom tenv venv vtab pos a1 (fun venv a1 ->
      close_atom tenv venv vtab pos a2 (fun venv a2 ->
            let venv = venv_add venv v ty in
               LetSubscript (v, ty, a1, a2, close_expr tenv venv vtab e)))

and close_addr_of_subscript_expr tenv venv vtab pos v ty a1 a2 e =
   let pos = string_pos "close_addr_of_subscript_expr" pos in
      close_atom tenv venv vtab pos a1 (fun venv a1 ->
      close_atom tenv venv vtab pos a2 (fun venv a2 ->
            let venv = venv_add venv v ty in
               LetAddrOfSubscript (v, ty, a1, a2, close_expr tenv venv vtab e)))

and close_set_project_expr tenv venv vtab pos ty a1 l a2 e =
   let pos = string_pos "close_set_project_expr" pos in
      close_atom tenv venv vtab pos a1 (fun venv a1 ->
      close_atom tenv venv vtab pos a2 (fun venv a2 ->
            SetProject (a1, l, ty,  a2, close_expr tenv venv vtab e)))

and close_project_expr tenv venv vtab pos v ty a l e =
   let pos = string_pos "close_project_expr" pos in
      close_atom tenv venv vtab pos a (fun venv a ->
            let venv = venv_add venv v ty in
               LetProject (v, ty, a, l, close_expr tenv venv vtab e))

and close_addr_of_project_expr tenv venv vtab pos v ty a l e =
   let pos = string_pos "close_addr_of_project_expr" pos in
      close_atom tenv venv vtab pos a (fun venv a ->
            let venv = venv_add venv v ty in
               LetAddrOfProject (v, ty, a, l, close_expr tenv venv vtab e))

and close_memcpy_expr tenv venv vtab pos a1 ty a2 e =
   let pos = string_pos "close_memcpy_expr" pos in
      close_atom tenv venv vtab pos a1 (fun venv a1 ->
      close_atom tenv venv vtab pos a2 (fun venv a2 ->
            Memcpy (a1, ty, a2, close_expr tenv venv vtab e)))

and close_alloc_expr tenv venv vtab pos v ty e =
   let venv = venv_add venv v (TyArray (ty, None)) in
      LetAlloc (v, ty, close_expr tenv venv vtab e)

(*
 * Print a list of vars.
 *)
let rec print_var_list firstp = function
 | [] -> ()
 | v :: vars ->
      if not firstp then
         begin
            Format.print_string ";";
            Format.print_space ()
         end;
      print_symbol v;
      print_var_list false vars

let print_var_set vars =
   Format.open_box 0;
   Format.print_string "{";
   print_var_list true (set_to_list vars);
   Format.print_string "}";
   Format.close_box ()

(*
 * Print the free vars equation.
 *)
let rec print_eqn = function
 | EqnNil ->
      Format.print_string "{}"
 | EqnAddFun (eqn, f) ->
      Format.open_box tabstop;
      print_eqn eqn;
      Format.print_space ();
      Format.print_string "+ ";
      print_symbol f;
      Format.close_box ()
 | EqnAdd (eqn, add) ->
      Format.open_box tabstop;
      print_eqn eqn;
      Format.print_space ();
      Format.print_string "+ ";
      print_var_set add;
      Format.close_box ()
 | EqnSub (eqn, sub) ->
      Format.open_box tabstop;
      print_eqn eqn;
      Format.print_space ();
      Format.print_string "- ";
      print_var_set sub;
      Format.close_box ()
 | EqnMerge (eqn1, eqn2) ->
      Format.open_box tabstop;
      Format.print_string "(";
      print_eqn eqn1;
      Format.print_string ")";
      Format.print_space ();
      Format.print_string "+ (";
      print_eqn eqn1;
      Format.print_string ")";
      Format.close_box ()

(*
 * Print the closure equations.
 *)
let print_ftab ftab =
   Format.open_vbox tabstop;
   Format.print_string "Dataflow equations:";
   Format.print_space ();
   tab_iter (fun f fvars ->
         Format.open_hvbox tabstop;
         print_symbol f;
         Format.print_string " = ";
         print_eqn fvars;
         Format.close_box ();
         Format.print_space ()) ftab;
   Format.close_box ();
   Format.print_newline ()

(*
 * Print the solved equations.
 *)
let print_vtab vtab =
   Format.open_vbox tabstop;
   Format.print_string "Free variables:";
   Format.print_space ();
   tab_iter (fun f fvars ->
         print_symbol f;
         Format.print_string " = ";
         List.iter (fun v ->
               Format.print_string " ";
               print_symbol v) fvars;
         Format.print_space ()) vtab;
   Format.close_box ();
   Format.print_newline ()

(*
 * Global version.
 * data_sym is the name of the data segment.
 * data_type is the struct type for the data segment.
 *)
let close_expr tenv venv data_sym data_type e =
   let ftab, eqn = free_expr (tab_empty ()) data_sym e in
   if !debug_closure then print_ftab ftab;

   let vtab = solve_vtab (ftab_to_vtab ftab) in
   if !debug_closure then print_vtab vtab;

   let venv = venv_add venv data_sym (TyArray (data_type, None)) in
   let e = close_expr tenv venv vtab e in
      standardize_expr_notype e
