(*
 * Escape analysis
 *
 * Ling Li, Xin Yu
 * Mar 13, 2001, Caltech
 *
 * This step is called escape analysis.  Each global function
 * allocates a record, called a "frame."  Variables that escape
 * are stored in the frame.  A variable escapes if:
 *    1. it is an aggregate (a struct), or
 *    2. it is used in a nested global function, or
 *    3. it's address is taken.
 *
 * We do this in two passes.  In the first pass, we walk the
 * program, and compute which variables escape.  We assume all
 * variable binding occurrences bind different variables (the
 * program has been standardized).  This means we can use a table
 * to record tyhe frame for each escaping variable.  A frame stack
 * is passed during the analysis for the nesting level of global
 * functions.
 *
 * In the second pass, we transform the code.
 *    1. Add an explicit frame allocation to each global function
 *       definition using LetAlloc.
 *    2. All accesses to escaping variables are transformed
 *       into record operations on the frame.  For struct
 *       assignments and initializations, we use Memcpy to
 *       copy the date.  For scalar assignments, we use
 *       SetProject.
 *
 * Global variables are also stored in an outermost frame.
 * There is no need to distinguish between top-level expressions
 * and expressions inside a function definition.  Another way to think
 * about it: it is as if the entire program were nested inside an implicit
 * function definition.
 *
 * All global flags for nested functions are cleared.
 *)

open Symbol
open Fc_ir_type
open Fc_ir_exn_type
open Fc_ir_exn
open Fc_ir_env
open Fc_ir_standardize

let debug_esc = ref false

(************************************************************************
 * TYPES
 ************************************************************************)

(*
 * Frames are just vars.
 *)
type frame = var
type full_frame =
   { frame_fun : var;
     frame_frame : frame
   }

(*
 * A variable has:
 *    1. a type
 *    2. an optional frame where the variable is bound
 *)
type venv =
   { venv_env : SymbolSet.t;
     venv_vars : (ty * full_frame) SymbolTable.t
   }

(************************************************************************
 * ENVIRONMENTS
 ************************************************************************)

(*
 * Default exception.
 *)
let raise_exn code v =
   IRException (code, v)

let tenv_lookup      = tenv_lookup raise_exn
let tenv_expand      = tenv_expand raise_exn

(*
 * Empty table.
 *)
let empty_venv =
   { venv_env = SymbolSet.empty;
     venv_vars = SymbolTable.create ()
   }

let venv_mem venv v =
   SymbolSet.mem v venv.venv_env

(*
 * Get the value.
 *)
let venv_lookup venv pos v =
   try SymbolTable.find venv.venv_vars v with
      Not_found ->
         raise (IRException (pos, UnboundVar v))

(*
 * Add a value.
 *)
let venv_add venv v ty frame =
   let { venv_env = env;
         venv_vars = vars
       } = venv
   in
      { venv_env = SymbolSet.add v env;
        venv_vars = SymbolTable.add vars v (ty, frame)
      }
(*
 * We adjust the calling convention here.
 * We add an extra argument to functions that
 * return an aggregate type.
 *)
let return_sym = Symbol.new_symbol_string "return"

(*
 * The data segment is also a frame.
 *)
let data_sym = new_symbol_string "data_seg"

(************************************************************************
 * UTILITIES
 ************************************************************************)

(*
 * Expand a function type.
 *)
let is_fun_type tenv pos ty =
   match tenv_expand tenv pos ty with
      TyFun _ -> true
    | _ -> false

let dest_fun_type tenv pos ty =
   match tenv_expand tenv pos ty with
      TyFun (ty_vars, ty_res) -> ty_vars, ty_res
    | ty -> raise (IRException (pos, NotAFunction ty))

(*
 * Get the type of an atom.
 *)
let get_atom_type venv pos = function
   AtomChar _ -> TyChar
 | AtomInt _ -> TyInt
 | AtomFloat _ -> TyFloat
 | AtomNil ty -> TyArray (ty, None)
 | AtomVar v -> fst (venv_lookup venv pos v)

(*
 * Default value in a non-aggregate type.
 *)
let default_value tenv pos ty =
   let ty = tenv_expand tenv pos ty in
      match ty with
         TyChar -> AtomChar '\000'
       | TyInt -> AtomInt 0
       | TyFloat -> AtomFloat 0.0
       | TyArray (ty, None) -> AtomNil ty

       | TyAny
       | TyArray (_, Some _)
       | TyStruct _
       | TyFun _ ->
            raise (IRException (pos, CantInitialize ty))

       | TyId _ ->
            raise (Invalid_argument "Fc_ir_closure1.default_value")

(*
 * Check for an aggregate type.
 *)
let is_aggregate_type tenv pos ty =
   match tenv_expand tenv pos ty with
      TyArray (_, Some _)
    | TyStruct _ ->
         true
    | TyAny
    | TyChar
    | TyInt
    | TyFloat
    | TyArray (_, None)
    | TyFun _ ->
         false
    | TyId _ ->
         raise (Invalid_argument "Fc_ir_closure.is_aggregate_type")

(*
 * Set intersection.
 *)
let set_intersect set1 set2 =
   SymbolSet.fold (fun set v ->
         if SymbolSet.mem v set2 then
            SymbolSet.add v set
         else
            set) SymbolSet.empty set1

(************************************************************************
 * FIRST PASS: COLLECT ESCAPING VARS
 ************************************************************************)

(*
 * Escape environment.  This phase calculates the variables
 * that "escape".  A variable escapes if it is an aggregate or
 * it is used in a function other than the one where it is defined.
 *
 * Note, it is not necessary for scalar variables to escape
 * (unless they have their address taken).  Closure conversion
 * can be used to handle them correctly.  However, we keep them
 * in the frame anyway, so that the environment is stored in a
 * single record (the frame).
 *
 * We keep several parts:
 *    esc_stack: a list of frames in scope
 *    esc_vars: a table of variables that escape,
 *       the frame they belong to, and whether access
 *       _must_ be through the frame.  Access must be through
 *       the frame for variables that are aggregates, or
 *       variables that have their address taken.
 *
 * For each function, we keep additional information:
 *    fun_frames: a list of frames in scope for each function
 *    fun_gflag: the global flag for the function
 *    fun_defs: the variables defined by the function
 *       (either through a LetCopy definition, or an assignment)
 *    fun_uses: the free variables in the function
 *    fun_calls: the functions that are called
 *
 * The defs and uses are for a dataflow analysis that will push
 * all frame writes to right before a tailcall, and all frame
 * reads on entry to a global fun or continuation.
 *)
type esc_env = SymbolSet.t
type esc_defs = SymbolSet.t
type esc_uses = SymbolSet.t
type esc_calls = (esc_defs * var * atom list) list

type fun_info =
   { fun_frames : frame list;
     fun_gflag : fun_class;
     fun_label : bool;
     fun_venv : esc_env;
     fun_oenv : esc_env;
     fun_ienv : esc_env;
     fun_defs : esc_defs;
     fun_uses : esc_uses;
     fun_calls : esc_calls
   }

type esc =
   { esc_stack : full_frame list;
     esc_vars : (ty * frame * bool) SymbolTable.t;
     esc_funs : fun_info SymbolTable.t;
     esc_defs : esc_defs;
     esc_uses : esc_uses;
     esc_calls : esc_calls
   }

(*
 * Empty escape environment include the
 * outermost frame.
 *)
let empty_esc =
   { esc_stack = [{ frame_fun = data_sym; frame_frame = data_sym }];
     esc_vars = SymbolTable.create ();
     esc_funs = SymbolTable.create ();
     esc_defs = SymbolSet.empty;
     esc_uses = SymbolSet.empty;
     esc_calls = []
   }

(************************************************
 * DEF/USE OPERATIONS
 *)

(*
 * Add a variable to the use set.
 *)
let esc_add_use esc v =
   { esc with esc_uses = SymbolSet.add v esc.esc_uses }

(*
 * Remove a variable from the use set.
 *)
let esc_remove_use esc v =
   { esc with esc_uses = SymbolSet.remove v esc.esc_uses }

(*
 * Add a def with a reference to another var.
 *)
let esc_add_def esc v =
   { esc with esc_defs = SymbolSet.add v esc.esc_defs }

(*
 * Add a function call.
 *)
let esc_add_call esc f args =
   { esc with esc_calls = (esc.esc_defs, f, args) :: esc.esc_calls }

(*
 * Combining defs/uses for a two-way branch.
 *)
let esc_defs esc =
   esc.esc_defs

let esc_reset_defs esc defs =
   { esc with esc_defs = defs }

(************************************************
 * FRAME PUSHING
 *)

(*
 * Add a function variable.
 *)
let esc_add_fun_var esc f venv =
   let { esc_stack = stack; esc_funs = funs } = esc in
   let frames = List.map (fun frame -> frame.frame_frame) (List.tl stack) in
   let fun_info =
      { fun_frames = frames;
        fun_gflag = FunGlobalClass;
        fun_label = false;
        fun_venv = venv.venv_env;
        fun_oenv = SymbolSet.empty;
        fun_ienv = venv.venv_env;
        fun_defs = SymbolSet.empty;
        fun_uses = SymbolSet.empty;
        fun_calls = []
      }
   in
      { esc with esc_funs = SymbolTable.add funs f fun_info }

(*
 * Add a new frame to the escape table.
 *)
let push_frame esc f frame gflag vars venv =
   let { esc_stack = stack; esc_funs = funs } = esc in
   let frames = frame :: stack in
   let frames' = List.map (fun frame -> frame.frame_frame) frames in
   let env = SymbolSet.of_list vars in
   let fun_info =
      { fun_frames = frames';
        fun_gflag = gflag;
        fun_label = true;
        fun_venv = SymbolSet.union venv.venv_env env;
        fun_oenv = SymbolSet.empty;
        fun_ienv = env;
        fun_defs = env;
        fun_uses = SymbolSet.empty;
        fun_calls = []
      }
   in
      { esc with esc_stack = frames;
                 esc_funs = SymbolTable.add funs f fun_info
      }

(*
 * Get the current frame, and the rest of the stack.
 *)
let pop_frame pos = function
   { esc_stack = frame :: funs } as esc ->
      frame, { esc with esc_stack = funs }
 | { esc_stack = [] } ->
      raise (IRException (pos, InternalError "pop_frame"))

(*
 * Add a new frame to the escape table.
 *)
let replace_frame esc pos f gflag vars venv =
   let { esc_stack = stack; esc_funs = funs } = esc in
      match stack with
         frame :: stack ->
            let frame' = { frame with frame_fun = f } in
            let frames = frame' :: stack in
            let frames' = List.map (fun frame -> frame.frame_frame) frames in
            let env1 = SymbolSet.of_list vars in
            let env2 = SymbolSet.union venv.venv_env env1 in
            let oenv, ienv =
               match gflag with
                  FunLocalClass -> venv.venv_env, env2
                | FunGlobalClass
                | FunContClass
                | FunDeclClass -> SymbolSet.empty, env1
            in
            let fun_info =
               { fun_frames = frames';
                 fun_gflag = gflag;
                 fun_label = true;
                 fun_venv = env2;
                 fun_oenv = oenv;
                 fun_ienv = ienv;
                 fun_defs = env1;
                 fun_uses = SymbolSet.empty;
                 fun_calls = []
               }
            in
            let esc =
               { esc with esc_stack = frames;
                          esc_funs = SymbolTable.add funs f fun_info
               }
            in
               frame, frame', esc
       | [] ->
            raise (IRException (pos, InternalError "frame stack is empty"))

(*
 * Restore the old frame.
 *)
let restore_frame frame pos = function
   { esc_stack = _ :: funs } as esc ->
      { esc with esc_stack = frame :: funs }
 | { esc_stack = [] } ->
      raise (IRException (pos, InternalError "restore_frame"))

(*
 * Clear defs/uses at entry to function call.
 *)
let esc_reset_dflow esc =
   let { esc_defs = defs;
         esc_uses = uses;
         esc_calls = calls
       } = esc
   in
   let esc =
      { esc with esc_defs = SymbolSet.empty;
                 esc_uses = SymbolSet.empty;
                 esc_calls = []
      }
   in
      esc, defs, uses, calls

(*
 * Save defs/uses at function exit.
 *)
let esc_restore_dflow esc f defs uses calls =
   let { esc_uses = uses';
         esc_calls = calls';
         esc_funs = funs'
       } = esc
   in
   let fun_info = SymbolTable.find funs' f in
   let uses' = SymbolSet.subtract_list uses' (SymbolSet.to_list fun_info.fun_defs) in
   let fun_info =
      { fun_info with fun_uses = uses';
                      fun_calls = calls'
      }
   in
   let funs = SymbolTable.add funs' f fun_info in
      { esc with esc_defs = defs;
                 esc_uses = uses;
                 esc_calls = calls;
                 esc_funs = funs
      }

(************************************************
 * BASIC ESC ANALYSIS
 *)

(*
 * Add an escaping variable to the frame.
 *)
let esc_add esc v ty aflag frame =
   { esc with esc_vars = SymbolTable.add esc.esc_vars v (ty, frame.frame_frame, aflag) }

(*
 * Add an atom.
 * If the atom is a var, and it is bound in
 * another frame, add it to the escape set.
 * Also compute the use set.
 *)
let esc_var tenv venv esc pos v =
   let add_var v =
      if venv_mem venv v then
         let frame', _ = pop_frame pos esc in
         let ty, frame = venv_lookup venv pos v in
         let esc = esc_add_use esc v in
            if is_aggregate_type tenv pos ty then
               esc_add esc v ty true frame
            else if frame' <> frame then
               esc_add esc v ty false frame
            else
               esc
      else
         esc
   in
      try
         let fun_info = SymbolTable.find esc.esc_funs v in
            if fun_info.fun_label then
               esc
            else
               add_var v
      with
         Not_found ->
            add_var v

let esc_var_def tenv venv esc pos v =
   let add_var v =
      if venv_mem venv v then
         let frame', _ = pop_frame pos esc in
         let ty, frame = venv_lookup venv pos v in
         let esc = esc_add_def esc v in
            if is_aggregate_type tenv pos ty then
               esc_add esc v ty true frame
            else if frame' <> frame then
               esc_add esc v ty false frame
            else
               esc
      else
         esc
   in
      try
         let fun_info = SymbolTable.find esc.esc_funs v in
            if fun_info.fun_label then
               esc
            else
               add_var v
      with
         Not_found ->
            add_var v

let esc_atom tenv venv esc pos a =
   let pos = string_pos "esc_atom" pos in
      match a with
         AtomChar _
       | AtomInt _
       | AtomFloat _
       | AtomNil _ ->
            esc
       | AtomVar v ->
            esc_var tenv venv esc pos v

let esc_atoms tenv venv esc pos atoms =
   List.fold_left (fun esc a ->
         esc_atom tenv venv esc pos a) esc atoms

let esc_atom_opt tenv venv esc pos a =
   match a with
      Some a ->
         esc_atom tenv venv esc pos a
    | None ->
         esc

(************************************************
 * EXPR ESC ANALYSIS
 *)

(*
 * Transform an expression.
 *)
let rec esc_expr tenv venv esc e =
   let pos = string_pos "esc_expr" (atom_pos e) in
      match e with
         LetTypes (types, e) ->
            esc_types_expr tenv venv esc pos types e
       | LetFuns2 (funs, e) ->
            esc_funs_expr tenv venv esc pos funs e
       | LetAtom (v, ty, a, e) ->
            esc_atom_expr tenv venv esc pos v ty a e
       | LetCopy (v, ty, a, e) ->
            esc_copy_expr tenv venv esc pos v ty a e
       | LetUnop (v, ty, op, a, e) ->
            esc_unop_expr tenv venv esc pos v ty op a e
       | LetBinop (v, ty, op, a1, a2, e) ->
            esc_binop_expr tenv venv esc pos v ty op a1 a2 e
       | TailCall (f, args) ->
            esc_tailcall_expr tenv venv esc pos f args
       | LetExtCall (v, f, ty, args, e) ->
            esc_extcall_expr tenv venv esc pos v f ty args e
       | LetString (v, s, e) ->
            esc_string_expr tenv venv esc pos v s e
       | IfThenElse (op, a1, a2, e1, e2) ->
            esc_if_expr tenv venv esc pos op a1 a2 e1 e2
       | SetVar (v, ty, a, e) ->
            esc_set_var_expr tenv venv esc pos v ty a e
       | LetAddrOfVar (v1, ty, v2, e) ->
            esc_addr_of_var_expr tenv venv esc pos v1 ty v2 e
       | SetSubscript (a1, ty, a2, a3, e) ->
            esc_set_subscript_expr tenv venv esc pos ty a1 a2 a3 e
       | LetSubscript (v, ty, a1, a2, e) ->
            esc_subscript_expr tenv venv esc pos v ty a1 a2 e
       | LetAddrOfSubscript (v, ty, a1, a2, e) ->
            esc_addr_of_subscript_expr tenv venv esc pos v ty a1 a2 e
       | SetProject (a1, l, ty, a2, e) ->
            esc_set_project_expr tenv venv esc pos ty a1 l a2 e
       | LetProject (v, ty, a, l, e) ->
            esc_project_expr tenv venv esc pos v ty a l e
       | LetAddrOfProject (v, ty, a, l, e) ->
            esc_addr_of_project_expr tenv venv esc pos v ty a l e
       | LetFuns _
       | LetApply _
       | Return _
       | Memcpy _
       | LetAlloc _
       | LetClosure _
       | IntCall _ ->
            raise (IRException (pos, IRLevel 1))

(*
 * Add the type definitions.
 *)
and esc_types_expr tenv venv esc pos types e =
   let pos = string_pos "esc_types_expr" pos in
   let tenv =
      List.fold_left (fun tenv (v, ty) -> tenv_add tenv v ty) tenv types
   in
      esc_expr tenv venv esc e

(*
 * Function definitions.
 * Global functions should push a new frame
 * onto the frame stack.  Also, all aggregate arguments
 * escape.
 *)
and esc_fun_global tenv venv esc pos f gflag ty vars body =
   (* Add a new frame for this function *)
   let pos = string_pos "esc_fun_global" pos in
   let ty_vars, _ = dest_fun_type tenv (int_pos 1 pos) ty in
   let frame = { frame_fun = f; frame_frame = new_symbol_string "frame" } in
   let esc, defs, uses, calls = esc_reset_dflow esc in
   let esc = push_frame esc f frame gflag vars venv in

   (* Add aggregate args to the frame *)
   let esc, venv =
      List.fold_left2 (fun (esc, venv) v ty ->
            let esc =
               if is_aggregate_type tenv pos ty then
                  esc_add esc v ty true frame
               else if is_fun_type tenv pos ty then
                  esc_add_fun_var esc v venv
               else
                  esc
            in
               esc, venv_add venv v ty frame) (esc, venv) vars ty_vars
   in

   (* Convert the body *)
   let esc = esc_expr tenv venv esc body in
   let _, esc = pop_frame pos esc in
   let esc = esc_restore_dflow esc f defs uses calls in
      esc

and esc_fun_local tenv venv esc pos f gflag ty vars body =
   (* Get the frame *)
   let pos = string_pos "esc_fun_local" pos in
   let ty_vars, _ = dest_fun_type tenv pos ty in
   let esc, defs, uses, calls = esc_reset_dflow esc in
   let frame_orig, frame, esc = replace_frame esc pos f gflag vars venv in

   (* Add aggregate args to the frame *)
   let esc, venv =
      List.fold_left2 (fun (esc, venv) v ty ->
            let esc =
               if is_aggregate_type tenv pos ty then
                  esc_add esc v ty true frame
               else if is_fun_type tenv pos ty then
                  esc_add_fun_var esc v venv
               else
                  esc
            in
               esc, venv_add venv v ty frame) (esc, venv) vars ty_vars
   in

   (* Convert the body *)
   let esc = esc_expr tenv venv esc body in
   let esc = restore_frame frame_orig pos esc in
   let esc = esc_restore_dflow esc f defs uses calls in
      esc

and esc_funs_expr tenv venv esc pos funs e =
   (* Add all the functions to the variable environment *)
   let pos = string_pos "esc_funs_expr" pos in
   let frame, _ = pop_frame (int_pos 0 pos) esc in
   let venv =
      List.fold_left (fun venv (f, _, ty, _, _) ->
            venv_add venv f ty frame) venv funs
   in

   (* Convert all the funs *)
   let esc =
      List.fold_left (fun esc (f, gflag, ty, vars, body) ->
            let esc_fun =
               match gflag with
                  FunGlobalClass ->
                     esc_fun_global
                | FunDeclClass
                | FunContClass
                | FunLocalClass ->
                     esc_fun_local
            in
               esc_fun tenv venv esc (int_pos 2 pos) f gflag ty vars body) esc funs
   in
      (* Convert the body *)
      esc_expr tenv venv esc e

(*
 * Atom assignment.  The atom assignment is
 * _not_ a copy.
 *)
and esc_atom_expr tenv venv esc pos v ty a e =
   (* Compute the rest using the current type *)
   let pos = string_pos "esc_atom_expr" pos in
   let frame, _ = pop_frame pos esc in
   let venv = venv_add venv v ty frame in
   let esc = esc_add_def esc v in
   let esc = esc_expr tenv venv esc e in
   let esc = esc_remove_use esc v in
   let esc = esc_atom tenv venv esc pos a in
      esc

(*
 * Copy expression is an allocation point.
 * If the value is an aggregate, it escapes.
 *)
and esc_copy_expr tenv venv esc pos v ty a e =
   (* Compute the rest using the current type *)
   let pos = string_pos "esc_copy_expr" pos in
   let frame, _ = pop_frame pos esc in
   let venv = venv_add venv v ty frame in
   let esc = esc_add_def esc v in
   let esc = esc_expr tenv venv esc e in
   let esc = esc_remove_use esc v in
   let esc = esc_atom_opt tenv venv esc pos a in
      if is_aggregate_type tenv pos ty then
         esc_add esc v ty true frame
      else
         esc

(*
 * Unary operation.
 * Arguments are always of scalar type, so
 * we don't worry about aggregates.
 *)
and esc_unop_expr tenv venv esc pos v ty op a e =
   let pos = string_pos "esc_unop_expr" pos in
   let frame, _ = pop_frame pos esc in
   let venv = venv_add venv v ty frame in
   let esc = esc_add_def esc v in
   let esc = esc_expr tenv venv esc e in
   let esc = esc_remove_use esc v in
   let esc = esc_atom tenv venv esc pos a in
      esc

(*
 * Binary operation.
 * Arguments are always of scalar type.
 *)
and esc_binop_expr tenv venv esc pos v ty op a1 a2 e =
   let pos = string_pos "esc_binop_expr" pos in
   let frame, _ = pop_frame pos esc in
   let venv = venv_add venv v ty frame in
   let esc = esc_add_def esc v in
   let esc = esc_expr tenv venv esc e in
   let esc = esc_remove_use esc v in
   let esc = esc_atom tenv venv esc pos a1 in
   let esc = esc_atom tenv venv esc pos a2 in
      esc

(*
 * Tailcall.
 * Add all the vars that escape.
 * The functions do not escape.
 *)
and esc_tailcall_expr tenv venv esc pos f args =
   let pos = string_pos "esc_tailcall_expr" pos in
   let esc = esc_var tenv venv esc pos f in
   let esc = esc_atoms tenv venv esc pos args in
   let esc = esc_add_call esc f args in
      esc

(*
 * External function call.
 *)
and esc_extcall_expr tenv venv esc pos v f ty args e =
   let pos = string_pos "esc_extcall_expr" pos in
   let frame, _ = pop_frame pos esc in
   let _, ty_res = dest_fun_type tenv pos ty in
   let venv = venv_add venv v ty_res frame in
   let esc = esc_add_def esc v in
   let esc = esc_expr tenv venv esc e in
   let esc = esc_remove_use esc v in
   let esc = esc_atoms tenv venv esc pos args in
      if is_aggregate_type tenv pos ty_res then
         esc_add esc v ty_res true frame
      else
         esc

(*
 * String allocation.
 *)
and esc_string_expr tenv venv esc pos v s e =
   let pos = string_pos "esc_string_expr" pos in
   let frame, _ = pop_frame pos esc in
   let ty = TyArray (TyChar, None) in
   let venv = venv_add venv v ty frame in
   let esc = esc_add_def esc v in
   let esc = esc_expr tenv venv esc e in
   let esc = esc_remove_use esc v in
      esc

(*
 * Conditional.
 *)
and esc_if_expr tenv venv esc pos op a1 a2 e1 e2 =
   let pos = string_pos "esc_if_expr" pos in
   let esc = esc_atom tenv venv esc pos a1 in
   let esc = esc_atom tenv venv esc pos a2 in
   let defs = esc_defs esc in
   let esc = esc_expr tenv venv esc e1 in
   let esc = esc_reset_defs esc defs in
   let esc = esc_expr tenv venv esc e2 in
      esc

(*
 * Set a variable.
 *)
and esc_set_var_expr tenv venv esc pos v ty a e =
   let pos = string_pos "esc_set_var_expr" pos in
   let esc = esc_var_def tenv venv esc pos v in
   let esc = esc_atom tenv venv esc pos a in
      esc_expr tenv venv esc e

(*
 * Address of a variable is eliminated.
 * If the variable is an aggregate, the AddressOf operation
 * is removed.  Otherwise, the variable is added to the escape
 * set.
 *)
and esc_addr_of_var_expr tenv venv esc pos v1 ty v2 e =
   (* Add the variable to the escape set *)
   let pos = string_pos "esc_addr_of_var_expr" pos in
   let frame, _ = pop_frame pos esc in
   let ty', frame' = venv_lookup venv pos v2 in
   let venv = venv_add venv v1 ty frame in
   let esc = esc_add_def esc v1 in
   let esc = esc_expr tenv venv esc e in
   let esc = esc_remove_use esc v1 in
   let esc = esc_add esc v2 ty' true frame' in
      esc

(*
 * Set an array entry.
 *)
and esc_subscript_expr tenv venv esc pos v ty a1 a2 e =
   let pos = string_pos "esc_subscript_expr" pos in
   let frame, _ = pop_frame pos esc in
   let venv = venv_add venv v ty frame in
   let esc = esc_add_def esc v in
   let esc = esc_expr tenv venv esc e in
   let esc = esc_remove_use esc v in
   let esc = esc_atom tenv venv esc pos a1 in
   let esc = esc_atom tenv venv esc pos a2 in
      esc

and esc_set_subscript_expr tenv venv esc pos ty a1 a2 a3 e =
   let pos = string_pos "esc_set_subscript_expr" pos in
   let esc = esc_atom tenv venv esc pos a1 in
   let esc = esc_atom tenv venv esc pos a2 in
   let esc = esc_atom tenv venv esc pos a3 in
      esc_expr tenv venv esc e

and esc_addr_of_subscript_expr tenv venv esc pos v ty a1 a2 e =
   let pos = string_pos "esc_addr_of_subscript_expr" pos in
   let frame, _ = pop_frame pos esc in
   let venv = venv_add venv v ty frame in
   let esc = esc_add_def esc v in
   let esc = esc_expr tenv venv esc e in
   let esc = esc_remove_use esc v in
   let esc = esc_atom tenv venv esc pos a1 in
   let esc = esc_atom tenv venv esc pos a2 in
      esc

(*
 * Struct projections.
 *)
and esc_project_expr tenv venv esc pos v ty a l e =
   let pos = string_pos "esc_project_expr" pos in
   let frame, _ = pop_frame pos esc in
   let venv = venv_add venv v ty frame in
   let esc = esc_add_def esc v in
   let esc = esc_expr tenv venv esc e in
   let esc = esc_remove_use esc v in
   let esc = esc_atom tenv venv esc pos a in
      esc

and esc_set_project_expr tenv venv esc pos ty a1 l a2 e =
   let pos = string_pos "esc_set_project_expr" pos in
   let esc = esc_atom tenv venv esc pos a1 in
   let esc = esc_atom tenv venv esc pos a2 in
      esc_expr tenv venv esc e

and esc_addr_of_project_expr tenv venv esc pos v ty a l e =
   let pos = string_pos "esc_addr_of_project_expr" pos in
   let frame, _ = pop_frame pos esc in
   let venv = venv_add venv v ty frame in
   let esc = esc_add_def esc v in
   let esc = esc_expr tenv venv esc e in
   let esc = esc_remove_use esc v in
   let esc = esc_atom tenv venv esc pos a in
      esc

(************************************************************************
 * DATAFLOW ANALYSIS
 ************************************************************************)

(*
 * Build a use-table: a table of functions that must be
 * updated when the uses for the given function are changed.
 *)
let dflow_use_table funs =
   let add_atom_use uses f = function
      AtomChar _
    | AtomInt _
    | AtomFloat _
    | AtomNil _ ->
         uses
    | AtomVar v ->
         try
            let fun_info = SymbolTable.find funs v in
               if fun_info.fun_label then
                  SymbolTable.add uses v f
               else
                  uses
         with
            Not_found ->
               uses
   in
   let add_atom_uses use_table f args' =
      List.fold_left (fun use_table a -> add_atom_use use_table f a) use_table args'
   in
   let use_table =
      SymbolTable.fold (fun use_table f fun_infos ->
            let { fun_calls = calls } = List.hd fun_infos in
               List.fold_left (fun use_table (_, f', args') ->
                     let use_table = SymbolTable.add use_table f' f in
                        add_atom_uses use_table f args') use_table calls) (**)
         (SymbolTable.create ()) funs
   in
   let use_table =
      SymbolTable.fold (fun use_table f callers ->
            SymbolTable.add use_table f (SymbolSet.of_list callers)) (**)
         (SymbolTable.create ()) use_table
   in
      use_table

(*
 * Add the callers to the workset.
 *)
let dflow_add_callers workset use_table f =
   try
      let callers = SymbolTable.find use_table f in
         SymbolSet.union workset callers
   with
      Not_found ->
         workset

(*
 * Add all the frame vars to the unknown functions.
 * For closures, we can't be sure exactly which values
 * they use.  To be safe, we assume they can access any
 * escaping variable in any of their frames.
 *)
let dflow_add_unknown_uses vars funs =
   let add_defs frames defs uses f =
      try
         let _, frame', _ = SymbolTable.find vars f in
            SymbolSet.fold (fun uses v ->
                  try
                     let _, frame, aflag = SymbolTable.find vars v in
                        if not aflag && List.mem frame frames && frame <> frame' then
                           SymbolSet.add v uses
                        else
                           uses
                  with
                     Not_found ->
                        uses) uses defs
      with
         Not_found ->
            uses
   in
      SymbolTable.map (fun f fun_info ->
            let { fun_label = label;
                  fun_frames = frames;
                  fun_uses = uses;
                  fun_ienv = defs
                } = fun_info
            in
               if label then
                  fun_info
               else
                  let uses = add_defs frames defs uses f in
                     { fun_info with fun_uses = uses }) funs

(*
 * Add the uses in the function and its arguments.
 *)
let dflow_add_fun_call funs uses f args =
   let add_var uses v =
      try
         let fun_info = SymbolTable.find funs v in
         let { fun_label = label;
               fun_frames = frames;
               fun_uses = uses'
             } = fun_info
         in
            SymbolSet.union uses uses'
      with
         Not_found ->
            uses
   in
   let add_atom uses = function
      AtomChar _
    | AtomInt _
    | AtomFloat _
    | AtomNil _ ->
         uses
    | AtomVar v ->
         add_var uses v
   in
   let uses = add_var uses f in
      List.fold_left add_atom uses args

(*
 * Calculate the variables that are "used" in the tailcall.
 *)
let dflow_uses_of_tailcall defs vars funs f args =
   let add_fun f =
      try
         let fun_info = SymbolTable.find funs f in
         let { fun_oenv = env; fun_uses = uses } = fun_info in
            env, uses
      with
         Not_found ->
            SymbolSet.empty, defs
   in
   let add_var uses v =
      try
         let fun_info = SymbolTable.find funs v in
            SymbolSet.union uses fun_info.fun_uses
      with
         Not_found ->
            uses
   in
   let add_atom uses = function
      AtomChar _
    | AtomInt _
    | AtomFloat _
    | AtomNil _ ->
         uses
    | AtomVar v ->
         add_var uses v
   in
   let env, uses = add_fun f in
      env, List.fold_left add_atom uses args

(*
 * Calculate the full def/use set for each function.
 * Propagate uses from known functions, and defs through
 * known functions.
 *)
let rec dflow_workset use_table workset funs =
   let dflow (workset, funs) f =
      (* Get the info about this fun *)
      let fun_info = SymbolTable.find funs f in

      (* Propagate defs down, and uses up *)
      let propagate_uses defs1 (workset, funs, uses) call =
         let defs, f', args = call in
            try
               let fun_info' = SymbolTable.find funs f' in
               let { fun_gflag = gflag;
                     fun_ienv = env';
                     fun_uses = uses';
                     fun_defs = defs'
                   } = fun_info'
               in

               (*
                * New defs for child (defs'') are:
                *   Incoming defs (defs1) +
                *   Our defs (defs) +
                *   Child defs (defs')
                * limited to variables that are bound
                * in the called function.
                *)
               let defs'' = SymbolSet.union defs' defs in
               let defs'' = SymbolSet.union defs'' defs1 in
               let defs'' = set_intersect defs'' env' in

               (*
                * New uses include
                *    All uses in the function and its arguments -
                *    The current defs
                *)
               let uses'' = dflow_add_fun_call funs uses' f' args in
               let uses''' = SymbolSet.subtract_list uses'' (SymbolSet.to_list defs) in
               let uses''' = SymbolSet.union uses uses''' in

                  (* If the defs have not changed, do nothing *)
                  if SymbolSet.equal defs'' defs' then
                     workset, funs, uses'''
                  else
                     (* Otherwise, save the new defs, and add the child to the workset *)
                     let workset = SymbolSet.add f' workset in
                     let fun_info' =
                        { fun_info' with fun_defs = defs'' }
                     in
                     let funs = SymbolTable.add funs f' fun_info' in
                        workset, funs, uses'''
            with
               Not_found ->
                  (*
                   * Even if the called function is not known,
                   * add the uses in the call.
                   *)
                  let uses' = dflow_add_fun_call funs SymbolSet.empty f' args in
                  let uses' = SymbolSet.subtract_list uses' (SymbolSet.to_list defs) in
                  let uses = SymbolSet.union uses uses' in
                     workset, funs, uses
      in

      (* Propagate *)
      let { fun_defs = defs; fun_uses = uses; fun_calls = calls } = fun_info in
      let workset, funs, uses' =
         List.fold_left (propagate_uses defs) (workset, funs, uses) calls
      in
         (* If the uses change, add the callers to the workset *)
         if SymbolSet.equal uses' uses then
            workset, funs
         else
            let workset = dflow_add_callers workset use_table f in
            let fun_info =
               { fun_info with fun_uses = uses';
                               fun_calls = calls
               }
            in
            let funs = SymbolTable.add funs f fun_info in
               workset, funs
   in

   (* Update all the funs in the workset *)
   let worklist = SymbolSet.to_list workset in
      if !debug_esc then
         begin
            Format.open_hvbox 3;
            Format.print_string "Worklist: ";
            List.iter (fun v ->
                  Format.print_space ();
                  print_symbol v) worklist;
            Format.print_newline ()
         end;
      if worklist <> [] then
         (* Recurse if there is still more work to do *)
         let workset, funs =
            List.fold_left dflow (SymbolSet.empty, funs) worklist
         in
            dflow_workset use_table workset funs
      else
         funs

(*
 * Print the dataflow.
 *)
let tabstop = 3

let print_symbol_set name vars =
   Format.print_space ();
   Format.open_hvbox tabstop;
   Format.print_string name;
   SymbolSet.iter (fun v ->
         Format.print_space ();
         print_symbol v) vars;
   Format.close_box ()

let print_dflow funs =
   Format.open_vbox 0;
   SymbolTable.iter (fun f fun_info ->
         let { fun_ienv = env;
               fun_defs = defs;
               fun_uses = uses
             } = fun_info
         in
            Format.print_space ();
            Format.open_vbox tabstop;
            Format.print_string "Function: ";
            print_symbol f;
            print_symbol_set "Env: " env;
            print_symbol_set "Defs: " defs;
            print_symbol_set "Uses: " uses;
            Format.close_box ()) funs;
   Format.close_box ()

(*
 * Dataflow.
 * Calculate the use_table, add all the
 * funs to the workset, then turn the crank until
 * the fixpoint is reached.
 *)
let dflow esc =
   let { esc_vars = vars; esc_funs = funs } = esc in
   let use_table = dflow_use_table funs in
   let workset =
      SymbolTable.fold (fun workset f _ ->
            SymbolSet.add f workset) SymbolSet.empty funs
   in
   let funs = dflow_add_unknown_uses esc.esc_vars funs in
   let funs = dflow_workset use_table workset funs in
   let funs =
      SymbolTable.fold (fun funs f entries ->
            SymbolTable.add funs f (List.hd entries)) (**)
         (SymbolTable.create ()) funs
   in
      if !debug_esc then
         begin
            print_dflow funs;
            Format.print_newline ()
         end;
      { esc with esc_funs = funs }

(************************************************************************
 * SECOND PASS: ALLOCATE FRAMES
 ************************************************************************)

(*
 * Conversion requires a set of variables that
 * escape, as well as the types of the frames.
 *)
type aelim =
   { aelim_esc : (ty * frame * bool) SymbolTable.t;
     aelim_frames : (var * ty) list SymbolTable.t;
     aelim_funs : fun_info SymbolTable.t
   }

(*
 * Convert the escaping vars.
 *)
let aelim_of_esc { esc_vars = vars; esc_funs = esc_funs } =
   (* Collect all the escaping vars and frame fields *)
   let esc, frames =
      SymbolTable.fold (fun (esc, frames) v entry ->
            let ty, frame, aflag = List.hd entry in
            let esc = SymbolTable.add esc v (ty, frame, aflag) in
            let frames =
               try
                  let fields = SymbolTable.find frames frame in
                  let fields = (v, ty) :: fields in
                     SymbolTable.add frames frame fields
               with
                  Not_found ->
                     SymbolTable.add frames frame [v, ty]
            in
               esc, frames) (SymbolTable.create (), SymbolTable.create ()) vars
   in
      { aelim_esc = esc; aelim_frames = frames; aelim_funs = esc_funs }

(*
 * Elim types.
 * Functions never take aggregate arguments.
 *)
let rec aelim_type tenv pos ty =
   let pos = string_pos "aelim_type" pos in
   match ty with
    | TyAny | TyChar | TyInt | TyFloat
    | TyId _ -> ty
    | TyArray (ty, dim) ->
         TyArray (aelim_type tenv pos ty, dim)
    | TyStruct fields ->
         let fields =
            List.map (fun (v, ty) -> v, aelim_type tenv pos ty) fields
         in
            TyStruct fields
    | TyFun (ty_vars, ty_ret) ->
         let to_pointer ty =
            let ty = aelim_type tenv pos ty in
               if is_aggregate_type tenv pos ty then
                  TyArray (ty, None)
               else ty
         in
         let ty_vars = List.map to_pointer ty_vars in
            TyFun (ty_vars, aelim_type tenv pos ty_ret)

(*
 * Type of the aggregate.
 *)
let type_of_aggregate tenv pos ty =
   let ty = aelim_type tenv pos (tenv_expand tenv pos ty) in
      match ty with
         TyArray (ty, Some _) -> TyArray (ty, None)
       | _ -> TyArray (ty, None)

(*
 * Lookup a variable.
 * Expand the type if the variable escapes.
 *)
type aelim_type =
   AggrType of ty * frame
 | ScalarType of ty * frame
 | NoType

let venv_lookup_var tenv esc pos v =
   try
      let ty, frame, aflag = SymbolTable.find esc.aelim_esc v in
      let ty = aelim_type tenv pos ty in
         if is_aggregate_type tenv pos ty then
            AggrType (type_of_aggregate tenv pos ty, frame)
         else if aflag then
            ScalarType (ty, frame)
         else
            NoType
   with
      Not_found ->
         NoType

(*
 * Look up the frame for a function.
 *)
let venv_lookup_fun esc pos f =
   try SymbolTable.find esc.aelim_funs f with
      Not_found ->
         raise (IRException (pos, UnboundVar f))

(*
 * Look up the type for a frame.
 *)
let venv_lookup_frame tenv esc pos frame =
   try
      let fields = SymbolTable.find esc.aelim_frames frame in
         List.map (fun (v, ty) -> v, aelim_type tenv pos ty) fields
   with
      Not_found ->
         []

(*
 * "Store" environment.
 * Keep track of variables that are defined.
 *)
let senv_empty =
   SymbolSet.empty

let senv_add senv v =
   SymbolSet.add v senv

let senv_add_vars senv vars =
   List.fold_left senv_add senv vars

(*
 * Save all modified scalars to the frame.
 *)
let save_scalars tenv esc pos defs env uses exp =
   let { aelim_esc = vars } = esc in
      SymbolSet.fold (fun exp v ->
            try
               let ty, frame, aflag = SymbolTable.find vars v in
                  if not aflag && SymbolSet.mem v uses && not (SymbolSet.mem v env) then
                     let ty = aelim_type tenv pos ty in
                        SetProject (AtomVar frame, v, ty, AtomVar v, exp)
                  else
                     exp
            with
               Not_found ->
                  exp) exp defs

(*
 * Load all the free scalars from the frame.
 *)
let restore_scalars tenv esc pos uses exp =
   let { aelim_esc = vars } = esc in
      SymbolSet.fold (fun exp v ->
            try
               let ty, frame, aflag = SymbolTable.find vars v in
                  if aflag then
                     exp
                  else
                     let ty = aelim_type tenv pos ty in
                     let v' = new_symbol v in
                        LetProject (v', ty, AtomVar frame, v,
                                    SetVar (v, ty, AtomVar v', exp))
            with
               Not_found ->
                  exp) exp uses

(*
 * Atom conversion.
 *)
let aelim_var tenv esc pos v cont =
   let pos = string_pos "aelim_var" pos in
      match venv_lookup_var tenv esc pos v with
         AggrType (ty, frame) ->
            LetAddrOfProject (v, ty, AtomVar frame, v, cont v)
       | ScalarType (ty, frame) ->
            LetProject (v, ty, AtomVar frame, v, cont v)
       | NoType ->
            cont v

let aelim_atom tenv esc pos a cont =
   let pos = string_pos "aelim_atom" pos in
      match a with
       | AtomChar _ | AtomInt _ | AtomFloat _ | AtomNil _ ->
            cont a
       | AtomVar v ->
            match venv_lookup_var tenv esc pos v with
             | AggrType (ty, frame) ->
                  LetAddrOfProject (v, ty, AtomVar frame, v, cont a)
             | ScalarType (ty, frame) ->
                  LetProject (v, ty, AtomVar frame, v, cont a)
             | NoType ->
                  cont a

let aelim_atoms tenv esc pos atoms cont =
   let rec loop args = function
      a :: atoms ->
         aelim_atom tenv esc pos a (fun a ->
               loop (a :: args) atoms)
    | [] ->
         cont (List.rev args)
   in
      loop [] atoms

let aelim_atom_opt tenv esc pos a cont =
   match a with
      Some a ->
         aelim_atom tenv esc pos a (fun a -> cont (Some a))
    | None ->
         cont None

(*
 * Transform an expression.
 *)
let rec aelim_expr tenv senv esc e =
   let pos = atom_pos e in
      match e with
         LetTypes (types, e) ->
            aelim_types_expr tenv senv esc pos types e
       | LetFuns2 (funs, e) ->
            aelim_funs_expr tenv senv esc pos funs e
       | LetAtom (v, ty, a, e) ->
            aelim_atom_expr tenv senv esc pos v ty a e
       | LetCopy (v, ty, a, e) ->
            aelim_copy_expr tenv senv esc pos v ty a e
       | LetUnop (v, ty, op, a, e) ->
            aelim_unop_expr tenv senv esc pos v ty op a e
       | LetBinop (v, ty, op, a1, a2, e) ->
            aelim_binop_expr tenv senv esc pos v ty op a1 a2 e
       | TailCall (f, args) ->
            aelim_tailcall_expr tenv senv esc pos f args
       | LetExtCall (v, f, ty, args, e) ->
            aelim_extcall_expr tenv senv esc pos v f ty args e
       | LetString (v, s, e) ->
            aelim_string_expr tenv senv esc pos v s e
       | IfThenElse (op, a1, a2, e1, e2) ->
            aelim_if_expr tenv senv esc pos op a1 a2 e1 e2
       | SetVar (v, ty, a, e) ->
            aelim_set_var_expr tenv senv esc pos v ty a e
       | LetAddrOfVar (v1, ty, v2, e) ->
            aelim_addr_of_var_expr tenv senv esc pos v1 ty v2 e
       | SetSubscript (a1, ty, a2, a3, e) ->
            aelim_set_subscript_expr tenv senv esc pos ty a1 a2 a3 e
       | LetSubscript (v, ty, a1, a2, e) ->
            aelim_subscript_expr tenv senv esc pos v ty a1 a2 e
       | LetAddrOfSubscript (v, ty, a1, a2, e) ->
            aelim_addr_of_subscript_expr tenv senv esc pos v ty a1 a2 e
       | SetProject (a1, l, ty, a2, e) ->
            aelim_set_project_expr tenv senv esc pos ty a1 l a2 e
       | LetProject (v, ty, a, l, e) ->
            aelim_project_expr tenv senv esc pos v ty a l e
       | LetAddrOfProject (v, ty, a, l, e) ->
            aelim_addr_of_project_expr tenv senv esc pos v ty a l e
       | LetFuns _
       | LetApply _
       | Return _
       | Memcpy _
       | LetAlloc _
       | LetClosure _
       | IntCall _ ->
            raise (IRException (pos, IRLevel 1))

(*
 * Add the type definitions.
 *)
and aelim_types_expr tenv senv esc pos types e =
   let pos = string_pos "aelim_types_expr" pos in
   let types = List.map (fun (v, ty) -> v, aelim_type tenv pos ty) types in
   let tenv =
      List.fold_left (fun tenv (v, ty) -> tenv_add tenv v ty) tenv types
   in
   let e = aelim_expr tenv senv esc e in
      LetTypes (types, e)

(*
 * Function definitions.
 * For global functions, allocate the frame,
 * copy the aggregate arguments to the frame, and
 * convert the body.
 *)
and aelim_fun_global tenv senv esc pos f ty vars body =
   let pos = string_pos "aelim_fun_global" pos in

   (* Get the function info *)
   let { fun_frames = frames;
         fun_gflag = gflag;
         fun_venv = env;
         fun_defs = defs;
         fun_uses = uses
       } = venv_lookup_fun esc pos f
   in

   (* Add the var definitions *)
   let senv = senv_add_vars defs vars in

   (* Convert the function body *)
   let body = aelim_expr tenv senv esc body in

   (* Copy all the aggregate arguments *)
   let ty_vars, ty_res = dest_fun_type tenv pos ty in
   let body =
      List.fold_left2 (fun body v ty ->
            match venv_lookup_var tenv esc pos v with
             | AggrType (ty', frame) ->
                  let v' = new_symbol v in
                     LetAddrOfProject (v', ty', AtomVar frame, v,
                                       Memcpy (AtomVar v', ty, AtomVar v, body))
                                               (*LetAtom (v, ty', AtomVar v', body)))???*)
             | ScalarType (ty, frame) ->
                  SetProject (AtomVar frame, v, ty, AtomVar v, body)
             | NoType ->
                  body) body vars ty_vars
   in
      (* Allocate the frame *)
      match frames with
         frame :: parent :: rest ->
            let fields = venv_lookup_frame tenv esc pos frame in

            (* Load all the scalar variables *)
            let uses = SymbolSet.subtract_list (set_intersect uses env) (SymbolSet.to_list defs) in
            let body = restore_scalars tenv esc pos uses body in

            (* Add all the inner frames to this frame *)
            let fields =
               List.fold_left (fun fields v -> (v, TyArray (TyId v, None)) :: fields) fields (parent :: rest)
            in

            (* Copy the frame pointers *)
            let rec copy_fields = function
               v :: rest ->
                  let ty = TyArray (TyId v, None) in
                     LetProject (v, ty, AtomVar parent, v,
                                 SetProject (AtomVar frame, v, ty, AtomVar v, copy_fields rest))
             | [] ->
                  let ty = TyArray (TyId parent, None) in
                     SetProject (AtomVar frame, parent, ty, AtomVar parent, body)
            in
            let body = copy_fields rest in

            (* Allocate the frame *)
            let ty_frame = TyStruct fields in
            let body =
               LetTypes ([frame, ty_frame], LetAlloc (frame, TyId frame, body))
            in
            let ty = aelim_type tenv pos ty in
            let gflag =
               if gflag = FunGlobalClass && rest = [] then
                  FunGlobalClass
               else
                  FunContClass
            in
               f, gflag, ty, vars, body
       | _ ->
            raise (IRException (pos, InternalError "frame stack is empty"))

(*
 * A continuation needs to restore the scalars from the
 * frame.
 *)
and aelim_fun_cont tenv senv esc pos f ty vars body =
   (* Get the frame *)
   let pos = string_pos "aelim_fun_cont" pos in

   (* Get the function info *)
   let { fun_frames = frames;
         fun_gflag = gflag;
         fun_venv = env;
         fun_defs = defs;
         fun_uses = uses
       } = venv_lookup_fun esc pos f
   in

   (* Add the var definitions *)
   let senv = senv_add_vars defs vars in

   (* Convert the body *)
   let body = aelim_expr tenv senv esc body in

   (* Copy all the aggregate arguments *)
   let ty_vars, ty_res = dest_fun_type tenv pos ty in
   let body =
      List.fold_left2 (fun body v ty ->
            match venv_lookup_var tenv esc pos v with
               AggrType (ty', frame) ->
                  let v' = new_symbol v in
                  let ty' = type_of_aggregate tenv pos ty in
                     LetAddrOfProject (v', ty', AtomVar frame, v,
                                       Memcpy (AtomVar v', ty, AtomVar v,
                                               LetAtom (v, ty', AtomVar v', body)))
             | ScalarType (ty, frame) ->
                  SetProject (AtomVar frame, v, ty, AtomVar v, body)

             | NoType ->
                  body) body vars ty_vars
   in
      (* Allocate the frame *)
      match frames with
         frame :: rest ->
            let fields = venv_lookup_frame tenv esc pos frame in

            (* Restore all the scalar variables *)
            let uses = SymbolSet.subtract_list (set_intersect uses env) (SymbolSet.to_list defs) in
            let body = restore_scalars tenv esc pos uses body in

            (* Copy the frame pointers *)
            let rec copy_fields = function
               v :: rest ->
                  let ty = TyArray (TyId v, None) in
                  let v' = new_symbol v in
                     LetProject (v', ty, AtomVar frame, v,
                                 SetVar (v, ty, AtomVar v', copy_fields rest))
             | [] ->
                  body
            in
            let body = copy_fields rest in

            (* Convert the function type *)
            let ty = aelim_type tenv pos ty in
               f, FunContClass, ty, vars, body

       | [] ->
            raise (IRException (pos, InternalError "frame stack is empty"))

(*
 * Local functions are the simplest,
 * the body is inlined normally.
 *)
and aelim_fun_local tenv senv esc pos f ty vars body =
   (* Get the frame *)
   let pos = string_pos "aelim_fun_local" pos in

   (* Get the function info *)
   let { fun_defs = defs } = venv_lookup_fun esc pos f in
   let senv = senv_add_vars defs vars in

   (* Convert the body *)
   let body = aelim_expr tenv senv esc body in

   (* Copy all the aggregate arguments *)
   let ty_vars, ty_res = dest_fun_type tenv pos ty in
   let body =
      List.fold_left2 (fun body v ty ->
            match venv_lookup_var tenv esc pos v with
               AggrType (ty', frame) ->
                  let v' = new_symbol v in
                  let ty' = type_of_aggregate tenv pos ty in
                     LetAddrOfProject (v', ty', AtomVar frame, v,
                                       Memcpy (AtomVar v', ty, AtomVar v,
                                               LetAtom (v, ty', AtomVar v', body)))
             | ScalarType (ty, frame) ->
                  SetProject (AtomVar frame, v, ty, AtomVar v, body)

             | NoType ->
                  body) body vars ty_vars
   in

   (* Convert the function type *)
   let ty = aelim_type tenv pos ty in
      f, FunLocalClass, ty, vars, body

and aelim_funs_expr tenv senv esc pos funs e =
   (* Add add the functions to the variable environment *)
   let pos = string_pos "aelim_funs_expr" pos in

   (* Convert all the funs *)
   let funs =
      List.map (fun (f, gflag, ty, vars, body) ->
            let aelim_fun =
               match gflag with
                  FunGlobalClass ->
                     aelim_fun_global
                | FunContClass ->
                     aelim_fun_cont
                | FunDeclClass
                | FunLocalClass ->
                     aelim_fun_local
            in
               aelim_fun tenv senv esc (int_pos 2 pos) f ty vars body) funs
   in

   (* Convert the body *)
   let e = aelim_expr tenv senv esc e in
      LetFuns2 (funs, e)

(*
 * Atom assignment.  If the value is an aggregate,
 * it becomes a pointer.
 *)
and aelim_atom_expr tenv senv esc pos v ty a e =
   let pos = string_pos "aelim_atom_expr" pos in
   let ty = aelim_type tenv pos ty in
   let e = aelim_expr tenv senv esc e in
      aelim_atom tenv esc pos a (fun a ->
            if is_aggregate_type tenv pos ty then
               LetAtom (v, type_of_aggregate tenv pos ty, a, e)
            else
               match venv_lookup_var tenv esc pos v with
                  AggrType (ty', frame) ->
                     let v' = new_symbol v in
                        LetAddrOfProject (v', ty', AtomVar frame, v,
                                          Memcpy (AtomVar v', ty, a, e))
                | ScalarType (ty, frame) ->
                     SetProject (AtomVar frame, v, ty, a, e)
                | NoType ->
                     LetAtom (v, ty, a, e))

(*
 * Copy expression is an allocation point.
 * If the value is an aggregate, copy it with Memcpy.
 *)
and aelim_copy_expr tenv senv esc pos v ty a e =
   (* Compute the rest using the current type *)
   let pos = string_pos "aelim_copy_expr" pos in
   let ty = aelim_type tenv pos ty in
   let senv = senv_add senv v in
   let e = aelim_expr tenv senv esc e in
      aelim_atom_opt tenv esc pos a (fun a ->
         try
            let a =
               match a with
                | None -> default_value tenv pos ty
                | Some a -> a
            in
               match venv_lookup_var tenv esc pos v with
                | AggrType (ty', frame) ->
                     let v' = new_symbol v in
                        LetAddrOfProject (v', ty', AtomVar frame, v,
                                          Memcpy (AtomVar v', ty, a, e))
                | ScalarType (ty, frame) -> (*??? If a is None, need it be init? *)
                     SetProject (AtomVar frame, v, ty, a, e)
                | NoType ->
                     LetAtom (v, ty, a, e)
         with
            IRException (_, CantInitialize _) -> e
      )

(*
 * Unary operation.
 * Arguments are always of scalar type, so
 * we don't worry about aggregates.
 *)
and aelim_unop_expr tenv senv esc pos v ty op a e =
   let pos = string_pos "aelim_unop_expr" pos in
   let ty = aelim_type tenv pos ty in
   let e = aelim_expr tenv senv esc e in
      aelim_atom tenv esc pos a (fun a ->
            match venv_lookup_var tenv esc pos v with
               AggrType _ ->
                  raise (IRException (pos, InternalError "aelim_unop_expr"))
             | ScalarType (ty, frame) ->
                  let v' = new_symbol v in
                     LetUnop (v', ty, op, a,
                              SetProject (AtomVar frame, v, ty, AtomVar v', e))
             | NoType ->
                  LetUnop (v, ty, op, a, e))

(*
 * Binary operation.
 * Arguments are always of scalar type.
 *)
and aelim_binop_expr tenv senv esc pos v ty op a1 a2 e =
   let pos = string_pos "aelim_binop_expr" pos in
   let ty = aelim_type tenv pos ty in
   let e = aelim_expr tenv senv esc e in
      aelim_atom tenv esc pos a1 (fun a1 ->
            aelim_atom tenv esc pos a2 (fun a2 ->
                  match venv_lookup_var tenv esc pos v with
                     AggrType _ ->
                        raise (IRException (pos, InternalError "aelim_binop_expr"))
                   | ScalarType (ty, frame) ->
                        let v' = new_symbol v in
                           LetBinop (v', ty, op, a1, a2,
                                     SetProject (AtomVar frame, v, ty, AtomVar v', e))
                   | NoType ->
                        LetBinop (v, ty, op, a1, a2, e)))

(*
 * Tailcall.
 * Check the function type.
 * If it returns an aggregate, we need to
 * pass the return argument.
 *)
and aelim_tailcall_expr tenv senv esc pos f args =
   let pos = string_pos "aelim_tailcall_expr" pos in
   let exp =
      aelim_atoms tenv esc pos args (fun args ->
            aelim_var tenv esc pos f (fun f ->
                  TailCall (f, args)))
   in

   (*
    * Calculate all the uses, and all the variables
    * bound in the destination environment.  Any of
    * the variables that have been modified and are
    * not bound, and are eventually used, must be written
    * to the frame.
    *)
   let env, uses = dflow_uses_of_tailcall senv esc.aelim_esc esc.aelim_funs f args in
      save_scalars tenv esc pos senv env uses exp

(*
 * External function call.
 *)
and aelim_extcall_expr tenv senv esc pos v f ty args e =
   let pos = string_pos "aelim_extcall_expr" pos in
   let _, ty_res = dest_fun_type tenv pos ty in
   let ty_res = aelim_type tenv pos ty_res in
   let ty = aelim_type tenv pos ty in
   let e = aelim_expr tenv senv esc e in
      aelim_atoms tenv esc pos args (fun args ->
            match venv_lookup_var tenv esc pos v with
               AggrType (ty', frame) ->
                  let v' = new_symbol v in
                  let v'' = new_symbol v in
                  let ty_vars, _ = dest_fun_type tenv pos ty in
                  let ty = TyFun (ty_vars, ty') in
                     LetExtCall (v', f, ty, args,
                                 LetAddrOfProject (v'', ty', AtomVar frame, v,
                                                   Memcpy (AtomVar v'', ty_res, AtomVar v', e)))
             | ScalarType (ty', frame) ->
                  let v' = new_symbol v in
                     LetExtCall (v', f, ty, args,
                                 SetProject (AtomVar frame, v, ty', AtomVar v', e))
             | NoType ->
                  LetExtCall (v, f, ty, args, e))

(*
 * String allocation.
 *)
and aelim_string_expr tenv senv esc pos v s e =
   let pos = string_pos "aelim_string_expr" pos in
   let ty = TyArray (TyChar, None) in
   let e = aelim_expr tenv senv esc e in
      match venv_lookup_var tenv esc pos v with
         AggrType _ ->
            raise (IRException (pos, InternalError "aelim_string_expr"))
       | ScalarType (ty, frame) ->
            let v' = new_symbol v in
               LetString (v', s, SetProject (AtomVar frame, v, ty, AtomVar v', e))
       | NoType ->
            LetString (v, s, e)

(*
 * Conditional.
 *)
and aelim_if_expr tenv senv esc pos op a1 a2 e1 e2 =
   let pos = string_pos "aelim_if_expr" pos in
   let e1 = aelim_expr tenv senv esc e1 in
   let e2 = aelim_expr tenv senv esc e2 in
      aelim_atom tenv esc pos a1 (fun a1 ->
            aelim_atom tenv esc pos a2 (fun a2 ->
                  IfThenElse (op, a1, a2, e1, e2)))

(*
 * Set a variable.
 * Add the variable to the senv (if the variable escapes).
 *)
and aelim_set_var_expr tenv senv esc pos v ty a e =
   let pos = string_pos "aelim_set_var_expr" pos in
   let senv = senv_add senv v in
   let e = aelim_expr tenv senv esc e in
   let ty = aelim_type tenv pos ty in
      aelim_atom tenv esc pos a (fun a ->
            match venv_lookup_var tenv esc pos v with
               AggrType (ty', frame) ->
                  let v' = new_symbol v in
                     LetAddrOfProject (v', ty', AtomVar frame, v,
                                       Memcpy (AtomVar v', ty, a, e))
             | ScalarType (ty, frame) ->
                  SetProject (AtomVar frame, v, ty, a, e)
             | NoType ->
                  SetVar (v, ty, a, e))

(*
 * Address of a variable is eliminated.
 * The variable better have escaped.
 *)
and aelim_addr_of_var_expr tenv senv esc pos v1 ty v2 e =
   (* Add the variable to the escape set *)
   let pos = string_pos "aelim_addr_of_var_expr" pos in
   let ty = aelim_type tenv pos ty in
   let e = aelim_expr tenv senv esc e in
   let inner ty' v' =
      match venv_lookup_var tenv esc pos v1 with
         AggrType _ ->
            raise (IRException (pos, InternalError "aelim_addr_of_var_expr1"))
       | ScalarType (ty, frame1) ->
            SetProject (AtomVar frame1, v1, ty, AtomVar v', e)
       | NoType ->
            LetAtom (v1, ty', AtomVar v', e)
   in
   let v' = new_symbol v2 in
      match venv_lookup_var tenv esc pos v2 with
         AggrType (ty', frame2) ->
            LetAddrOfProject (v', ty', AtomVar frame2, v2, inner ty' v')
       | ScalarType (_, frame2) ->
            LetAddrOfProject (v', ty, AtomVar frame2, v2, inner ty v')
       | NoType ->
            raise (IRException (pos, InternalError "aelim_addr_of_var_expr2"))

(*
 * Set an array entry.
 *)
and aelim_set_subscript_expr tenv senv esc pos ty a1 a2 a3 e =
   let pos = string_pos "aelim_set_subscript_expr" pos in
   let ty = aelim_type tenv pos ty in
   let e = aelim_expr tenv senv esc e in
      aelim_atom tenv esc pos a1 (fun a1 ->
            aelim_atom tenv esc pos a2 (fun a2 ->
                  aelim_atom tenv esc pos a3 (fun a3 ->
                        if is_aggregate_type tenv pos ty then
                           let v' = new_symbol_string "aelim_set_subscript_expr" in
                           let ty' = type_of_aggregate tenv pos ty in
                              LetAddrOfSubscript (v', ty', a1, a2,
                                                  Memcpy (AtomVar v', ty, a3, e))
                        else
                           SetSubscript (a1, ty, a2, a3, e))))

and aelim_subscript_expr tenv senv esc pos v ty a1 a2 e =
   let pos = string_pos "aelim_subscript_expr" pos in
   let ty = aelim_type tenv pos ty in
   let e = aelim_expr tenv senv esc e in
      aelim_atom tenv esc pos a1 (fun a1 ->
            aelim_atom tenv esc pos a2 (fun a2 ->
                  match venv_lookup_var tenv esc pos v with
                     AggrType (ty', frame) ->
                        let v' = new_symbol v in
                        let v'' = new_symbol v in
                           LetAddrOfSubscript (v', ty', a1, a2,
                                               LetAddrOfProject (v'', ty', AtomVar frame, v,
                                                                 Memcpy (AtomVar v'', ty, AtomVar v', e)))
                   | ScalarType (ty, frame) ->
                        let v' = new_symbol v in
                           LetSubscript (v', ty, a1, a2,
                                         SetProject (AtomVar frame, v, ty, AtomVar v', e))
                   | NoType ->
                        LetSubscript (v, ty, a1, a2, e)))

and aelim_addr_of_subscript_expr tenv senv esc pos v ty a1 a2 e =
   let pos = string_pos "aelim_addr_of_subscript_expr" pos in
   let ty = aelim_type tenv pos ty in
   let v' = new_symbol v in
   let e = aelim_expr tenv senv esc e in
   let e =
      match venv_lookup_var tenv esc pos v with
         AggrType _ ->
            raise (IRException (pos, InternalError "aelim_addr_of_subscript_expr"))
       | ScalarType (ty, frame) ->
            SetProject (AtomVar frame, v, ty, AtomVar v', e)
       | NoType ->
            LetAtom (v, ty, AtomVar v', e)
   in
      aelim_atom tenv esc pos a1 (fun a1 ->
            aelim_atom tenv esc pos a2 (fun a2 ->
                  LetAddrOfSubscript (v', ty, a1, a2, e)))

(*
 * Struct projections.
 *)
and aelim_set_project_expr tenv senv esc pos ty a1 l a2 e =
   let pos = string_pos "aelim_set_project_expr" pos in
   let ty = aelim_type tenv pos ty in
   let e = aelim_expr tenv senv esc e in
      aelim_atom tenv esc pos a1 (fun a1 ->
            aelim_atom tenv esc pos a2 (fun a2 ->
                  if is_aggregate_type tenv pos ty then
                     let v' = new_symbol_string "aelim_set_project_expr" in
                     let ty' = type_of_aggregate tenv pos ty in
                        LetAddrOfProject (v', ty', a1, l,
                                          Memcpy (AtomVar v', ty, a2, e))
                  else
                     SetProject (a1, l, ty, a2, e)))

and aelim_project_expr tenv senv esc pos v ty a l e =
   let pos = string_pos "aelim_project_expr" pos in
   let ty = aelim_type tenv pos ty in
   let e = aelim_expr tenv senv esc e in
      aelim_atom tenv esc pos a (fun a ->
            match venv_lookup_var tenv esc pos v with
               AggrType (ty', frame) ->
                  let v' = new_symbol v in
                  let v'' = new_symbol v in
                     LetAddrOfProject (v', ty', a, l,
                                       LetAddrOfProject (v'', ty', AtomVar frame, v,
                                                         Memcpy (AtomVar v'', ty, AtomVar v', e)))
             | ScalarType (ty, frame) ->
                  let v' = new_symbol v in
                     LetProject (v', ty, a, l,
                                 SetProject (AtomVar frame, v, ty, AtomVar v', e))
             | NoType ->
                  LetProject (v, ty, a, l, e))

and aelim_addr_of_project_expr tenv senv esc pos v ty a l e =
   let pos = string_pos "aelim_addr_of_project_expr" pos in
   let ty = aelim_type tenv pos ty in
   let v' = new_symbol v in
   let e = aelim_expr tenv senv esc e in
   let e =
      match venv_lookup_var tenv esc pos v with
         AggrType _ ->
            raise (IRException (pos, InternalError "aelim_addr_of_project_expr"))
       | ScalarType (ty, frame) ->
            SetProject (AtomVar frame, v, ty, AtomVar v', e)
       | NoType ->
            LetAtom (v, ty, AtomVar v', e)
   in
      aelim_atom tenv esc pos a (fun a ->
            LetAddrOfProject (v', ty, a, l, e))

(************************************************************************
 * GLOBAL FUNCTIONS
 ************************************************************************)

(*
 * For the global function:
 *    1. Compute the escape set
 *    2. Rewrite the code
 *)
let aelim_expr tenv e =
   (* Compute the set of escaping variables *)
   let esc = esc_expr tenv empty_venv empty_esc e in
   let esc = dflow esc in

   (* Rewrite the code *)
   let aelim = aelim_of_esc esc in
   let e = aelim_expr tenv senv_empty aelim e in

   (* Get the outermost data frame *)
   let pos = atom_pos e in
   let frame, _ = pop_frame pos esc in
   let frame = frame.frame_frame in
   let ty_frame = TyStruct (venv_lookup_frame tenv aelim pos frame) in
      frame, ty_frame, standardize_expr_notype e
