(*
 * Eliminate variable aliases: that is, any instructions of
 * the form:
 *
 *    let v1 = v2 in e
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Symbol
open Fc_ir_type

(*
 * Variable alias table.
 *)
type venv = var SymbolTable.t

let venv_empty = SymbolTable.create ()

let venv_add = SymbolTable.add

(*
 * Variable aliasing.
 *)
let inline_var venv v =
   try SymbolTable.find venv v with
      Not_found ->
         v
(*
 * Atom aliasing.
 *)
let inline_atom venv a =
   match a with
      AtomChar _
    | AtomInt _
    | AtomFloat _
    | AtomNil _ ->
         a
    | AtomVar v ->
         AtomVar (inline_var venv v)

let inline_atom_opt venv = function
   Some a ->
      Some (inline_atom venv a)
 | None ->
      None

let inline_atoms venv args =
   List.map (inline_atom venv) args

(*
 * Inline all variable aliases.
 *)
let rec inline_expr venv e =
   match e with
      LetTypes (types, e) ->
         LetTypes (types, inline_expr venv e)
    | LetFuns (funs, e) ->
         let funs =
            List.map (fun (f, gflag, ty, vars, e) ->
                  f, gflag, ty, vars, inline_expr venv e) funs
         in
         let e = inline_expr venv e in
            LetFuns (funs, e)
    | LetFuns2 (funs, e) ->
         let funs =
            List.map (fun (f, gflag, ty, vars, e) ->
                  f, gflag, ty, vars, inline_expr venv e) funs
         in
         let e = inline_expr venv e in
            LetFuns2 (funs, e)
    | LetAtom (v, ty, a, e) ->
         inline_atom_expr venv v ty a e
    | LetCopy (v, ty, a, e) ->
         LetCopy (v, ty, inline_atom_opt venv a, inline_expr venv e)
    | LetUnop (v, ty, op, a, e) ->
         LetUnop (v, ty, op, inline_atom venv a, inline_expr venv e)
    | LetBinop (v, ty, op, a1, a2, e) ->
         LetBinop (v, ty, op, inline_atom venv a1, inline_atom venv a2,
                   inline_expr venv e)
    | TailCall (f, args) ->
         TailCall (f, inline_atoms venv args)
    | LetExtCall (v, f, ty, args, e) ->
         LetExtCall (v, f, ty, inline_atoms venv args, inline_expr venv e)
    | IntCall (ty, f, args) ->
         IntCall (ty, f, inline_atoms venv args)
    | LetApply (v, ty, f, args, e) ->
         LetApply (v, ty, f, inline_atoms venv args, inline_expr venv e)
    | Return a ->
         Return (inline_atom venv a)
    | IfThenElse (op, a1, a2, e1, e2) ->
         IfThenElse (op, inline_atom venv a1, inline_atom venv a2,
                     inline_expr venv e1, inline_expr venv e2)
    | SetVar (v, ty, a, e) ->
         SetVar (inline_var venv v, ty, inline_atom venv a, inline_expr venv e)
    | LetAddrOfVar (v1, ty, v2, e) ->
         LetAddrOfVar (v1, ty, v2, inline_expr venv e)
    | LetString (v, s, e) ->
         LetString (v, s, inline_expr venv e)
    | SetSubscript (a1, ty, a2, a3, e) ->
         SetSubscript (inline_atom venv a1, ty,
                       inline_atom venv a2, inline_atom venv a3,
                       inline_expr venv e)
    | LetSubscript (v, ty, a1, a2, e) ->
         LetSubscript (v, ty, inline_atom venv a1, inline_atom venv a2,
                       inline_expr venv e)
    | LetAddrOfSubscript (v, ty, a1, a2, e) ->
         LetAddrOfSubscript (v, ty, inline_atom venv a1, inline_atom venv a2,
                             inline_expr venv e)
    | SetProject (a1, l, ty, a2, e) ->
         SetProject (inline_atom venv a1, l, ty, inline_atom venv a2, inline_expr venv e)
    | LetProject (v, ty, a, l, e) ->
         LetProject (v, ty, inline_atom venv a, l, inline_expr venv e)
    | LetAddrOfProject (v, ty, a, l, e) ->
         LetAddrOfProject (v, ty, inline_atom venv a, l, inline_expr venv e)
    | Memcpy (a1, ty, a2, e) ->
         Memcpy (inline_atom venv a1, ty, inline_atom venv a2, inline_expr venv e)
    | LetAlloc (v, ty, e) ->
         LetAlloc (v, ty, inline_expr venv e)
    | LetClosure (v, ty, f, args, e) ->
         LetClosure (v, ty, f, inline_atoms venv args, inline_expr venv e)

(*
 * For atoms, remove aliases.
 *)
and inline_atom_expr venv v ty a e =
   match a with
      AtomChar _
    | AtomInt _
    | AtomFloat _
    | AtomNil _ ->
         LetAtom (v, ty, a, inline_expr venv e)
    | AtomVar v' ->
         let v' = inline_var venv v' in
         let venv = venv_add venv v v' in
            inline_expr venv e

(************************************************************************
 * GLOBAL FUNCTION
 ************************************************************************)

(*
 * Inline all variable aliases.
 *)
let inline_expr e =
   inline_expr venv_empty e

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
