(*
 * Type and variable environments.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Symbol
open Fc_ir_type
open Fc_ir_exn_type

(************************************************************************
 * TYPES
 ************************************************************************)

(*
 * Type and variable environments.
 *)
type tenv = ty SymbolTable.t
type venv = ty SymbolTable.t

(************************************************************************
 * ENVIRONMENTS
 ************************************************************************)

(*
 * New environments.
 *)
let builtin_types =
   ["char",  TyChar;
    "int",   TyInt;
    "float", TyFloat]

let builtin_vals =
   ["exit",  TyFun ([TyInt], TyInt)]

let init_normal_table env vals =
   List.fold_left (fun env (s, ty) ->
         SymbolTable.add env (Symbol.add s) ty) env vals

let empty_tenv =
   init_normal_table (SymbolTable.create ()) builtin_types

let empty_venv =
   init_normal_table (SymbolTable.create ()) builtin_vals

let global_vars = List.map (fun (v, _) -> Symbol.add v) builtin_vals

(*
 * Add a type definition.
 *)
let tenv_add tenv v ty =
   SymbolTable.add tenv v ty

(*
 * Lookup a type.
 *)
let tenv_lookup raise_exn tenv earg v =
   try SymbolTable.find tenv v with
      Not_found ->
         raise (raise_exn earg (UnboundType v))

(*
 * Expand a type one level.
 *)
let rec tenv_expand raise_exn tenv earg ty =
   match ty with
      TyAny
    | TyChar
    | TyInt
    | TyFloat
    | TyArray _
    | TyStruct _
    | TyFun _ ->
         ty
    | TyId v ->
         tenv_expand raise_exn tenv earg (tenv_lookup raise_exn tenv earg v)

(*
 * Fold over the table.
 *)
let tenv_fold = SymbolTable.fold

(*
 * Add a variable to the table.
 *)
let venv_add venv v ty =
   SymbolTable.add venv v ty

(*
 * Lookup a var.
 *)
let venv_mem venv v =
   SymbolTable.mem venv v

let venv_lookup raise_exn venv earg v =
   try SymbolTable.find venv v with
      Not_found ->
         raise (raise_exn earg (UnboundVar v))

let venv_lookup_opt venv v =
   try Some (SymbolTable.find venv v) with
      Not_found ->
         None

let venv_lookup_expand raise_exn tenv venv earg v =
   tenv_expand raise_exn tenv earg (venv_lookup raise_exn venv earg v)

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
