(*
 * FIR test function.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Fc_ir_type
open Fc_ir_exn
open Fc_ir_eval

(*
 * Arguments.
 *)
let saw_input = ref false
let debug_print_ast = ref false
let debug_print_ir = ref false
let debug_check_ir = ref false
let debug_eval_ir = ref false

let usage =
   "usage: compile [options] [files...] [-- <program arguments>]"

(*
 * Environment.
 *)
type env =
   { mutable ty_tenv : Fc_ir_env.tenv;
     mutable ty_venv : Fc_ir_env.venv;
     mutable eval_tenv : Fc_ir_env.tenv;
     mutable eval_venv : Fc_ir_eval.venv
   }

let env =
   { ty_tenv = Fc_ir_typeof.empty_tenv;
     ty_venv = Fc_ir_typeof.empty_venv;
     eval_tenv = Fc_ir_typeof.empty_tenv;
     eval_venv = Fc_ir_eval.empty_venv
   }

(*
 * Program arguments.
 *)
let rev_argv = ref []

(*
 * Name for the data segment.
 * Right now, there is none, so we just
 * make up a name.
 *)
let data_sym = Symbol.new_symbol_string "data_seg"

(*
 * Parser wrappers.
 *)
let parse_prog inx =
   Fc_ast_state.push_tenv ();
   let lexbuf = Lexing.from_channel inx in
   let prog = Fc_ast_parse.prog Fc_ast_lex.main lexbuf in
   let decls = Fc_ast_state.pop_tenv () in
      match decls with
         (_, _, pos) :: _ -> Fc_ast_type.TypeDefs (decls, pos) :: prog
       | [] -> prog

let parse_stmt lexbuf =
   Fc_ast_state.push_tenv ();
   let stmt = Fc_ast_parse.topstmt Fc_ast_lex.main lexbuf in
   let decls = Fc_ast_state.pop_tenv () in
      match stmt, decls with
         Some stmt, (_, _, pos) :: _ -> [Fc_ast_type.TypeDefs (decls, pos); stmt]
       | Some stmt, [] -> [stmt]
       | None, _ -> []

(*
 * Print and typecheck the IR.
 *)
let print_check_ir tenv venv exp =
   if !debug_print_ir then
      begin
         Format.set_formatter_out_channel stdout;
         Fc_ir_util.print_expr exp;
         Format.print_newline ()
      end;
   if !debug_check_ir then
      ignore (Fc_ir_typeof.check_expr tenv venv exp)

(*
 * Read a program.
 *)
let compile_prog prog =
   (* Print the program *)
   let _ =
      if !debug_print_ast then
         begin
            Format.set_formatter_out_channel stdout;
            Fc_ast_util.print_prog prog;
            Format.print_newline ()
         end
   in

   (* Produce the IR *)
   let tenv, venv, ty, exp = Fc_ir_ast.make_prog env.ty_tenv env.ty_venv prog in
   let _ = print_check_ir env.ty_tenv env.ty_venv exp in

   (* Evaluate it *)
   let _ =
      if !debug_eval_ir then
         begin
            let tenv, venv, v = Fc_ir_eval.eval_expr env.eval_tenv env.eval_venv data_sym exp in
               Format.set_formatter_out_channel stdout;
               Format.print_string "- val = ";
               Fc_ir_eval.print_val v;
               Format.print_newline ();
               env.eval_tenv <- tenv;
               env.eval_venv <- venv
         end
   in
      env.ty_tenv <- tenv;
      env.ty_venv <- venv

(*
 * Compilation.
 *)
let compile name =
   (* Note that we compiled a file *)
   saw_input := true;

   (* Open the input channel and set the position *)
   let inx = open_in name in
   let _ = Fc_ast_state.set_current_position (name, 1, 0, 1, 0) in
      try
         let prog = parse_prog inx in
            compile_prog prog;
            close_in inx
      with
         exn ->
            Fc_ir_exn.print_exn_chan stderr exn;
            close_in inx;
            exit 1

(*
 * Evaluate the main function.
 *)
let call_main () =
   if !debug_eval_ir then
      ignore (Fc_ir_eval.eval_main env.eval_tenv env.eval_venv data_sym (Array.of_list (List.rev !rev_argv)))

(*
 * Toploop.
 *)
let toploop () =
   (* Set the filename *)
   let _ = Fc_ast_state.set_current_position ("<stdin>", 1, 0, 1, 0) in

   (* Add a return type *)
   let _ =
      let venv = Fc_ir_typeof.venv_add_return env.ty_venv TyAny in
         env.ty_venv <- venv
   in

   (* Create the lexer *)
   let lexbuf = Lexing.from_channel stdin in

   (* Parse and print the expressions *)
   let donep = ref false in
      while not !donep do
         (* Print a prompt *)
         output_string stdout "# ";
         flush stdout;

         (* Get the declaration *)
         try
            match parse_stmt lexbuf with
               [] ->
                  donep := true
             | prog ->
                  compile_prog prog
         with
            exn ->
               Fc_ir_exn.print_exn_chan stderr exn;
               flush stderr
      done

(*
 * Call the main program.
 *)
let spec =
   ["-print_ast", Arg.Unit (fun () -> debug_print_ast := true), "print expressions";
    "-print_ir", Arg.Unit (fun () -> debug_print_ir := true), "print intermediate code";
    "-check_ir", Arg.Unit (fun () -> debug_check_ir := true), "typecheck intermediate code";
    "-eval", Arg.Unit (fun () -> debug_eval_ir := true), "evaluate intermediate code";
    "-debug_eval", Arg.Unit (fun () ->
          debug_eval_ir := true;
          debug_eval := true), "use the debugger during evaluation";
    "-debug_pos", Arg.Unit (fun () -> debug_pos := true), "print position information";
    "--", Arg.Rest (fun s -> rev_argv := s :: !rev_argv), "program arguments"]

let _ =
   Arg.parse spec compile usage;
   let rest =
      if !saw_input then
         call_main
      else
         toploop
   in
      Fc_ir_exn.catch rest ()

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
